/*
 *  linux/arch/arm/mach-pxa/cache.c
 *
 *  Intel Memory Management
 *
 *  Cache Allocation API
 *
 *  Todd Brandt
 *  Copyright (c) 2004, Intel Corporation (todd.e.brandt@intel.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *

 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

#include <linux/kernel.h>
#include <linux/cache.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/module.h>
#include <linux/vmalloc.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/hardware.h>
#include <asm/cpu-single.h>
#include "imm.h"

#define L2_CACHE_ALIGN_MASK		(~(PHYS_L2CACHE_LINE_BYTES - 1))
#define L2_CACHE_ALIGN_LENGTH(x)	(((x)+(PHYS_L2CACHE_LINE_BYTES - 1)) \
					& L2_CACHE_ALIGN_MASK)
#define L2_CACHE_ALIGN_ADDRESS(x)	((x) & L2_CACHE_ALIGN_MASK)
#define L2_CACHE_LOCK_SIZE(a, l)	(L2_CACHE_ALIGN_LENGTH((a)+(l)) \
					-L2_CACHE_ALIGN_ADDRESS(a))
#define CACHE_SET(x)			(((x)&PHYS_L2CACHE_SET_MASK) \
					>>PHYS_L2CACHE_SET_SHIFT)
#define CACHE_ADDRESS(x)		((x)<<PHYS_L2CACHE_SET_SHIFT)
#define CACHE_SET_INDEX(x)		(CACHE_SET(x) \
					* PHYS_L2CACHE_WAYS_PER_SET)
#define CACHE_MAP_SIZE			(PHYS_L2CACHE_SIZE \
					/ (32*PHYS_L2CACHE_LINE_BYTES))

extern struct imm_info_t immid_list;

/* flag showing that this module was initialized */
static int module_initialized = 0;

/* a map of which entries in L2 cache are locked */
static unsigned long cache_map[CACHE_MAP_SIZE];

/* a simple value showing how many lockable bytes are left */
static u32 cache_space_left = IMM_L2CACHE_LOCK;

/*----------------------------------------------------------------------
 *
 *  High Level Call Tree:
 *
 *  [Application or Driver Call]
 * 	imm_cache_fetch_and_lock
 * 		imm_cache_lock_helper
 * 			create_cache_entry
 * 			xscale_l2_cache_fetch_and_lock
 * 			cache_left
 * 	imm_cache_allocate_and_lock
 * 		imm_cache_lock_helper
 * 			create_cache_entry
 * 			xscale_l2_cache_allocate_and_lock
 * 			cache_left
 * 	imm_cache_unlock
 * 		imm_cache_unlock_helper
 * 			remove_cache_entry
 * 			xscale_l2_cache_unlock_range
 * 			cache_left
 * 	imm_cache_invalidate
 * 		imm_cache_unlock_helper
 * 			remove_cache_entry
 * 			xscale_l2_cache_invalidate_range
 * 			cache_left
 *
 *  [Process Exit - from Policy Manager]
 * 	imm_cache_unlock_immid
 * 		imm_cache_invalidate
 * 			imm_cache_unlock_helper
 * 				remove_cache_entry
 * 				xscale_l2_cache_invalidate_range
 * 				cache_left
 *
 *  [Memory Free - from Policy Manager or SRAM API]
 * 	imm_cache_handle_unmap
 * 		imm_cache_unlock
 * 			imm_cache_unlock_helper
 * 				remove_cache_entry
 * 				xscale_l2_cache_unlock_range
 * 				cache_left
 *
 *  [Proc Filesystem]
 * 	/proc/imm/cachelock_usage (read-only)
 * 		imm_cachelock_usage_read_proc
 * 	/proc/imm/cachelock_entries (read-only)
 * 		imm_cachelock_entries_read_proc
 *
 *  [Module Initialization - from Policy Manager]
 *	imm_dau_cache_init
 *
 *--------------------------------------------------------------------*/

/*************************************************************************
 * Function: cache_left
 * Description: This function takes in the address range of a memory block
 * 	that was either locked or unlocked, determines how many actual lines
 * 	were used, and updates the available cache space value.
 *
 * Arguments:
 * entry    - the cache lock entry being created or destroyed
 * imm_info - the imm struct for this client
 * remove   - 1 = unlock, 0 = lock
 *
 * Return Value: none
 *************************************************************************/

void cache_left(struct imm_l2lock_t *entry, struct imm_info_t *imm_info,
		int remove)
{
	unsigned int vstart, vend, vp, pstart, pend;
	int i, set;

	/* vstart and vend are the cache aligned lock range */
	vstart = L2_CACHE_ALIGN_ADDRESS((u32)entry->addr);
	vend = L2_CACHE_ALIGN_LENGTH((u32)entry->addr+entry->len);

	/* use vp to loop through the pages in this range because */
	/* the physical page addresses may not be contiguous */
	for (vp = vstart&PAGE_MASK; vp < vend; vp += PAGE_SIZE) {

		/* align the range for this page to the page boundaries */
		pstart = imm_get_physical((void *)vp, imm_info->immid);
		pend = pstart + PAGE_SIZE;

		/* if this is the first page, move start forward */
		if (vp == (vstart&PAGE_MASK))
			pstart += vstart - vp;

		/* if this is the last page, move end backward */
		if ((vp + PAGE_SIZE) >= vend)
			pend -= vp + PAGE_SIZE - vend;

		imm_debug("pstart = %08X, pend = %08X\n", pstart, pend);
		/* pstart and pend now represent a valid physical lock range */
		/* loop through each cache line and find its set/way */
		for (;pstart < pend; pstart += PHYS_L2CACHE_LINE_BYTES) {

			/* find the index for the 1st way of the target set
			 * in cache_map
			 */
			set = CACHE_SET_INDEX(pstart);
			if (remove) {
				/* we're unlocking, remove the upper most
				 * used way
				 */
				for (i = set+PHYS_L2CACHE_WAYS_PER_SET-2;
					(i >= set)&&!__test_bit(i, cache_map);
					i--);
				if (i >= set) {
					/* There is a used way in this set */
					__clear_bit(i, cache_map);
					cache_space_left +=
						PHYS_L2CACHE_LINE_BYTES;
				}
			} else {
				/* we're locking, check each way looking for
				 * free space
				 */
				for (i = set;
					(i < set + PHYS_L2CACHE_WAYS_PER_SET-1)
					&&__test_bit(i, cache_map); i++);
				if (i < set+PHYS_L2CACHE_WAYS_PER_SET-1) {
					/* This is not the last way, the set
					 * isn't full
					 */
					__set_bit(i, cache_map);
					cache_space_left -=
						PHYS_L2CACHE_LINE_BYTES;
				} else {
					entry->missdata++;
				}
			}
		}
	}
}

/*************************************************************************
 *
 * Cache Entry list handlers
 *
 * remove_cache_entry: given a virtual address, find the cache entry and
 * 	remove it from the list. Pass back the actual entry to the caller
 *	as it will be used to carry out the physical unlock, and passed
 *	to cache_left to update the cache usage.
 * create_cache_entry: Given an address and length, create a new cache
 *	lock entry. Pass back the entry as it will eventually be passed to
 *	cache_left to update the cache usage.
 *
 *************************************************************************/

struct imm_l2lock_t * remove_cache_entry(u32 *addr,
					struct imm_info_t *imm_info)
{
	struct imm_l2lock_t *target = imm_info->l2lock_list.next;

        while (target != &imm_info->l2lock_list) {
                if (target->addr == addr) {
			target->prev->next = target->next;
			target->next->prev = target->prev;
			return(target);
                }
                target = target->next;
        }
	return(NULL);
}

struct imm_l2lock_t * create_cache_entry(u32 *addr, u32 len,
					struct imm_info_t *imm_info)
{
	struct imm_l2lock_t *new;

	new = kmalloc(sizeof(struct imm_l2lock_t), GFP_ATOMIC);
	new->addr = addr;
	new->len = len;
	new->missdata = 0;

	new->prev = &imm_info->l2lock_list;
	new->next = imm_info->l2lock_list.next;
	new->prev->next = new;
	new->next->prev = new;
	return(new);
}

#ifdef CONFIG_IMM_DEBUG
void debug_cache_stats(void) {
	int i, j, total[PHYS_L2CACHE_WAYS_PER_SET];

	/* print out a basic usage statistic */
	printk(KERN_DEBUG "%d%% of cache locked, %dKB available\n",
		(cache_space_left*100)/PHYS_L2CACHE_LOCK,
		cache_space_left/1024);
	for (i = 0; i < PHYS_L2CACHE_WAYS_PER_SET; i++) total[i] = 0;
	/* loop through the cache map (created by function cache_left) and */
	/* figure out which sets have what number of ways locked */
	for (i = 0; i < PHYS_L2CACHE_SIZE/PHYS_L2CACHE_LINE_BYTES;
		i+=PHYS_L2CACHE_WAYS_PER_SET) {
		for (j = i; (j < i + PHYS_L2CACHE_WAYS_PER_SET - 1)
			&&__test_bit(j, cache_map); j++);
		total[j-i]++;
	}
	/* Print out the stats, ideally only one of these bins will be full */
	/* (meaning that cache is perfectly evenly locked), but in reality */
	/* there will be some uneven focus on certain sets */
	for (i = 0; i < PHYS_L2CACHE_WAYS_PER_SET; i++) {
		printk(KERN_DEBUG "Total number of sets with %d of %d ways locked: %d\n",
			i, PHYS_L2CACHE_WAYS_PER_SET, total[i]);
	}
	for (i = 0; i < CACHE_MAP_SIZE; i++) {
		if (i%8 == 0)
			printk(KERN_DEBUG "\n%04X ", CACHE_ADDRESS(i*(32
					/ PHYS_L2CACHE_WAYS_PER_SET)));
		printk(KERN_DEBUG "%08X ", cache_map[i]);
	}
}
#endif

/*************************************************************************
 * Function: imm_cache_lock_helper
 * Description: This is the primary cache lock function. You give it an
 * 	address, length, and an immid and it'll create the entry and lock
 * 	the cache for you. Depending on which front-end calls it (fetch
 * 	and lock or allocate and lock), the fetch variable is used to alter
 *	how the actual locking is carried out.
 *
 * Arguments:
 * addr  - the virtual address of some block of memory to lock
 * len   - the length in bytes of the block of memory to lock
 * immid - the immid of the client this lock is to be asscoiated with
 * fetch - 1 = use fetch_and_lock, 0 = use allocate_and_lock
 *
 * Return Value: 0 if successful, non-zero on error
 *************************************************************************/

int imm_cache_lock_helper(void *addr, u32 len, u32 immid, u8 fetch)
{
	u32 size = L2_CACHE_LOCK_SIZE((u32)addr, len);
	struct imm_info_t *imm_info;
	struct vm_area_struct *vm_search;
	struct imm_l2lock_t *new, *c;
	u32 	vm_found, start, end;

	imm_debug("imm_cache_lock_helper(addr=0x%08X, len=%d, immid=%d,
		fetch=%d)\n", (u32)addr, len, IMMID_VALUE(immid), fetch);

	/* if this client isn't registered, forget it */
	if ((imm_info = imm_info_getentry(immid)) == NULL) {
                return -1;
        }

	/* if this module isn't initialized (very rare), forget it */
	if (!module_initialized) {
		register_error(immid, IMM_ERROR_UNINIT);
		return -1;
	}

	/* if the address is obviously out of bounds, forget it */
	if ((addr == NULL) || (IMMID_USER(immid)&&((u32)addr >= PAGE_OFFSET))){
		register_error(immid, IMM_ERROR_PERMISSION_DENIED);
		return -1;
        }

	/* if the size is invalid, fail */
        if (len < 1) {
		register_error(immid, IMM_ERROR_SIZETOOLOW);
		return -1;
	}

	/* align start and end to the L2 cache line boundary */
	start = L2_CACHE_ALIGN_ADDRESS((u32)addr);
	end = L2_CACHE_ALIGN_LENGTH((u32)addr + len);

	/* perform a test to see if this address range is permissible */
	down_read(&imm_info->mm->mmap_sem);
	vm_search = imm_info->mm->mmap;
	vm_found = 0;

	do {
		if ( (start >= vm_search->vm_start) && (start
				< vm_search->vm_end)) {
			if ( end <= vm_search->vm_end) {
				vm_found = 1;
				break;
			} else {
				start = vm_search->vm_end;
			}
		}
		vm_search = vm_search->vm_next;

	} while ( ( start >= vm_search->vm_start) && ( vm_search ) );

	up_read(&imm_info->mm->mmap_sem);

	/* if the address is not in this client's space, forget it */
	if ( !vm_found ) {
		register_error(immid, IMM_ERROR_PERMISSION_DENIED);
		return -1;
	}

	/* if the request overlaps an existing lock, forget it */
	for (c = imm_info->l2lock_list.next; c != &imm_info->l2lock_list;
			c = c->next) {
		if (((u32)addr < ((u32)c->addr+c->len))&&((u32)addr+len
				> ((u32)c->addr))) {
			register_error(immid, IMM_ERROR_RANGEOVERLAP);
			return -1;
		}
	}

	/* if we don't have enough space to accomodate the lock, forget it */
	if (size >  cache_space_left) {
		register_error(immid, IMM_ERROR_NOSPACE);
		return -1;
	}

	/* ok, everything looks good, create the entry */
	new = create_cache_entry((u32 *)addr, len, imm_info);
	imm_debug("xscale_l2_cache_lock(address %08X, len %d)\n",
			(u32)addr, len);

	/* now perform the actual lock */
	if (fetch)
		xscale_l2_cache_fetch_and_lock((u32)addr, (u32)addr + len);
	else
		xscale_l2_cache_allocate_and_lock((u32)addr, (u32)addr + len);

	/* update how much space we have left in the cache to lock */
	cache_left(new, imm_info, 0);
	imm_debug("Lock was %d%% effective, %d/%d lines lost\n",
		((len-(PHYS_L2CACHE_LINE_BYTES*new->missdata))*100)/len,
		new->missdata, len/PHYS_L2CACHE_LINE_BYTES);
#ifdef CONFIG_IMM_DEBUG
	debug_cache_stats();
#endif

	/* no errors happened, we made it */
	register_error(immid, IMM_ERROR_NONE);
	return 0;
}

/*************************************************************************
 *
 * Cache Lock Front End functions
 *
 * imm_cache_fetch_and_lock: Lock a range of memory and preserve its
 *	contents.
 * imm_cache_allocate_and_lock: Lock a range of memory but leave its
 *	contents undefined. (will take slightly less time)
 *
 *************************************************************************/

int imm_cache_fetch_and_lock(void *addr, u32 len, u32 immid)
{
	int res;

	down(&imm_sem);
	res = imm_cache_lock_helper(addr, len, immid, 1);
	up(&imm_sem);
	return res;
}

int imm_cache_allocate_and_lock(void *addr, u32 len, u32 immid)
{
	int res;

	down(&imm_sem);
	res = imm_cache_lock_helper(addr, len, immid, 0);
	up(&imm_sem);
	return res;
}

/*************************************************************************
 * Function: imm_cache_unlock_helper
 * Description: This is the primary cache unlock function. You give it an
 * 	address and an immid and it'll find the entry and unlock
 * 	the cache for you. Depending on which front-end calls it (invalidate
 * 	or unlock), the clean variable is used to alter how the actual
 *	unlocking is carried out.
 *
 * Arguments:
 * addr  - the virtual address of some block of memory to unlock
 * immid - the immid of the client this lock is asscoiated with
 * clean - 1 = use unlock, 0 = use invalidate
 * for_cleanup - used to circumvent the mmap semaphor if this unlock call
 *	is embedded in a syscall.
 *
 * Return Value: 0 if successful, non-zero on error
 *************************************************************************/

int imm_cache_unlock_helper(void *addr, u32 immid, int clean, int for_cleanup)
{
	struct imm_l2lock_t *entry;
	struct imm_info_t *imm_info;
	struct vm_area_struct *vm_search;
	u32 	vm_found, start, end;

	imm_debug("imm_cache_unlock_helper(addr=0x%08X, immid=%d, clean=%d, \
			for_cleanup=%d)\n", (u32)addr, IMMID_VALUE(immid),
			clean, for_cleanup);

	/* if this client isn't registered, forget it */
        if ((imm_info = imm_info_getentry(immid)) == NULL) {
                return -1;
        }

	/* if this module isn't initialized (very rare), forget it */
	if (!module_initialized) {
		register_error(immid, IMM_ERROR_UNINIT);
		return -1;
	}

	/* if the address is obviously out of bounds, forget it */
        if ((addr == NULL) || (IMMID_USER(immid)&&((u32)addr >= PAGE_OFFSET))) {
		register_error(immid, IMM_ERROR_PERMISSION_DENIED);
		return -1;
        }

	/* we have enough to proceed, go grab the entry */
        entry = remove_cache_entry((u32 *)addr, imm_info);

	/* if the entry doesn't exist, we're done */
	if (!entry) {
		register_error(immid, IMM_ERROR_ENTRY_NOTFOUND);
	        return -1;
	}

	/* align start and end to the L2 cache line boundary */
	start = L2_CACHE_ALIGN_ADDRESS((u32)addr);
	end = L2_CACHE_ALIGN_LENGTH((u32)addr + entry->len);

	/* perform a test to see if this address range is permissible */
	/* ultimately this should be redundant since we should never */
	/* be told to unlock an entry which is no longer in valid memory */
	/* IMM should be designed well enough to handle these cases */
	/* But, just in case, we do the test to see if we missed something */

	if (!for_cleanup) down_read(&imm_info->mm->mmap_sem);
	vm_search = imm_info->mm->mmap;
	vm_found = 0;

	do {
		if ((start >= vm_search->vm_start) && (start
				< vm_search->vm_end)) {
			if ( end <= vm_search->vm_end) {
				vm_found = 1;
				break;
			} else {
				start = vm_search->vm_end;
			}
		}
		vm_search = vm_search->vm_next;

	} while (( start >= vm_search->vm_start) && ( vm_search ));

	if (!for_cleanup) up_read(&imm_info->mm->mmap_sem);

	/* if this memory is no longer valid, then we can't clean it */
	/* we need a valid virt-phys mapping to use the unlock instructions */
	/* so if this ever happens this is very bad. We call these instances */
	/* cache leaks and we forcibly display an error message if they occur */
	/* NOTE - CACHE LEAKS ARE NEVER USER ERROR, if this case ever occurs, */
	/* IMM is at fault and must be upgraded (hopefully we've covered all */
	/* the bases) */

	if ( !vm_found ) {
		printk(KERN_DEBUG "Cache leak of %d bytes at 0x%08X\n",
			L2_CACHE_LOCK_SIZE((u32)addr, entry->len), (u32)addr);
	        kfree(entry);
		register_error(immid, IMM_ERROR_PERMISSION_DENIED);
		return(-1);
	}
	imm_debug("xscale_l2_cache_unlock(address %08X, len %d)\n",
			(u32)addr, entry->len);

	/* now perform the actual unlock or invalidate */
	if (clean)
		xscale_l2_cache_unlock_range((u32)addr,
				(u32)addr + entry->len);
	else
		xscale_l2_cache_invalidate_range((u32)addr,
				(u32)addr + entry->len);

	/* update how much space we have left in the cache to lock */
	cache_left(entry, imm_info, 1);

	/* destroy the memory used for the cache_entry */
        kfree(entry);

	/* no errors happened, we made it */
	register_error(immid, IMM_ERROR_NONE);
        return 0;
}

/*************************************************************************
 *
 * Cache UNLock Front End functions
 *
 * imm_cache_unlock: Unlock a range of memory and preserve its
 *      contents.
 * imm_cache_invalidate: UNLock a range of memory but leave its
 *      contents undefined. (will take slightly less time)
 *
 *************************************************************************/

int imm_cache_unlock(void *addr, u32 immid)
{
	int res;

	down(&imm_sem);
	res = imm_cache_unlock_helper(addr, immid, 1, 0);
	up(&imm_sem);
	return res;
}

int imm_cache_invalidate(void *addr, u32 immid)
{
	int res;

	down(&imm_sem);
	res = imm_cache_unlock_helper(addr, immid, 0, 0);
	up(&imm_sem);
	return res;
}

/*************************************************************************
 * Function: imm_cache_handle_unmap
 * Description: This function is used to destroy cache entries that happen
 *	to reside in memory that the system is about to unmap. It's needed
 *	so that we don't try to kill cache entries after their memory is
 *	gone, because this is impossible and will cause cache leaks. It
 *	should be tied to whatever syscalls get triggered to unmap memory
 *	prior to process exit.
 *
 * Arguments:
 * addr  - the virtual address of some block of memory that's disappearing
 * len   - the length in bytes of the memory block that's disappearing
 *
 * Return Value: 0 if successful, non-zero on error
 *************************************************************************/

int imm_cache_handle_unmap(unsigned long addr, unsigned long len)
{
	struct imm_info_t *imm_info;
	struct imm_l2lock_t *c;

	/* if the process this memory is a part of is not a client */
	/* of IMM, we don't care if it's going away. */
	if ((imm_info = imm_info_getentry(current->pid)) == NULL) {
		return -1;
	}

	/* check every entry to see if any of part of them reside */
	/* in this memory region */
	for (c = imm_info->l2lock_list.next; c != &imm_info->l2lock_list;
			c = c->next) {
		if ((addr < L2_CACHE_ALIGN_LENGTH((u32)c->addr+c->len))&&
			addr+len > L2_CACHE_ALIGN_ADDRESS((u32)c->addr)) {
			c = c->prev;
			/* this entry falls in the target region so
			 * we need to unlock it to preserve the contents.
			 * Even though some of it's memory is disappearing,
			 * we preserve just in case a user has locked two
			 * independant mallocs into cache using a single
			 * lock (not a smart * idea but if they do it, we
			 * can't just destroy the other malloc). Also, we
			 * disable the mmap semaphor as we are probably
			 * being called from a syscall which already has
			 * that semaphor set (leading to a hang if we don't)
			 */
			imm_cache_unlock_helper(c->next->addr,
					imm_info->immid, 0, 1);
		}
	}
	return(0);
}

/*************************************************************************
 * Function: imm_cache_unlock_immid
 * Description: This function is used to destroy all cache entries for a
 *	given client as that client is about to exit.
 *
 * Arguments:
 * immid - the immid of the client that's exitting
 *
 * Return Value: 0 if successful, non-zero on error
 *************************************************************************/

int imm_cache_unlock_immid(u32 immid)
{
	struct imm_info_t *imm_info;

	/* if this immid isn't actually a client, forget it */
	if ((!module_initialized) || ((imm_info = imm_info_getentry(immid))
		== NULL)) {
		return -1;
	}

	/* loop through and invalidate everything. We don't need to preserve */
	/* the contents because the client's whole memory space is about */
	/* to be torn down */
	while (imm_info->l2lock_list.next != &imm_info->l2lock_list)
		imm_cache_unlock_helper(imm_info->l2lock_list.next->addr,
			imm_info->immid, 0, 0);

	return 0;
}

/*************************************************************************
 *
 * PROC FS Functions for the IMM Cache API
 *
 * READ /proc/imm/cachelock_entries --> imm_cachelock_entries_read_proc
 * This is used to display all cache lock entries.
 *
 * READ /proc/imm/cachelock_usage --> imm_cachelock_usage_read_proc
 * This is used to display a usage statistic of how evenly cache has been
 * locked accross each of the Sets.
 *
 *************************************************************************/

READ_PROC_PROTO(imm_cachelock_entries_read_proc)
{
	READ_PROC_VARS
	struct imm_l2lock_t *c;
	struct imm_info_t *head = &immid_list, *p;

	READ_PROC_INIT
	/* loop through all the entries and print them out */
	for (p = head->next; p != head; p = p->next) {
		for (c = p->l2lock_list.next; c != &p->l2lock_list;
				c = c->next) {
			pprintf("ADDR(%08X) LEN(%08X) IMMID(%s - %u)\n",
				(u32)c->addr, c->len, IMMID_TYPE(p->immid),
				IMMID_VALUE(p->immid));
		}
	}
	READ_PROC_RETURN
}

READ_PROC_PROTO(imm_cachelock_usage_read_proc)
{
	READ_PROC_VARS
	int i, j, total[PHYS_L2CACHE_WAYS_PER_SET];

	READ_PROC_INIT
	/* print out a basic usage statistic */
	pprintf("%d%% of cache locked, %dKB available\n",
		(cache_space_left*100)/PHYS_L2CACHE_LOCK,
		cache_space_left/1024);
	for (i = 0; i < PHYS_L2CACHE_WAYS_PER_SET; i++) total[i] = 0;
	/* loop through the cache map (created by function cache_left) and */
	/* figure out which sets have what number of ways locked */
	for (i = 0; i < PHYS_L2CACHE_SIZE/PHYS_L2CACHE_LINE_BYTES;
		i+=PHYS_L2CACHE_WAYS_PER_SET) {
		for (j = i; (j < i + PHYS_L2CACHE_WAYS_PER_SET - 1)
			&&__test_bit(j, cache_map); j++);
		total[j-i]++;
	}
	/* Print out the stats, ideally only one of these bins will be full */
	/* (meaning that cache is perfectly evenly locked), but in reality */
	/* there will be some uneven focus on certain sets */
	for (i = 0; i < PHYS_L2CACHE_WAYS_PER_SET; i++) {
		pprintf("Total number of sets with %d of %d ways locked: %d\n",
			i, PHYS_L2CACHE_WAYS_PER_SET, total[i]);
	}
	READ_PROC_RETURN
}

void imm_dau_cache_init(struct proc_dir_entry *imm_dir)
{
	struct proc_dir_entry *entry;
	int i;

	/* the cache map starts out completely vacant */
	for (i = 0; i < CACHE_MAP_SIZE; i++)
		cache_map[i] = 0;

	/* create the PROC FS files */
	if (imm_dir) {
		REG_READ_PROC(entry, imm_dir, "cachelock_entries",
				imm_cachelock_entries_read_proc)
		REG_READ_PROC(entry, imm_dir, "cachelock_usage",
				imm_cachelock_usage_read_proc)
	}

	printk(KERN_INFO "Intel(c) Memory Management - L2 Cache Locking\n");

	module_initialized = 1;
}

EXPORT_SYMBOL(imm_cache_fetch_and_lock);
EXPORT_SYMBOL(imm_cache_allocate_and_lock);
EXPORT_SYMBOL(imm_cache_unlock);
EXPORT_SYMBOL(imm_cache_invalidate);
EXPORT_SYMBOL(imm_cache_unlock_immid);
