/*
 *  linux/arch/arm/mach-pxa/hotswap.c
 *
 *  Intel Memory Management
 *
 *  Todd Brandt
 *  Copyright (c) 2004, Intel Corporation (todd.e.brandt@intel.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.

 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/linkage.h>
#include <linux/proc_fs.h>
#include <linux/compiler.h>

#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/pgalloc.h>
#include <asm/io.h>
#include <asm/cacheflush.h>
#include "sram.h"

inline void invalidate_l1dcache_page(u32 addr)
{
	u32 end;

	for (end = addr+PAGE_SIZE; addr < end; addr += 32) {
		asm volatile("mcr p15, 0, %0, c7, c6, 1" : : "r" (addr));
	}
}

inline void update_priorities(struct imm_page_t **page_map,
			s16 pages_needed, s8 priority)
{
	int i, pc, pn = PRIORITY_INDEX(priority);
	struct imm_page_t *p;

	for (i = 0; i < pages_needed; i++) {
		p = page_map[i];
		pc = PRIORITY_INDEX(p->priority);

		if (PAGE_IS_L2CACHE(p->flag)) {
			num_l2cache_pages[pc]--;
			num_l2cache_pages[pn]++;
		} else if (PAGE_IS_SRAM(p->flag)) {
			num_sram_pages[pc]--;
			num_sram_pages[pn]++;
		} else if (PAGE_IS_DRAM(p->flag)) {
			num_dram_pages[pc]--;
			num_dram_pages[pn]++;
		}
		p->priority = priority;
	}
}

inline void reserve_pages(struct imm_page_t **page_map, s16 pages_needed)
{
	int i;
	struct imm_page_t *p;

	for (i = 0; i < pages_needed; i++) {
		p = page_map[i];
		if (PAGE_IS_MAPPED(p->flag) && !PAGE_IS_DMA(p->flag)) {
			p->flag |= IMM_INFO_RESERVED;
		}
	}
}

inline void
clear_priority_flag(struct imm_page_t **page_map, s16 pages_needed)
{
	int i;
	struct imm_page_t *p;

	for (i = 0; i < pages_needed; i++) {
		p = page_map[i];
		if (PAGE_IS_MAPPED(p->flag) && !PAGE_IS_DMA(p->flag)) {
			p->flag &= ~IMM_INFO_REPRIORITIZE;
		}
	}
}

inline u32
remove_my_pages(struct imm_page_t **page_map, s16 pages_needed, u32 immid)
{
	int i, j;

	for (i = 0, j = 0; i < pages_needed; i++) {
		if (page_map[i]->immid != immid) {
			if (j < i)
				page_map[j] = page_map[i];
			j++;
		}
	}
	return(j);
}

#define USE_VIRT_ADDR(page)	(!IMMID_USER((page)->immid)||((page)->immid \
				== current->pid))

extern void vunmap_pud_range(pgd_t *pgd, unsigned long addr,
                                        unsigned long end);
extern void xsc3_l2_cache_allocate_and_lock(u32, u32);
extern void xsc3_l2_cache_invalidate_range(u32, u32);
inline void imm_remap_page1way(struct imm_page_t *src, struct imm_page_t *dst)
{
	pgd_t *dir;
	struct vm_area_struct *vma;
	struct imm_info_t *imm_info;

	if (PAGE_IS_DRAM(dst->flag)) {
		num_free_dram_pages--;
	} else if (PAGE_IS_L2CACHE(dst->flag)) {
		xsc3_l2_cache_allocate_and_lock(dst->kern_addr,
			dst->kern_addr + PAGE_SIZE);
		cache_space_left -= PAGE_SIZE;
	} else if (PAGE_IS_SRAM(dst->flag)) {
		num_free_sram_pages--;
	}

	INCREMENT_PAGECOUNT(dst->flag, src->priority);

	if (PAGE_IS_DRAM(src->flag)) {
		num_free_dram_pages++;
	} else if (PAGE_IS_L2CACHE(src->flag)) {
		xsc3_l2_cache_invalidate_range(src->kern_addr,
			src->kern_addr + PAGE_SIZE);
		cache_space_left += PAGE_SIZE;
	} else {
		num_free_sram_pages++;
	}

	DECREMENT_PAGECOUNT(src->flag, src->priority);

	imm_info = imm_info_getentry(src->immid);
	if (unlikely(imm_info == NULL)) {
		imm_failure("imm_info_getentry error");
		return;
	}

	if (USE_VIRT_ADDR(src)) {
		copy_page((u32 *)dst->kern_addr, (u32 *)src->virt_addr);
	} else {
		copy_page((u32 *)dst->kern_addr, (u32 *)src->kern_addr);
	}
	__cpuc_flush_dcache_page((u32 *)dst->kern_addr);

	dst->virt_addr = src->virt_addr;
	dst->immid = src->immid;
	dst->priority = src->priority;
	dst->flag |= IMM_INFO_USED;
	src->virt_addr = 0;
	src->immid = 0;
	src->priority = 0;
	src->flag &= IMM_INFO_SRAM | IMM_INFO_DRAM | IMM_INFO_L2CACHE
			| IMM_INFO_RESERVED;

	imm_debug("Remapping 1way immid %d V%08X --> P%08X (I%03d) to \
		P%08X (I%03d)\n", dst->immid, dst->virt_addr, src->phys_addr,
		src->index, dst->phys_addr, dst->index);
	if (IMMID_USER(imm_info->immid)) {
		vma = find_vma(imm_info->mm, dst->virt_addr);
		if (unlikely(vma == NULL)) {
			imm_failure("find_vma failed\n");
			return;
		}
		/* first destroy the existing mapping */
		zap_page_range(vma, dst->virt_addr, PAGE_SIZE, NULL);

		/* now map in the underlying physical memory */
		if (io_remap_pfn_range(vma, dst->virt_addr, (dst->phys_addr)
			>> PAGE_SHIFT, PAGE_SIZE, PAGE_SHARED)) {
			imm_failure("io_remap_pfn_range failed\n");
		}
	} else {
		dir = pgd_offset_k(dst->virt_addr);

		/* first destroy the existing mapping (handle coherency) */
		/*flush_cache_vunmap(page->virt_addr, dst->virt_addr + PAGE_SIZE);*/
		vunmap_pud_range(dir, dst->virt_addr, dst->virt_addr + PAGE_SIZE);
		flush_tlb_kernel_range(dst->virt_addr, dst->virt_addr + PAGE_SIZE);

		/* now map in the underlying physical memory */
		if (remap_area_pages(dst->virt_addr, dst->phys_addr, PAGE_SIZE,
			L_PTE_BUFFERABLE | L_PTE_CACHEABLE)) {
			imm_failure("remap_area_pages failed\n");
		}
	}
}

u32 temp_page[PAGE_SIZE/sizeof(u32)];

inline void imm_remap_page2way(struct imm_page_t *src, struct imm_page_t *dst)
{
	int i;
	pgd_t *dir;
	u32 *saddr, *daddr, virt_addr, immid, priority;
	struct vm_area_struct *vma;
	struct imm_info_t *imm_info;
	struct imm_page_t *p;

	DECREMENT_PAGECOUNT(dst->flag, dst->priority);
	DECREMENT_PAGECOUNT(src->flag, src->priority);
	INCREMENT_PAGECOUNT(dst->flag, src->priority);
	INCREMENT_PAGECOUNT(src->flag, dst->priority);

	if (USE_VIRT_ADDR(src) && USE_VIRT_ADDR(dst)) {
		saddr = (u32 *)src->virt_addr;
		daddr = (u32 *)dst->virt_addr;
	} else if (USE_VIRT_ADDR(src) && !USE_VIRT_ADDR(dst)) {
		saddr = (u32 *)src->virt_addr;
		daddr = (u32 *)dst->kern_addr;
		invalidate_l1dcache_page((u32)daddr);
	} else if (!USE_VIRT_ADDR(src) && USE_VIRT_ADDR(dst)) {
		saddr = (u32 *)src->kern_addr;
		daddr = (u32 *)dst->virt_addr;
		invalidate_l1dcache_page((u32)saddr);
	} else {
		saddr = (u32 *)src->kern_addr;
		daddr = (u32 *)dst->kern_addr;
		invalidate_l1dcache_page((u32)saddr);
		invalidate_l1dcache_page((u32)daddr);
	}

	copy_page(temp_page, daddr);
	copy_page(daddr, saddr);
	copy_page(saddr, temp_page);

	if (!USE_VIRT_ADDR(src)) __cpuc_flush_dcache_page(saddr);
	if (!USE_VIRT_ADDR(dst)) __cpuc_flush_dcache_page(daddr);

	src->flag &= ~IMM_INFO_REPRIORITIZE;
	dst->flag &= ~IMM_INFO_REPRIORITIZE;

	virt_addr = dst->virt_addr;
	immid = dst->immid;
	priority = dst->priority;

	dst->virt_addr = src->virt_addr;
	dst->immid = src->immid;
	dst->priority = src->priority;

	src->virt_addr = virt_addr;
	src->immid = immid;
	src->priority = priority;

	imm_debug("Remapping 2way immid %d V%08X --> P%08X (I%03d)\n"
		"               immid %d V%08X --> P%08X (I%03d)\n",
		src->immid, src->virt_addr, dst->phys_addr, dst->index,
		dst->immid, dst->virt_addr, src->phys_addr, src->index);
	for (i = 0, p = src; i < 2; i++, p = dst) {
		imm_info = imm_info_getentry(p->immid);
		if (unlikely(imm_info == NULL))
		{
			imm_failure("imm_info_getentry error");
			return;
		}
		if (IMMID_USER(imm_info->immid)) {
			vma = find_vma(imm_info->mm, p->virt_addr);
			if (unlikely(vma == NULL)) {
				imm_failure("find_vma failed\n");
				return;
			}
			/* first destroy the existing mapping */
			zap_page_range(vma, p->virt_addr, PAGE_SIZE, NULL);

			/* now map in the underlying physical memory */
			if (io_remap_pfn_range(vma, p->virt_addr,
				(p->phys_addr) >> PAGE_SHIFT, PAGE_SIZE,
				PAGE_SHARED)) {
				imm_failure("io_remap_pfn_range failed\n");
			}
		} else {
			dir = pgd_offset_k(p->virt_addr);

			/* first destroy the existing mapping (handle
			 * coherency)
			 */
			vunmap_pud_range(dir, p->virt_addr, p->virt_addr
					+ PAGE_SIZE);
			flush_tlb_kernel_range(p->virt_addr, p->virt_addr
					+ PAGE_SIZE);

			/* now map in the underlying physical memory */
			if (remap_area_pages(p->virt_addr, p->phys_addr,
				PAGE_SIZE, L_PTE_BUFFERABLE
				| L_PTE_CACHEABLE)) {
				imm_failure("remap_area_pages failed\n");
			}
		}
	}
}

/*************************************************************************
 * Function: imm_remap_pages
 * Description: Hotswaps a list of pages from the src list to the dst list.
 *
 * Arguments:
 * src_page_map - an index array of IMM pages within page_list that are to
 *	be hotswapped out.
 * dst_page_map - an index array of IMM pages within page_list that are to
 *	be hotswapped to.
 * num_pages - the number of pages to eb hotswapped.
 *
 * Assumptions when entering this function:
 *	(1) The source and destination lists will be exacly the same
 *		length.
 *	(2) Remaps can be carried out in any order. If there are
 *		dependencies between two sets of hotswap lists they
 *		must be sent to this function in separate calls.
 *************************************************************************/

void imm_remap_pages(struct imm_page_t **src_map, struct imm_page_t **dst_map,
			s16 num_pages)
{
	int i;
	struct imm_page_t *src, *dst;

#ifdef CONFIG_IMM_DEBUG
	imm_debug("imm_remap_pages called with %d pages\n", num_pages);
	imm_debug("src_map:\n");
	print_page_map(50, src_map, num_pages);
	imm_debug("dst_map:\n");
	print_page_map(51, dst_map, num_pages);
#endif
	for (i = 0; i < num_pages; i++)
	{
		src = src_map[i];
		dst = dst_map[i];
		if (!PAGE_IS_MAPPED(src->flag)&&!PAGE_IS_MAPPED(dst->flag)) {
			continue;
		} else if (!PAGE_IS_MAPPED(src->flag)) {
			src = dst_map[i];
			dst = src_map[i];
		}

		if (!PAGE_IS_MAPPED(dst->flag)) {
			imm_remap_page1way(src, dst);
#ifdef CONFIG_IMM_DEBUG
			set_plist(src->index, dst->index, 2);
#endif
		} else {
			imm_remap_page2way(src, dst);
#ifdef CONFIG_IMM_DEBUG
			set_plist(src->index, dst->index, 3);
#endif
		}
	}
}

s16 find_used_pages_to_promote(struct imm_page_t **page_map, s16 pages_needed,
			u32 min_priority, struct imm_info_t *ref_client,
			u32 targetmem)
{
	int i, m, mtypes=0;
	u32 *num_pages[2];
	s16 pages_found = 0;
	s8 min_index = PRIORITY_INDEX(min_priority);
	struct imm_page_t *start[2], *end[2];
	struct imm_page_t **tmp_page_max[IMM_CLIENT_NUM_PRIORITIES];
	struct imm_page_t **tmp_page_ptr[IMM_CLIENT_NUM_PRIORITIES];
	struct imm_page_t **my_page_ptr = NULL, *p;

	imm_debug("find_used_pages_to_promote(<list>, pages_needed=%d, \
		min_priority=%d, targetmem=%s)\n", pages_needed,
		min_priority, PAGE_IS_TYPE(targetmem));

	/* we've been asked to swap pages into the free spaces */
	/* in page_map, the free spaces are of type srcflag */
	/* we'll start with the highest priority first and move */
	/* downward. For each priority we will search all of lower */
	/* memory starting with the slowest memory first */
	/* priority 1 DRAM->SRAM, priority 2 DRAM->SRAM, etc */

	switch(targetmem) {
	case IMM_INFO_L2CACHE:
		/* first we will search DRAM */
		num_pages[0] = num_dram_pages;
		start[0] = DRAMSTARTP;
		end[0] = DRAMENDP;
		/* then SRAM */
		num_pages[1] = num_sram_pages;
		start[1] = SRAMSTARTP;
		end[1] = SRAMENDP;
		mtypes = 2;
		break;
	case IMM_INFO_SRAM:
		/* first we will search DRAM */
		num_pages[0] = num_dram_pages;
		start[0] = DRAMSTARTP;
		end[0] = DRAMENDP;
		/* and that's it */
		num_pages[1] = NULL;
		start[1] = NULL;
		end[1] = NULL;
		mtypes = 1;
		break;
	default:
		return 0;
	}

	/* now we setup pointers for each priority into the source map */
	/* this way we can just funnel the pages directly to the source */
	/* map with one scan of the page_list */
	for (i = IMM_CLIENT_NUM_PRIORITIES-1; i > min_index; i--) {
		if (pages_found < pages_needed) {
			tmp_page_ptr[i] = page_map + pages_found;
			pages_found += num_pages[0][i];
			if (num_pages[1]) pages_found += num_pages[1][i];
			if (pages_found > pages_needed)
				pages_found = pages_needed;
			tmp_page_max[i] = page_map + pages_found;
			if ((ref_client)&&(i == PRIORITY_INDEX(ref_client->priority)))
				my_page_ptr = tmp_page_max[i] - 1;
		} else {
			tmp_page_ptr[i] = NULL;
			tmp_page_max[i] = NULL;
		}
	}
	pages_needed = pages_found;
	pages_found = 0;

	for (m = 0; m < mtypes; m++) {
		for (p = start[m]; (p < end[m])&&(pages_found < pages_needed); p++) {
			/* only grab mapped, non-DMA pages */
			if (PAGE_IS_MAPPED(p->flag)&&
				(PRIORITY1_IS_GREATER(p->priority, min_priority))&&
				!PAGE_IS_DMA(p->flag)) {
				/* check the whole range and store the matches */
				i = PRIORITY_INDEX(p->priority);
				if ((ref_client)&&(p->priority == ref_client->priority)&&
					(p->immid != ref_client->immid)) {
					/* for priority pages that match my client's priority */
					/* I want my pages to go before anyone else's, so I store */
					/* everyone else's starting at the end of the list. My */
					/* entries may overwrite them */
					if ((tmp_page_ptr[i])&&
						(my_page_ptr >= tmp_page_ptr[i]))
					{
						*my_page_ptr-- = p;
					}
				} else {
					if ((tmp_page_ptr[i])&&
						(tmp_page_ptr[i] < tmp_page_max[i]))
					{
						*tmp_page_ptr[i]++ = p;
						pages_found++;
					}
				}
			}
		}
	}

	if (ref_client)
	{
		i = PRIORITY_INDEX(ref_client->priority);
		if (tmp_page_max[i] > tmp_page_ptr[i])
		{
			pages_found += tmp_page_max[i]-tmp_page_ptr[i];
			page_map_reverse(tmp_page_ptr[i], tmp_page_max[i]-tmp_page_ptr[i]);
		}
	}

	if (unlikely(pages_found < pages_needed))
		imm_failure("Bad index");

#ifdef CONFIG_IMM_DEBUG
	print_page_map(66, page_map, pages_found);
#endif
	return(pages_found);
}

s16 find_used_pages_to_demote(struct imm_page_t **page_map, s16 pages_needed,
				s8 ref_priority, u32 targetmem)
{
	int i;
	u32 *num_pages;
	s16 pages_found = 0;
	s8 num_lower = PRIORITY_INDEX(ref_priority);
	struct imm_page_t *p, *start, *end;
	struct imm_page_t **tmp_page_end[IMM_CLIENT_NUM_PRIORITIES];
	struct imm_page_t **tmp_page_ptr[IMM_CLIENT_NUM_PRIORITIES];

	imm_debug("find_used_pages_to_demote(<list>, pages_needed=%d, \
		ref_priority=%d, targetmem=%s)\n", pages_needed, ref_priority,
		PAGE_IS_TYPE(targetmem));

	/* if the current process is lowest priority, there's no point */
	if (num_lower == 0) return(0);

	switch(targetmem) {
	case IMM_INFO_L2CACHE:
		num_pages = num_l2cache_pages;
		start = L2CACHESTARTP;
		end = L2CACHEENDP;
		break;
	case IMM_INFO_SRAM:
		num_pages = num_sram_pages;
		start = SRAMSTARTP;
		end = SRAMENDP;
		break;
	case IMM_INFO_DRAM:
		num_pages = num_dram_pages;
		start = DRAMSTARTP;
		end = DRAMENDP;
		break;
	default:
		return 0;
	}

	/* now we setup pointers for each priority into the source map */
	/* this way we can just funnel the pages directly to the source */
	/* map with one scan of the page_list */
	for (i = 0; i < num_lower; i++)
	{
		if (pages_found < pages_needed) {
			tmp_page_ptr[i] = page_map + pages_found;
			pages_found += num_pages[i];
			if (pages_found > pages_needed) pages_found = pages_needed;
			tmp_page_end[i] = page_map + pages_found;
		} else {
			tmp_page_ptr[i] = NULL;
			tmp_page_end[i] = NULL;
		}
	}

	pages_needed = pages_found;
	pages_found = 0;

	for (p = start; (pages_found < pages_needed)&&(p < end); p++)
	{
		/* only grab mapped, non-DMA pages of lower priority */
		if (PAGE_IS_MAPPED(p->flag)&&
			!PAGE_IS_DMA(p->flag)&&
			PRIORITY1_IS_GREATER(ref_priority, p->priority))
		{
			/* check the whole range and store the matches */
			i = PRIORITY_INDEX(p->priority);
			if ((tmp_page_ptr[i])&&
				(tmp_page_ptr[i] < tmp_page_end[i]))
			{
				*tmp_page_ptr[i]++ = p;
				pages_found++;
			}
		}
	}

	if (unlikely(pages_found < pages_needed))
		imm_failure("Bad index");

#ifdef CONFIG_IMM_DEBUG
	print_page_map(66, page_map, pages_found);
#endif
	return pages_found;
}

s16
hotswap_pages(struct imm_page_t **page_map, s16 pages_needed,
	u32 targetmem)
{
	int i;
	s8 ref_priority = IMM_CLIENT_MAX_PRIORITY;
	s16 dst_pages_found = 0;
	struct imm_page_t **dst_page_map, *p, **src_page_map = page_map;
	struct imm_page_t *start, *end;

	imm_debug("hotswap_pages(<list>, pages_needed=%d, targetmem=%s\n",
		pages_needed, PAGE_IS_TYPE(targetmem));

	/* we've been asked to swap the pages listed in page_map */
	/* into the memory type designated by targetmem */
	/* we will do so first with free pages, then mapped */
	/* lower priority pages, and finally will try the next */
	/* lower memory type if necessary, this function must not fail */
	/* L2CACHE priority 5->1, SRAM priority 5->1, etc */

	/* the lowest priority in the page list is the reference */
	for (i = 0; i < pages_needed; i++)
		if (PRIORITY1_IS_GREATER(ref_priority, page_map[i]->priority))
			ref_priority = page_map[i]->priority;

	switch(targetmem) {
	case IMM_INFO_L2CACHE:
		start = L2CACHESTARTP;
		end = L2CACHEENDP;
		break;
	case IMM_INFO_SRAM:
		start = SRAMSTARTP;
		end = SRAMENDP;
		break;
	case IMM_INFO_DRAM:
		start = DRAMSTARTP;
		end = DRAMENDP;
		break;
	default:
		return 0;
	}

	/* create a dst page map for the src page map */
	if ((dst_page_map = kmalloc(pages_needed*sizeof(struct imm_page_t *), GFP_ATOMIC)) == NULL)
		return 0;

	/* find REPRIORITIZE pages to swap into */
	for (p = start; (dst_pages_found < pages_needed)&&
		(p < end); p++)
	{
		if (PAGE_IS_MAPPED(p->flag)&&
			PAGE_IS_REPRIORITIZE(p->flag))
		{
			dst_page_map[dst_pages_found++] = p;
		}
	}

	/* find FREE pages first to swap into */
	for (p = start; (dst_pages_found < pages_needed)&&
		(p < end); p++)
	{
		if (!PAGE_IS_MAPPED(p->flag)&&
			!PAGE_IS_RESERVED(p->flag))
		{
			dst_page_map[dst_pages_found++] = p;
		}
	}

	/* map into any FREE/REPRIORITIZE pages immediately */
	if (dst_pages_found)
	{
		/* carry out the remapping of each of the pages */
		imm_remap_pages(src_page_map, dst_page_map, dst_pages_found);
	}

	/* if we found all we need in FREE pages, leave */
	if (dst_pages_found == pages_needed)
	{
		kfree(dst_page_map);
		return dst_pages_found;
	}

	/* at this point we need to also scan through mapped pages of */
	/* lower priority to fit the requested hotswap into this memory */
	/* type, first let's reset the src page list */
	src_page_map = &page_map[dst_pages_found];
	pages_needed -= dst_pages_found;
	dst_pages_found = 0;

	/* if we are targetting DRAM, we can't hotswap anything as there */
	/* are no lower levels of memory, so something went WRONG */
	if (unlikely(PAGE_IS_DRAM(targetmem))) {
		imm_failure("IMM Hotswap failure, hotswap won't work at DRAM level");
		kfree(dst_page_map);
		return(0);
	}

	dst_pages_found = find_used_pages_to_demote(dst_page_map,
		pages_needed, ref_priority, targetmem);

	switch(targetmem) {
	case IMM_INFO_L2CACHE:
#if (IMM_SRAM_PAGES > 0)
		if (dst_pages_found)
		{
			/* recursive call to this function to move out the dst pages */
			/* into lower memory so I can use them */
			hotswap_pages(dst_page_map, dst_pages_found, IMM_INFO_SRAM);
		}

		/* if there are still missing pages, we recursively call this */
		/* function to find them in the next lower memory type */
		if (dst_pages_found < pages_needed)
		{
			i = hotswap_pages(&src_page_map[dst_pages_found],
				pages_needed-dst_pages_found, IMM_INFO_SRAM);
			if (unlikely(i != pages_needed - dst_pages_found)) {
				imm_failure("IMM Hotswap failure");
			}
		}
		break;
#endif
	case IMM_INFO_SRAM:
#if (IMM_DRAM_PAGES > 0)
		if (dst_pages_found)
		{
			/* recursive call to this function to move out the dst pages */
			/* into lower memory so I can use them */
			hotswap_pages(dst_page_map, dst_pages_found, IMM_INFO_DRAM);
		}

		/* if there are still missing pages, we recursively call this */
		/* function to find them in the next lower memory type */
		if (dst_pages_found < pages_needed)
		{
			i = hotswap_pages(&src_page_map[dst_pages_found],
				pages_needed-dst_pages_found, IMM_INFO_DRAM);
			if (unlikely(i != pages_needed - dst_pages_found)) {
				imm_failure("IMM Hotswap failure");
			}
		}
		break;
#endif
	default:
		imm_failure("IMM Hotswap failure, hotswap won't work at DRAM level");
		kfree(dst_page_map);
		return(0);
	}

	/* now we should be ready to go */
	imm_remap_pages(src_page_map, dst_page_map, dst_pages_found);

	kfree(dst_page_map);
	return dst_pages_found;
}

/*************************************************************************
 * Function: imm_alloc_pages_hotswap
 * Description: Adds pages to an allocation by allocating them from lower
 * priority clients. This function only retrieves pages that are mapped
 * into a lower priority client's adress space. The caller specifies
 * what type of pages to look for (L2CACHE, SRAM, or DMA)
 *
 * Arguments:
 * page_map - an index array of IMM pages within page_list that comprise
 *	the physical side of this allocation. The map comes in with needed
 *	pages in the list set to -1. This function will fill as many of
 *	them in as possible with L2CACHE and/or SRAM.
 * imm_info - imm_info struct of the calling client.
 * pages_needed - the number of pages needed to satisfy this allocation.
 *	This number must be set to the number of pages left to satisfy
 *	the allocation.
 * flags
 *	IMM_INFO_L2CACHE to look for L2CACHE pages
 *	IMM_INFO_SRAM to look for SRAM pages
 *
 * Return Value: the number of pages successfully mapped in.
 *
 * Assumptions when entering this function:
 *	(1) There are enough SRAM and/or DRAM pages currently free to cover
 *		pages_needed pages. This function should *not* be called
 *		otherwise (this condition results in malloc fail anyway).
 *	(2) This function is being called in a strict order.
 *		(a) We are called with the L2CACHE flag *only* after free
 *		L2CACHE pages have been added to page_map.
 *		(b) We are called with the SRAM flag *only* after free
 *		SRAM pages have been added to page_map.
 *		(c) Immediately after we are called with the L2CACHE flag,
 *		the caller will be adding free SRAM pages to page_map if
 *		the current call has not satisfied all of pages_needed.
 *************************************************************************/

int imm_alloc_pages_hotswap(struct imm_page_t **page_map,
	struct imm_info_t *imm_info, s16 pages_needed)
{
	s16 pages_found = 0;
#if (IMM_L2CACHE_PAGES > 0)
	s16 l2start = 0, l2_pages_found = 0;
#endif
	s16 sramstart = 0, sram_pages_found = 0;
	struct imm_page_t *p;

#if (IMM_L2CACHE_PAGES > 0)
	/* find free L2CACHE pages */
	for (p = L2CACHESTARTP; (pages_found < pages_needed)&&
		(p < L2CACHEENDP); p++)
	{
		if (!PAGE_IS_MAPPED(p->flag))
		{
			page_map[pages_found++] = p;
			p->flag |= IMM_INFO_RESERVED;
		}
	}
#if (IMM_SRAM_PAGES > 0) || (IMM_DRAM_PAGES > 0)
	/* find mapped L2CACHE pages to swap out */
	if (pages_found < pages_needed)
	{
		l2start = pages_found;
		l2_pages_found = find_used_pages_to_demote(&page_map[l2start],
			pages_needed-pages_found, imm_info->priority, IMM_INFO_L2CACHE);
		pages_found += l2_pages_found;
		reserve_pages(&page_map[l2start], l2_pages_found);
	}
#endif
#endif

#if (IMM_SRAM_PAGES > 0)
	/* find free SRAM pages */
	for (p = SRAMSTARTP; (pages_found < pages_needed)&&
		(p < SRAMENDP); p++)
	{
		if (!PAGE_IS_MAPPED(p->flag))
		{
			page_map[pages_found++] = p;
			p->flag |= IMM_INFO_RESERVED;
		}
	}
#if (IMM_DRAM_PAGES > 0)
	/* find mapped SRAM pages to swap out */
	if (pages_found < pages_needed)
	{
		sramstart = pages_found;
		sram_pages_found = find_used_pages_to_demote(&page_map[sramstart],
			pages_needed-pages_found, imm_info->priority, IMM_INFO_SRAM);
		pages_found += sram_pages_found;
		reserve_pages(&page_map[sramstart], sram_pages_found);
	}
#endif
#endif

#if (IMM_DRAM_PAGES > 0)
	/* find free DRAM pages */
	for (p = DRAMSTARTP; (pages_found < pages_needed)&&
		(p < DRAMENDP); p++)
	{
		if (!PAGE_IS_MAPPED(p->flag))
		{
			page_map[pages_found++] = p;
			p->flag |= IMM_INFO_RESERVED;
		}
	}
#endif

#if (IMM_SRAM_PAGES > 0) && (IMM_DRAM_PAGES > 0)
	/* hotswap the mapped SRAM pages to slower memory */
	if (sram_pages_found)
	{
		hotswap_pages(&page_map[sramstart], sram_pages_found,
			IMM_INFO_DRAM);
	}
#endif

#if (IMM_L2CACHE_PAGES > 0)
	/* hotswap the mapped L2CACHE pages to slower memory */
	if (l2_pages_found)
	{
		hotswap_pages(&page_map[l2start], l2_pages_found,
#if (IMM_SRAM_PAGES > 0)
			IMM_INFO_SRAM);
#else
			IMM_INFO_DRAM);
#endif
	}
#endif

	return pages_found;
}

/*************************************************************************
 * Function: imm_malloc_dma_lowpriority
 * Description: Adds pages to an allocation by allocating them from lower
 * priority clients. This function only retrieves pages that are mapped
 * into a lower priority client's adress space.
 *
 * Arguments:
 * page_map - an index array of IMM pages within page_list that comprise
 *	the physical side of this allocation. The map comes in with needed
 *	pages in the list set to -1. This function will fill as many of
 *	them in as possible with L2CACHE and/or SRAM.
 * imm_info - imm_info struct of the calling client.
 * pages_needed - the number of pages needed to satisfy this allocation.
 *	This number must be set to the number of pages left to satisfy
 *	the allocation.
 *
 * Return Value: the number of pages successfully mapped in.
 *
 * Assumptions when entering this function:
 *	(1) There are enough SRAM and/or DRAM pages currently free to cover
 *		pages_needed pages. This function should *not* be called
 *		otherwise (this condition results in malloc fail anyway).
 *	(2) If called with the DMA flag, no previous scans have been
 *		done beyond Assumption (1). We assume that all memory is
 *		open to us as it is described in page_list.
 *************************************************************************/

int
imm_alloc_pages_hotswap_dma(struct imm_page_t **src_page_map, struct imm_info_t *imm_info,
	s16 pages_needed)
{
	int i, j;
	/* for DMA we can also hotswap out our own pages */
	s8 num_lower = PRIORITY_INDEX(imm_info->priority) + 1;
	struct imm_page_t **dst_page_map, *p;
	s16 tmp_pages_found[IMM_CLIENT_NUM_PRIORITIES];
	struct imm_page_t **tmp_page_map[IMM_CLIENT_NUM_PRIORITIES];

	/* make a single call to kmalloc and arrange the memory manually */
	if ((dst_page_map = kmalloc(num_lower*pages_needed*sizeof(struct imm_page_t *), GFP_ATOMIC)) == NULL)
		return 0;

        /* temporary priority sort memory goes here */
	for (i = 0; i < num_lower; i++) {
		tmp_page_map[i] = &dst_page_map[i*pages_needed];
 		tmp_pages_found[i] = 0;
	}

	/* look for a contiguous block of SRAM pages, mapped or unmapped */
	for (p = SRAMSTARTP; p < SRAMENDP; p++)
	{
		/* grab mapped, non-DMA pages of lower priority */
		if (PAGE_IS_MAPPED(p->flag)&&
			!PAGE_IS_DMA(p->flag)&&
			(PRIORITY1_IS_GREATER(imm_info->priority, p->priority)||
			(PRIORITY_EQUAL(p->priority, imm_info->priority)&&
			(p->immid == imm_info->immid))))
		{
			/* add the page to all temporary lists with */
			/* priorities equal to or greater than this page */
			for (i = PRIORITY_INDEX(p->priority);
				i < num_lower; i++)
			{
				/* if this priority level already has enough */
				/* pages, disregard the new one, we only need */
				/* one matching block */
				if (tmp_pages_found[i] < pages_needed)
					tmp_page_map[i][tmp_pages_found[i]++] = p;
			}
			for (i = 0; i < PRIORITY_INDEX(p->priority); i++)
			{
				/* for all lower priority lists */
				/* this page represents a gap and they need to */
				/* start over */
				if (tmp_pages_found[i] < pages_needed)
					tmp_pages_found[i] = 0;
			}
		/* also grab unmapped pages */
		} else if (!PAGE_IS_MAPPED(p->flag)) {
			for (i = 0; i < num_lower; i++)
			{
				/* this page is a wildcard that can be added to */
				/* all priority lists */
				if (tmp_pages_found[i] < pages_needed)
					tmp_page_map[i][tmp_pages_found[i]++] = p;
			}
		/* all other pages are gaps */
		} else {
			for (i = 0; i < num_lower; i++)
			{
				/* this page represents a gap and all lists need */
				/* to start over */
				if (tmp_pages_found[i] < pages_needed)
					tmp_pages_found[i] = 0;
			}
		}
	}

	/* find the lowest priority group of pages that will fit */
	for (i = 0; (i < num_lower)&&(tmp_pages_found[i] < pages_needed); i++);

	if (i < num_lower) {
		/* it's all or nothing with DMA, to get here all pages */
		/* must have been found and are contiguous in tmp_page_map[i] */

		/* first we fill out src_page_map and tag all the new pages */
		for (j = 0, p = tmp_page_map[i][0]; j < pages_needed; p++, j++)
		{
			if (!PAGE_IS_MAPPED(p->flag))
				p->flag |= IMM_INFO_RESERVED;
			src_page_map[j] = p;
		}

		/* now we have to sort the mapped pages from highest to */
		/* lowest priority for the remap process */

		/* first we reinitialize the tmp counters */
		for (i = 0; i < num_lower; i++) tmp_pages_found[i] = 0;

		/* now we split all the pages into the tmp buffers */
		for (i = 0; i < pages_needed; i++)
		{
			p = src_page_map[i];
			if (PAGE_IS_MAPPED(p->flag))
			{
				j = PRIORITY_INDEX(p->priority);
				tmp_page_map[j][tmp_pages_found[j]++] = p;
			}
		}

		/* now we put all the pages back into the highest priority */
		/* slot in order */
		for (i = num_lower-2; i >= 0; i--)
		{
			for (j = 0; j < tmp_pages_found[i]; j++)
			{
				tmp_page_map[num_lower-1][tmp_pages_found[num_lower-1]++] =
					tmp_page_map[i][j];
			}
		}

		/* if there are pages mapped, send them out */
		if (tmp_pages_found[num_lower-1])
		{
			hotswap_pages(tmp_page_map[num_lower-1],
				tmp_pages_found[num_lower-1],
				IMM_INFO_SRAM);
		}

		kfree(dst_page_map);
		return pages_needed;
	} else {
		kfree(dst_page_map);
		return 0;
	}
}

/*************************************************************************
 * Function: imm_free_pages_hotswap
 * Description: Swaps pages into space that has been freed by a call to
 * imm_free. The space freed is filled with existing pages in the
 * following order. Note that if pages are swapped in from the priority
 * that matches the calling client, its pages are mapped in first
 * before other clients with the same priority.
 *
 * A) Find DRAM pages to swap into FREE L2CACHE
 *  1) DST: FREE L2CACHE 	SRC: PRIORITY 1 DRAM
 *  2) DST: FREE L2CACHE 	SRC: PRIORITY 2 DRAM...
 *  3) DST: FREE L2CACHE 	SRC: PRIORITY N DRAM
 * B) Find SRAM pages to swap into FREE L2CACHE
 *  1) DST: FREE L2CACHE 	SRC: PRIORITY 1 SRAM
 *  2) DST: FREE L2CACHE 	SRC: PRIORITY 2 SRAM...
 *  3) DST: FREE L2CACHE 	SRC: PRIORITY N SRAM
 *
 *  perform remap
 *  add any SRC pages from B as DST pages in C
 *
 * C) Find DRAM pages to swap into FREE SRAM
 *  1) DST: FREE SRAM		SRC: PRIORITY 1 DRAM
 *  2) DST: FREE SRAM		SRC: PRIORITY 2 DRAM...
 *  3) DST: FREE SRAM		SRC: PRIORITY N DRAM
 *
 *  perform remap
 *  done
 *
 * Arguments:
 * dst_page_map - an index array of IMM pages within page_list that comprise
 *	the pages just freed. The map is ordered with free L2CACHE pages first
 *	and free SRAM pages last.
 * sram_start - the index of the first SRAM page in dst_page_map.
 * dst_pages_found - the total number of SRAM+L2CACHE pages freed.
 * imm_info - imm_info struct of the calling client.
 *
 * Return Value: the number of pages successfully mapped in.
 *************************************************************************/

int
imm_free_pages_hotswap(struct imm_page_t **dst_page_map, s16 sram_start,
	s16 dst_pages_found, struct imm_info_t *imm_info)
{
	int i, total = 0;
	s16 src_pages_found = 0;
	struct imm_page_t **src_page_map;

	if (!dst_pages_found) return(0);

#ifdef CONFIG_IMM_DEBUG
	imm_debug("sram_start = %d\n", sram_start);
	print_page_map(30, dst_page_map, dst_pages_found);
#endif

	/* make a single call to kmalloc and arrange the memory manually */
	if ((src_page_map = kmalloc(dst_pages_found*sizeof(struct imm_page_t *), GFP_ATOMIC)) == NULL)
		return 0;

	/* if L2CACHE pages have been freed up */
	if (sram_start) {
		total = src_pages_found = find_used_pages_to_promote(src_page_map,
			sram_start, IMM_CLIENT_MIN_PRIORITY+1, imm_info, IMM_INFO_L2CACHE);

#ifdef CONFIG_IMM_DEBUG
		print_page_map(31, src_page_map, src_pages_found);
#endif

		/* carry out the remapping of each of the pages */
		imm_remap_pages(src_page_map, dst_page_map, src_pages_found);

		/* if there weren't enough pages to fill out the L2CACHE gaps */
		/* there won't be any for the SRAM gaps, so quit */
		if (src_pages_found < sram_start) {
			kfree(src_page_map);
			return total;
		}

		/* if any of the source pages we just hotswapped from was SRAM, */
		/* they are now free and we must process them as well. */
		for (i = src_pages_found-1; (sram_start >= 0)&&(i >= 0); i--) {
			if (PAGE_IS_SRAM(src_page_map[i]->flag)) {
				dst_page_map[--sram_start] = src_page_map[i];
			}
		}

		if (unlikely(i >= 0)) {
			imm_failure("Bad index");
			kfree(src_page_map);
			return total;
		}
	}

	/* if SRAM pages have been freed up */
	if (dst_pages_found - sram_start) {
#ifdef CONFIG_IMM_DEBUG
		print_page_map(33, &dst_page_map[sram_start],
			dst_pages_found-sram_start);
#endif
		/* Search for DRAM pages to fill the SRAM gap */
		src_pages_found = find_used_pages_to_promote(&src_page_map[sram_start],
			dst_pages_found-sram_start, IMM_CLIENT_MIN_PRIORITY+1, imm_info, IMM_INFO_SRAM);

#ifdef CONFIG_IMM_DEBUG
		print_page_map(34, &src_page_map[sram_start],
			src_pages_found);
#endif
		total += src_pages_found;
		/* carry out the remapping of each of the pages */
		imm_remap_pages(&src_page_map[sram_start],
			&dst_page_map[sram_start], src_pages_found);
	}

	return total;
}

int imm_priority_change(struct imm_info_t *imm_info, int priority)
{
	int upgrade, i = 0, total = 0;
	int pages_needed = imm_info->num_pages;
	s16 src_l2_pages_found = 0;
	s16 src_sram_pages_found = 0;
	s16 src_dram_pages_found = 0;
	s16 src_pages_found = 0;
	struct imm_page_t **src_page_map, *p;
	struct imm_page_t **dst_page_map;
	s16 dst_l2_pages_found = 0;
	s16 dst_sram_pages_found = 0;
	s16 dst_pages_found = 0;

	imm_debug("%d priority change from %d to %d\n",
		imm_info->immid, imm_info->priority, priority);

	upgrade = PRIORITY1_IS_GREATER(priority, imm_info->priority)?1:0;
	if (PRIORITY_EQUAL(priority, imm_info->priority)) return(0);
	imm_info->priority = priority;
	if (pages_needed < 1) return(0);

	imm_debug("%s %d pages\n",
		(upgrade)?"upgrading":"downgrading", pages_needed);

	/* make a single call to kmalloc and arrange the memory manually */
	if ((src_page_map = kmalloc(pages_needed*sizeof(struct imm_page_t *),
		GFP_ATOMIC)) == NULL) return 0;

#if (IMM_L2CACHE_PAGES > 0)
	for (p = L2CACHESTARTP; (p < L2CACHEENDP)&&(i < pages_needed); p++)
	{
		/* only grab mapped, non-DMA pages that belong to this client*/
		if (PAGE_IS_MAPPED(p->flag)&&
			(p->immid == imm_info->immid))
		{
			src_page_map[src_pages_found++] = p;
			src_l2_pages_found++; i++;
			if (!upgrade) p->flag |= IMM_INFO_REPRIORITIZE;
		}
	}
#endif
#if (IMM_SRAM_PAGES > 0)
	for (p = SRAMSTARTP; (p < SRAMENDP)&&(i < pages_needed); p++)
	{
		/* only grab mapped, non-DMA pages that belong to this client*/
		if (PAGE_IS_MAPPED(p->flag)&&
			(p->immid == imm_info->immid))
		{
			if (!PAGE_IS_DMA(p->flag)) {
				src_page_map[src_pages_found++] = p;
				src_sram_pages_found++;
				p->flag |= IMM_INFO_REPRIORITIZE;
			} else {
				p->priority = imm_info->priority;
			}
			i++;
		}
	}
#endif
#if (IMM_DRAM_PAGES > 0)
	for (p = DRAMSTARTP; (p < DRAMENDP)&&(i < pages_needed); p++)
	{
		/* only grab mapped, non-DMA pages that belong to this client*/
		if (PAGE_IS_MAPPED(p->flag)&&
			(p->immid == imm_info->immid))
		{
			src_page_map[src_pages_found++] = p;
			src_dram_pages_found++; i++;
			if (upgrade) p->flag |= IMM_INFO_REPRIORITIZE;
		}
	}
#endif
	if (unlikely(i < pages_needed)) {
		imm_failure("Bad page count");
		kfree(src_page_map);
		return(0);
	}

	imm_debug("page breakdown: %d L2CACHE pages, %d SRAM pages, %d DRAM pages\n",
		src_l2_pages_found, src_sram_pages_found, src_dram_pages_found);

	/* make a single call to kmalloc and arrange the memory manually */
	if ((dst_page_map = kmalloc(src_pages_found*sizeof(struct imm_page_t *),
		GFP_ATOMIC)) == NULL)
		return 0;

	if (upgrade) {
		/* on upgrade, we ignore the L2CACHE pages because they're */
		/* already most optimal, and focus on upgrading DRAM and SRAM */

#if (IMM_L2CACHE_PAGES > 0) && ((IMM_SRAM_PAGES > 0) || (IMM_DRAM_PAGES > 0))
		/* find mapped L2CACHE pages to swap out to make way for our SRAM/DRAM */
		if (src_sram_pages_found+src_dram_pages_found)
		{
			/* since our client's pages are still set to their original */
			/* priority, we may also find them in the search */
			dst_l2_pages_found = find_used_pages_to_demote(dst_page_map,
				src_l2_pages_found+src_sram_pages_found+src_dram_pages_found,
				imm_info->priority, IMM_INFO_L2CACHE);
			dst_l2_pages_found = remove_my_pages(dst_page_map,
				dst_l2_pages_found, imm_info->immid);
			dst_l2_pages_found = min(dst_l2_pages_found, src_sram_pages_found+src_dram_pages_found);
		}
#endif

#if (IMM_SRAM_PAGES > 0) && (IMM_DRAM_PAGES > 0)
		/* find mapped SRAM pages to swap out to make way for our remaining DRAM */
		if (dst_l2_pages_found < src_dram_pages_found)
		{
			/* since our client's pages are still set to their original */
			/* priority, we may also find them in the search */
			dst_sram_pages_found = find_used_pages_to_demote(&dst_page_map[dst_l2_pages_found],
				src_sram_pages_found+src_dram_pages_found, imm_info->priority,
				IMM_INFO_SRAM);
			dst_sram_pages_found = remove_my_pages(&dst_page_map[dst_l2_pages_found],
				dst_sram_pages_found, imm_info->immid);
			dst_sram_pages_found = min(dst_sram_pages_found, src_dram_pages_found);
		}
#endif
		update_priorities(src_page_map, src_pages_found,
				imm_info->priority);

#if (IMM_SRAM_PAGES > 0) && (IMM_DRAM_PAGES > 0)
		/* hotswap the mapped SRAM pages to slower memory */
		if (dst_sram_pages_found)
			hotswap_pages(&dst_page_map[dst_l2_pages_found],
				dst_sram_pages_found, IMM_INFO_DRAM);
#endif

#if (IMM_L2CACHE_PAGES > 0)
		/* hotswap the mapped L2CACHE pages to slower memory */
		if (dst_l2_pages_found)
			hotswap_pages(dst_page_map, dst_l2_pages_found,
#if (IMM_SRAM_PAGES > 0)
				IMM_INFO_SRAM);
#else
				IMM_INFO_DRAM);
#endif
#endif
		total = dst_sram_pages_found + dst_l2_pages_found;
	} else {
		/* on downgrade, we ignore the DRAM pages because they're */
		/* already least optimal, we focus on downgrading L2CACHE and SRAM */

		update_priorities(src_page_map, src_pages_found, imm_info->priority);

#if (IMM_L2CACHE_PAGES > 0) && ((IMM_SRAM_PAGES > 0) || (IMM_DRAM_PAGES > 0))
		/* find mapped SRAM/DRAM pages to swap into our L2CACHE */
		if (src_l2_pages_found)
		{
			dst_pages_found = find_used_pages_to_promote(dst_page_map,
				src_l2_pages_found, imm_info->priority, NULL, IMM_INFO_L2CACHE);

			/* carry out the remapping of each of the pages */
			imm_remap_pages(src_page_map, dst_page_map,
					dst_pages_found);

			total = dst_pages_found;

			/* if any of the source pages we just hotswapped from
			 * was SRAM, they are now ours and we must process
			 * them as well.
			 */
			for (i = dst_pages_found-1; i >= 0; i--) {
				if (PAGE_IS_SRAM(dst_page_map[i]->flag)) {
					src_page_map[--src_l2_pages_found]
						= dst_page_map[i];
					src_sram_pages_found++;
				}
			}
		}
#endif
#if (IMM_SRAM_PAGES > 0) && (IMM_DRAM_PAGES > 0)
		/* find mapped DRAM pages to swap into L2CACHE */
		if (src_sram_pages_found) {
			/* since our client's pages are still set to their original */
			/* priority, we may also find them in the search */
			dst_pages_found = find_used_pages_to_promote(dst_page_map,
				src_sram_pages_found, imm_info->priority, NULL, IMM_INFO_SRAM);

			/* carry out the remapping of each of the pages */
			imm_remap_pages(&src_page_map[src_l2_pages_found], dst_page_map, dst_pages_found);

			total += dst_pages_found;
		}
#endif
	}
	clear_priority_flag(src_page_map, src_pages_found);
	return total;
}
