/*
 *  linux/arch/arm/mach-pxa/imm/memattr.c
 *
 *  Intel Memory Management
 *
 *  Memory Attribute Change API
 *
 *  Todd Brandt
 *  Copyright (c) 2004, Intel Corporation (todd.e.brandt@intel.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.

 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

#include <linux/kernel.h>
#include <linux/cache.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/shm.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/linkage.h>
#include <linux/proc_fs.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/hardware.h>
#include <asm/cpu-single.h>
#include <asm/irq.h>
#include <asm/page.h>
#include <asm/pgalloc.h>
#include <asm/cacheflush.h>
#include "imm.h"

#define ATTR_L1(x)	((x)&IMM_ATTR_L1_CACHEABLE)
#define ATTR_LLR(x)	((x)&IMM_ATTR_LLR_CACHEABLE)
#define ATTR_L2(x)	((x)&IMM_ATTR_L2_CACHEABLE)
#define ATTR_WT(x)	((x)&IMM_ATTR_L1_WRITETHROUGH)

/* this call is needed to flush L2 and ensure it's unlocked */
extern int xsc3_l2_cache_unlock_range(u32, u32);

/* flag showing that this module was initialized */
static int module_initialized = 0;

/* a vessel for IMM flags, TEX, C, and B bit values */
struct memattr_t {
	u32 flags;
	u8 tex;
	u8 cb;
};

/*----------------------------------------------------------------------
 *
 *  High Level Call Tree:
 *
 *  [Application or Driver Call]
 * 	imm_get_mem_attr
 * 		imm_get_ptep
 *		get_attr_from_pte
 * 	imm_set_mem_attr
 *		get_texcb
 *		imm_get_ptep
 *
 *  [Module Initialization - from Policy Manager]
 *	imm_dau_memattr_init
 *
 *--------------------------------------------------------------------*/

/*************************************************************************
 * Function: get_texcb
 * Description: Given a set of IMM Memory attribute flags, figure out what
 *	TEX, C, and B bit values are needed in the Page Table Entries.
 *
 * Arguments:
 * m->flags: The IMM memory attribute flags
 *
 * Return Values:
 * m->tex: The TEX value for the PTE
 * m->cb: The C and B bit values for the PTE
 *************************************************************************/

void get_texcb(struct memattr_t *m) {
	u32 flags = m->flags;

#ifdef CONFIG_PXA3xx
	if(flags & IMM_ATTR_LLR_CACHEABLE) {
		/* TEX = 001, C = 1, B = 0 */
		/* LLR */
		m->tex = 1;
		m->cb = 2;
	} else if(ATTR_L1(flags)&&ATTR_L2(flags)&&!ATTR_WT(flags)) {
		/* TEX = 101, C = 1, B = 1 */
		/* Inner write-back, read-allocate */
		/* Outer write-back, write-allocate */
		m->tex = 5;
		m->cb = 3;
	} else if(ATTR_L1(flags)&&ATTR_L2(flags)&&ATTR_WT(flags)) {
		/* TEX = 101, C = 1, B = 0 */
		/* Inner write-through, read-allocate */
		/* Outer write-back, write-allocate */
		m->tex = 5;
		m->cb = 2;
	} else if(ATTR_L1(flags)&&!ATTR_L2(flags)&&!ATTR_WT(flags)) {
		/* TEX = 100, C = 0, B = 1 */
		/* Inner write-back, read-allocate */
		/* Outer uncacheable */
		m->tex = 4;
		m->cb = 1;
	} else if(ATTR_L1(flags)&&!ATTR_L2(flags)&&ATTR_WT(flags)) {
		/* TEX = 000, C = 1, B = 0 */
		/* Inner write-through, read-allocate */
		/* Outer uncacheable */
		m->tex = 0;
		m->cb = 2;
	} else if(!ATTR_L1(flags)&&ATTR_L2(flags)) {
		/* TEX = 101, C = 0, B = 0 */
		/* Inner uncacheable */
		/* Outer write-back, write-allocate */
		m->tex = 5;
		m->cb = 0;
	} else {
		/* TEX = 000, C = 0, B = 1 */
		/* Inner uncacheable */
		/* Outer uncacheable */
		m->tex = 0;
		m->cb = 1;
	}
#else
	if(flags & IMM_ATTR_LLR_CACHEABLE) {
		/* X = 1, C = 1, B = 0 */
		/* LLR */
		m->tex = 1;
		m->cb = 2;
	} else if(ATTR_L1(flags)&&!ATTR_WT(flags)) {
		/* X = 0, C = 1, B = 1 */
		/* Inner write-back, read-allocate */
		/* Outer write-back, write-allocate */
		m->tex = 0;
		m->cb = 3;
	} else if(ATTR_L1(flags)&&ATTR_WT(flags)) {
		/* X = 0, C = 1, B = 0 */
		/* Inner write-through, read-allocate */
		/* Outer write-back, write-allocate */
		m->tex = 0;
		m->cb = 2;
	} else {
		/* X = 0, C = 0, B = 0 */
		/* Inner uncacheable */
		/* Outer uncacheable */
		m->tex = 0;
		m->cb = 0;
	}
#endif
}

/*************************************************************************
 * Function: get_attr_from_pte
 * Description: The reverse of get_texcb. Given a hardware PTE value,
 *	figure out what IMM memory attribute flags describe it.
 *
 * Arguments:
 * pte - a hardware PTE value
 *
 * Return Value:
 * IMM flags describing the PTE cache attributes
 * -1 on failure (failure only if page type not supported)
 *************************************************************************/

int
get_attr_from_pte(u32 pte) {
	struct memattr_t m = {0, 0, 0};
	int flags = -1;

	if((pte & 0x3) == 0x1) { /* 64K page */
		flags = IMM_ATTR_64K_PAGE;
#ifdef CONFIG_PXA3xx
		m.tex = (pte >> 12) & 0x7;
#else
		m.tex = (pte >> 12) & 0x1;
#endif
		m.cb = (pte >> 2) & 0x3;
	} else if((pte & 0x3) == 0x3) { /* 4K page */
		flags = 0;
#ifdef CONFIG_PXA3xx
		m.tex = (pte >> 6) & 0x7;
#else
		m.tex = (pte >> 6) & 0x1;
#endif
		m.cb = (pte >> 2) & 0x3;
	} else {
		return -1;
	}

#ifdef CONFIG_PXA3xx
	if((m.tex == 1) && (m.cb == 2)) {
		/* TEX = 001, C = 1, B = 0 */
		/* LLR */
		flags |= IMM_ATTR_LLR_CACHEABLE;
	} else if((((m.tex & 5) == 5) && ((m.cb & 1) == 1)) ||
		(((m.tex & 6) == 0) && (m.cb == 3))) {
		/* TEX = 1X1, C = X, B = 1, TEX = 00X, C = 1, B = 1 */
		/* Inner write-back, read-allocate */
		/* Outer write-back, write-allocate */
		flags |= IMM_ATTR_L1_CACHEABLE | IMM_ATTR_L2_CACHEABLE;
	} else if(((m.tex & 5) == 5) && (m.cb == 2)) {
		/* TEX = 1X1, C = 1, B = 0 */
		/* Inner write-through, read-allocate */
		/* Outer write-back, write-allocate */
		flags |= IMM_ATTR_L1_CACHEABLE | IMM_ATTR_L2_CACHEABLE | IMM_ATTR_L1_WRITETHROUGH;
	} else if(((m.tex & 5) == 4) && ((m.cb & 1) == 1)) {
		/* TEX = 1X0, C = X, B = 1 */
		/* Inner write-back, read-allocate */
		/* Outer uncacheable */
		flags |= IMM_ATTR_L1_CACHEABLE;
	} else if(((m.tex & 1) == 0) && (m.cb == 2)) {
		/* TEX = XX0, C = 1, B = 0 */
		/* Inner write-through, read-allocate */
		/* Outer uncacheable */
		flags |= IMM_ATTR_L1_CACHEABLE | IMM_ATTR_L1_WRITETHROUGH;
	} else if(((m.tex & 5) == 5) && (m.cb == 0)) {
		/* TEX = 1X1, C = 0, B = 0 */
		/* Inner uncacheable */
		/* Outer write-back, write-allocate */
		flags |= IMM_ATTR_L2_CACHEABLE;
	} else {
		/* Inner uncacheable */
		/* Outer uncacheable */
		flags = 0;
	}
#else
	if((m.tex == 1) && (m.cb == 2)) {
		/* X = 1, C = 1, B = 0 */
		/* LLR */
		flags |= IMM_ATTR_LLR_CACHEABLE;
	} else if((m.tex == 0) && (m.cb == 3)) {
		/* X = 0, C = 1, B = 1 */
		/* Inner write-back, read-allocate */
		flags |= IMM_ATTR_L1_CACHEABLE;
	} else if((m.tex == 0) && (m.cb == 2)) {
		/* X = 0, C = 1, B = 0 */
		/* Inner write-through, read-allocate */
		flags |= IMM_ATTR_L1_CACHEABLE | IMM_ATTR_L1_WRITETHROUGH;
	} else {
		/* X = 0, C = 0, B = 0 */
		/* Inner uncacheable */
		/* Outer uncacheable */
		flags = 0;
	}
#endif
	return flags;
}

/*************************************************************************
 * Function: imm_get_ptep
 * Description: Retreive a pointer to the hardware PTE that describes a
 *	given virtual address.
 *
 * Arguments:
 * v     - a virtual address
 * immid - the immid of the client this address belongs to
 *
 * Return Value:
 * (u32 *) pointer to the hardware PTE, NULL on failure
 *************************************************************************/

u32 *
imm_get_ptep(void *v, u32 immid)
{
	pmd_t *pmd;
	pte_t *pte;
	pgd_t *pgd;
	u32 *val = NULL, virtual = (u32)v;
	struct mm_struct* mm;

	if(IMMID_USER(immid))
		mm = current->mm;
	else
		mm = &init_mm;

	down_read(&mm->mmap_sem);

	/* get the 1st level page table address */
	pgd = pgd_offset(mm, virtual);
	if (!pgd_none(*pgd) && !pgd_bad(*pgd)) {
		/* get the 2nd level page table address */
                pmd = pmd_offset(pgd, virtual);
		if (!pmd_none(*pmd) && !pmd_bad(*pmd)) {
			/* get the software PTE address */
                        pte = pte_offset_kernel(pmd, virtual);
			/* in linux 2.6, HW PTE is 2K below SW */
			if(pte)
	                        val = (u32 *)((u32)pte-2048);
                }
        }

	up_read(&mm->mmap_sem);
        return val;
}

/*************************************************************************
 *
 * Function: imm_get_mem_attr
 * Description: Retrieve the page attributes for a given virtual
 * address
 *
 * Arguments:
 * virtual - virtual address
 * immid - the id used to tell if this is kernel or user space
 *
 * Return Value:
 * (int) the attributes, -1 if unmapped
 *
 *************************************************************************/

int
imm_get_mem_attr(void *v, u32 immid)
{
	int attr = -1;
	u32 *ptep;

	down(&imm_sem);
	/* grab the PTE pointer */
	ptep = imm_get_ptep(v, immid);
	/* if it's valid, determine the flags */
	if(ptep) attr = get_attr_from_pte(*ptep);
	up(&imm_sem);
	return attr;
}

/*************************************************************************
 *
 * Function: imm_set_mem_attr
 * Description: Changes the TEX, C, B value for a range of memory
 *
 * Arguments:
 * virtual_address - the virtual address of the block to be changed
 * size - the size in bytes of the target block
 * flags - cacheability flags
 * immid - the id to assign this to
 *
 * Return Value:
 * the error condition, can be sent to imm_error to retrieve the
 * error string
 *************************************************************************/

int
imm_set_mem_attr(void *addr, size_t size, u32 flags, u32 immid)
{
	struct memattr_t page_instr;
	u32 address, *ptep_hw, pte_hw, temp;
	u32 isubpage;
	u32 content, mask, vm_found, start, end;
	struct imm_info_t *imm_info;
	struct	mm_struct* mm;
	struct vm_area_struct *vm_search;

	down(&imm_sem);
	imm_debug("imm_set_mem_attr(%08X, %d, %x, %d)\n",
		(u32)addr, size, flags, immid);

	/* if this client isn't registered, forget it */
	if((imm_info = imm_info_getentry(immid)) == NULL) {
		up(&imm_sem);
		return -1;
	}

	/* if this module isn't initialized (very rare), forget it */
	if(!module_initialized) {
		register_error(immid, IMM_ERROR_UNINIT);
		up(&imm_sem);
		return -1;
	}

	/* if the address is obviously out of bounds, forget it */
	if((addr == NULL) || (IMMID_USER(immid)&&((u32)addr >= PAGE_OFFSET))) {
		register_error(immid, IMM_ERROR_PERMISSION_DENIED);
		up(&imm_sem);
		return -1;
	}

        /* if the size is invalid, fail */
	if(size < 1) {
		register_error(immid, IMM_ERROR_SIZETOOLOW);
		up(&imm_sem);
		return -1;
	}

	/* determine the TEX, C, and B bit values for the target attributes */
	page_instr.flags = flags;
	get_texcb(&page_instr);

	imm_debug("TEX = %x, CB = %x\n",
		page_instr.tex, page_instr.cb);

	/* if this memory is user, we have to check permissions, if it's */
	/* kernel and we haven't failed yet, we assume we have permission */
	if(IMMID_USER(immid)) {
		mm = current->mm;

		/* perform a test to see if this address range is permissible */
		down_read(&mm->mmap_sem);

		start = (u32)addr;
		end = (u32)addr+size;
		vm_search = mm->mmap;
		vm_found = 0;

		do {
			if ( start >= vm_search->vm_start && start < vm_search->vm_end) {
				if ( end <= vm_search->vm_end) {
					vm_found = 1;
					break;
				} else {
					start = vm_search->vm_end;
				}
			}
			vm_search = vm_search->vm_next;

		} while ( ( start >= vm_search->vm_start) && ( vm_search ) );

		up_read(&mm->mmap_sem);

		/* if the address is not in this client's space, forget it */
		if ( !vm_found ) {
			register_error(immid, IMM_ERROR_PERMISSION_DENIED);
			up(&imm_sem);
			return(-1);
		}
	} else {
		mm = &init_mm;
	}

	imm_debug("found the memory, proceeding with attribute change\n");
	start = (u32)addr;
	end = (u32)addr+size;

	/* flush L1 and L2 cache prior to messing with the attributes */
	/* This will ensure full coherency in case we lose or gain a */
	/* cache level */
	flush_cache_all();
	xsc3_l2_cache_unlock_range(start, end); /* flush and unlock L2 */

	/* loop through each page of the target range separately */
	for(address = start; address < end; address += PAGE_SIZE) {
		/* perform a read+write operation to trigger an exception */
		/* if this page happens to be copy-on-write (i.e. unmapped */
		/* until written) */
		temp = *(volatile u32 *)address;
		*(volatile u32 *)address = temp;

		/* grab the hardware pte pointer and its value */
		ptep_hw = imm_get_ptep((u32 *)address, immid);
		pte_hw = *ptep_hw;

		/* if this PTE is inaccessible, fail the rest of the range */
		/* this shouldn't happen but is here just in case */
		if(!ptep_hw || !pte_hw) {
			register_error(immid, IMM_ERROR_PERMISSION_DENIED);
			up(&imm_sem);
			return -1;
		}

		imm_debug("BEFORE: address[%08X], pte_hw[%08X]\n", (u32)(address&PAGE_MASK), pte_hw);

		/* we support only two page entry types */
		if ((pte_hw & 0x3) == 0x01) {  // Large page in Coarse page table
			/* Ensure L2 coherency on the full page boundary */
			xsc3_l2_cache_unlock_range(address&PAGE_MASK, PAGE_ALIGN(address+15*PAGE_SIZE)); /* flush and unlock L2 */
#ifdef CONFIG_PXA3xx
			mask = 0xFFFF8FF3;
			content = (((page_instr.tex)&7) << 12) | (((page_instr.cb)&3) << 2);
#else
			mask = 0xFFFFEFF3;
			content = (((page_instr.tex)&7) << 12) | (((page_instr.cb)&3) << 2);
#endif
			/* Large pages are actually just 16 contiguous Extended */
			/* Small entries. Make sure all 16 are identical */
			for ( isubpage = 0; isubpage < 16; isubpage++ ) {
				*ptep_hw = ((*ptep_hw) & mask) | (content & ~mask);

				/* make sure the new PTE value is flushed from L1/L2 */
				xsc3_flush_user_cache_range((u32)ptep_hw, (u32)ptep_hw+4, 0);
				xsc3_l2_cache_unlock_range((u32)ptep_hw, (u32)ptep_hw+4);

				/* Now flush the TLBs so the change will take effect */
				flush_tlb_kernel_page((address&PAGE_MASK) + (isubpage*PAGE_SIZE));

				imm_debug("AFTER: address[%08X], pte_hw[%08X]\n",
					address, *ptep_hw);
				ptep_hw++;
			}

			address += PAGE_SIZE*15;

		} else if ((pte_hw & 0x3) == 0x03) {
			/* Ensure L2 coherency on the full page boundary */
			xsc3_l2_cache_unlock_range(address&PAGE_MASK, PAGE_ALIGN(address)); /* flush and unlock L2 */
#ifdef CONFIG_PXA3xx
			mask = 0xFFFFFE33;
			content = (((page_instr.tex)&7) << 6) | (((page_instr.cb)&3) << 2);
#else
			mask = 0xFFFFFFB3;
			content = (((page_instr.tex)&7) << 6) | (((page_instr.cb)&3) << 2);
#endif
			pte_hw = (pte_hw & mask) | (content & ~mask);
			*ptep_hw = pte_hw;
			imm_debug("AFTER: address[%08X], pte_hw[%08X]\n",
				address, *ptep_hw);

			/* make sure the new PTE value is flushed from L1/L2 */
			xsc3_flush_user_cache_range((u32)ptep_hw, (u32)ptep_hw+4, 0);
			xsc3_l2_cache_unlock_range((u32)ptep_hw, (u32)ptep_hw+4);

			/* Now flush the TLBs so the change will take effect */
			flush_tlb_kernel_page(address&PAGE_MASK);
		}
	}

	/* after the attribute change, flush the cache again to ensure */
	/* coherency for any cache levels just added */
	flush_cache_all();

	/* no errors happened, we made it */
	register_error(immid, IMM_ERROR_NONE);
	up(&imm_sem);
	return(0);
}

void
imm_dau_memattr_init(struct proc_dir_entry *imm_dir)
{

        printk(KERN_INFO "Intel(c) Memory Management - Page Attribute Manipulation\n");

	module_initialized = 1;
}

EXPORT_SYMBOL(imm_get_mem_attr);
EXPORT_SYMBOL(imm_set_mem_attr);
