/*
 *  linux/arch/arm/mach-pxa/imm/policy.c
 *
 *  Intel Memory Management
 *
 *  Policy Manager - focal point for IMM functionality
 *
 *  Todd Brandt
 *  Copyright (c) 2004, Intel Corporation (todd.e.brandt@intel.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.

 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

#include <linux/mm.h>
#include <linux/rmap.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/hardware.h>
#include <linux/module.h>
#include <linux/notifier.h>
#include <linux/profile.h>
#include <asm/uaccess.h>
#include "imm.h"

/* the imm semaphore */
DECLARE_MUTEX(imm_sem);

/* error list */
const char *imm_errors[] = {
        "Operation completed successfully",
        "System memory is too low, kernel malloc failed",
        "Insufficient space left",
        "The size argument is too small, size must be > 0 bytes",
        "The requested entry does not exist",
        "The address range requested is off limits to this process",
        "This functionality is currently unsupported",
	"The requested address range overlaps with an existing entry's \
		address range",
	"The requested IMM module failed to initialize due to a kernel error",
        "Unknown Error!"
};

/* immid list */
#ifdef CONFIG_IMM_SIMPLE_PRIORITY_CONTROL
struct imm_priority_t priority_list;
#define NAME_MATCH(x, y)					\
       	((min(strlen(y), (size_t)(IMM_CLIENT_NAME_SIZE-1))	\
			== strlen((const char*)(x))) && 	\
	 	(!strncmp((const char*)(x), (const char*)(y), strlen(x))))
#define VALID_NAME(x)	((x) != NULL)
#endif
struct imm_info_t immid_list;
struct proc_dir_entry *proc_imm = NULL;
static u32 kernel_immid = IMM_FIRST_KERNEL_IMMID;
static int imm_initialized = 0;

/*----------------------------------------------------------------------
 *
 *  High Level Call Tree:
 *
 *  [Application or Driver Call]
 *	imm_register_kernel
 *		imm_info_newentry
 *	imm_register_user
 *		imm_info_newentry
 *	imm_get_physical
 *	imm_get_virtual
 *	imm_error
 *
 *  [IMM Module call (SRAM, Cache, Memattr, Power)]
 *	imm_info_getentry
 *	register_error
 *	pprintf_helper
 *
 *  [Process Exit - Kernel Hook from System Profile Interface]
 * 	imm_task_exit
 *		imm_free_immid (sram.c)
 *		imm_cache_unlock_immid (cache.c)
 *		imm_info_delentry
 *
 *  [Memory Free - Kernel Hook from System Profile Interface]
 * 	imm_do_munmap
 *		imm_info_getentry
 *		imm_cache_handle_unmap (cache.c)
 *
 *  [Module Initialization]
 *	imm_init
 *		imm_dau_sram_init (sram.c)
 *		imm_dau_cache_init (cache.c)
 *		imm_dau_memattr_init (memattr.c)
 *		imm_autopower_init (power.c)
 *
 *--------------------------------------------------------------------*/

/*************************************************************************
 *
 * Function: imm_get_physical
 * Description: Retrieve the physical address for a given imm virtual
 * address
 *
 * Arguments:
 * virtual - address returned from imm_malloc
 * immid - the id of the caller
 *
 * Return Value:
 * the physical address, 0 on failure
 *************************************************************************/

u32
imm_get_physical(void *v, u32 immid)
{
	pmd_t *pmd;
	pte_t *pte;
	pgd_t *pgd;
	u32 val = 0, virtual = (u32)v;
	struct mm_struct* mm;

	if (IMMID_USER(immid))
		mm = current->mm;
	else
		mm = &init_mm;

	pgd = pgd_offset(mm, virtual);
	if (!pgd_none(*pgd) && !pgd_bad(*pgd)) {
		/* 1st level entry pointer */
		pmd = pmd_offset(pgd, virtual);
		if (!pmd_none(*pmd) && !pmd_bad(*pmd)) {
			/* 2nd level entry pointer */
			pte = pte_offset_kernel(pmd, virtual);
			if (pte) {
				val = (*(u32 *)((u32)pte-2048))&PAGE_MASK;
				val += virtual%PAGE_SIZE;
			}
		} else if (!pmd_none(*pmd)) {
			/* work around */
			val = (unsigned long)virt_to_phys((void*)virtual);
			/*
			 * old method address conversion is not correct
			val = (*(u32 *)pmd) & 0xFFF00000;
			val += virtual%0x100000;
			*/
		}
	}

	return val;
}

/*************************************************************************
 *
 * Function: imm_get_virtual
 * Description: Retrieve the virtual address for a given physical
 * address
 *
 * Arguments:
 * physical - address of some memory in the immid's address space
 * immid - the id of the caller
 *
 * Return Value:
 * the virtual address, 0 on failure
 *************************************************************************/

u32
imm_get_virtual(u32 p, struct task_struct *tsk)
{
	pmd_t *pmd;
	pte_t *pte;
	pgd_t *pgd;
	u32 val, v;
	struct mm_struct* mm;
	struct vm_area_struct *vm_search;

	down(&imm_sem);
	if ((tsk == NULL)||(tsk->mm == NULL)) return 0;
	mm = tsk->mm;

	for (vm_search = mm->mmap; vm_search; vm_search = vm_search->vm_next) {
		for (v = vm_search->vm_start; v < vm_search->vm_end;
				v+=PAGE_SIZE) {
			pgd = pgd_offset(mm, v);
			if (!pgd_none(*pgd) && !pgd_bad(*pgd)) {
				pmd = pmd_offset(pgd, v);
				if (!pmd_none(*pmd) && !pmd_bad(*pmd)) {
					pte = pte_offset_kernel(pmd, v);
					if (pte) {
						val = (*(u32 *)((u32)pte-2048))&PAGE_MASK;
						val += p%PAGE_SIZE;
						if (p == val) {
							up(&imm_sem);
							return(v+(p%PAGE_SIZE));
						}
					}
				}
			}
		}
	}
	up(&imm_sem);
	return 0;
}

/*************************************************************************
 * imm_info list handlers
 ************************************************************************/

/* retrieve an imm_info object from the list */
struct imm_info_t *
imm_info_getentry(u32 immid)
{
	struct imm_info_t *head, *p;

	if (immid < 1)
		return NULL;
	head = &immid_list;
	p = head->next;

	/* loop through the list til you hit the end or the immid */
	while ((p != head) && (p->immid != immid)) p = p->next;
	if (p->immid == immid) {
		return p;
	} else {
		return NULL;
	}
}

/* create a new imm_info entry for a new immid */
static struct imm_info_t *
imm_info_newentry(u32 immid, char *name, u32 user_imm_space)
{
	struct imm_info_t *head, *p, *new;
#ifdef CONFIG_IMM_SIMPLE_PRIORITY_CONTROL
	struct imm_priority_t *ptr;
#endif

	head = &immid_list;
	p = head->next;

	/* loop through the list til you hit the end or the immid */
	while ((p != head) && (p->immid != immid)) p = p->next;
	if (p->immid == immid) {
		return p;
	} else {
		imm_debug("imm_info_newentry called, initializing a new \
				imm_info entry\n");
	        new = kmalloc(sizeof(struct imm_info_t), GFP_ATOMIC);
	        if (new == NULL) {
			imm_debug("imm_info_newentry failed, kmalloc \
					returned NULL: %s, line %d\n",
					__FILE__, __LINE__);
			return NULL;
		}

		new->immid = immid;
		new->last_error = 0;
		memcpy(new->name, name, min(strlen(name)+1,
			(size_t)(IMM_CLIENT_NAME_SIZE-1)));
		new->name[IMM_CLIENT_NAME_SIZE-1] = '\0';
		if (IMMID_USER(immid)) {
			new->mm = current->mm;
		} else {
			new->mm = &init_mm;
		}

#ifdef CONFIG_IMM_ALLOCATION_PRIORITY_SCHEME
		if (IMMID_USER(immid)) {
			new->priority = IMM_DEFAULT_USER_PRIORITY;
		} else {
			new->priority = IMM_DEFAULT_KERNEL_PRIORITY;
		}
		new->num_pages = 0;
#ifdef CONFIG_IMM_SIMPLE_PRIORITY_CONTROL
		for (ptr = priority_list.next;
			((ptr != &priority_list) &&
				!NAME_MATCH(ptr->name, name));
			ptr = ptr->next);
		if (ptr != &priority_list) {
			imm_debug("process %s is reassigned priority %d\n",
				name, ptr->priority);
			new->priority = ptr->priority;
		}
#endif
#endif
#ifdef CONFIG_IMM_API_SRAM
		if (IMMID_USER(immid)) {
			new->malloc_list = kmalloc(sizeof(struct imm_virt_t),
						GFP_ATOMIC);
			if (new->malloc_list == NULL) {
				imm_debug("imm_info_newentry failed, kmalloc \
						returned NULL: %s, line %d\n",
						__FILE__, __LINE__);
				kfree(new);
				return NULL;
			}
			new->malloc_list->immid = immid;
			new->malloc_list->start = user_imm_space
				+ IMM_MALLOC_MAP_SIZE;
			new->malloc_list->end = user_imm_space;
			new->malloc_list->prev = new->malloc_list;
			new->malloc_list->next = new->malloc_list;
		} else {
			new->malloc_list = &malloc_list;
		}
		new->used_space = 0;
#endif

#ifdef CONFIG_IMM_API_CACHE
		new->l2lock_list.addr = NULL;
		new->l2lock_list.len = 0;
		new->l2lock_list.prev = &new->l2lock_list;
		new->l2lock_list.next = &new->l2lock_list;
#endif

	        new->next = p;
	        new->prev = p->prev;
	        new->next->prev = new;
	        new->prev->next = new;

		return new;
	}
}

#if defined(CONFIG_IMM_API_SRAM) || defined(CONFIG_IMM_API_CACHE)
/* remove an immid from the list */
static int
imm_info_delentry(u32 immid)
{
	struct imm_info_t *head, *p;

	head = &immid_list;
	p = head->next;

	/* loop through the list til you hit the end or the immid */
	while ((p != head) && (p->immid != immid)) p = p->next;
	if (p->immid == immid) {
#ifdef CONFIG_IMM_API_SRAM
		if (IMMID_USER(immid)) {
			struct imm_virt_t *headv = p->malloc_list;
			struct imm_virt_t *pv = headv->next;
			while (pv != headv) {
				/* pv now points to the memory block
				 * to be removed
				 */
				pv->prev->next = pv->next;
				pv->next->prev = pv->prev;
				kfree(pv);
				pv = headv->next;
			}
			kfree(p->malloc_list);
		}
#endif
		p->prev->next = p->next;
		p->next->prev = p->prev;
		kfree(p);
		return 0;
	}
	return -1;
}

static int
imm_task_exit(struct notifier_block * self, unsigned long val, void * data) {
struct task_struct * tsk = (struct task_struct *)data;

	down(&imm_sem);
        if (tsk && imm_initialized) {
	/* cache must be cleared first since some of it may be */
	/* locked over SRAM */
#ifdef CONFIG_IMM_API_CACHE
		imm_cache_unlock_immid(tsk->pid);
#endif
#ifdef CONFIG_IMM_API_SRAM
		imm_free_immid(tsk->pid);
#endif
		imm_info_delentry(tsk->pid);
        }
	up(&imm_sem);

	return NOTIFY_OK;
}

static struct notifier_block imm_task_exit_nb = {
	.notifier_call  = imm_task_exit,
};
#endif

#ifdef CONFIG_IMM_API_CACHE
static int
imm_do_munmap(struct notifier_block * self, unsigned long val, void * data)
{
	unsigned long addr = (unsigned long)data;
	struct imm_info_t *imm_info;

	down(&imm_sem);

	if ((imm_info = imm_info_getentry(current->pid)) != NULL) {
		imm_cache_handle_unmap(addr, val);
	}
	up(&imm_sem);

	return 0;
}

static struct notifier_block imm_do_munmap_nb = {
	.notifier_call  = imm_do_munmap,
};
#endif

/* associate an error with an immid */
void
register_error(u32 immid, u32 err)
{
	struct imm_info_t *head, *p;

	head = &immid_list;
	p = head->next;

	/* loop through the list til you hit the end or the immid */
	while ((p != head) && (p->immid != immid)) p = p->next;
	if (p->immid == immid)
		p->last_error = err;
}

/*************************************************************************
 *
 * Function: imm_register(kernel and user)
 * Description: Create a new immid instance for either the kernel or user
 *   space.
 *
 * Return Value:
 * immid - a new unique id which will never conflict with user immids
 *
 *************************************************************************/

u32
imm_register_kernel(char *name)
{
	struct imm_info_t *imm_info;

	down(&imm_sem);
	if (name == NULL) name = IMM_DEFAULT_DRIVER_NAME;
	imm_debug("imm_register_kernel called for %s, new ID = %08X\n",
		name, kernel_immid);
	if ((imm_info = imm_info_newentry(kernel_immid, name, 0)) == NULL) {
		up(&imm_sem);
		return 0;
	}
	up(&imm_sem);
	return(kernel_immid++);
}

u32
imm_register_user(u32 vstart, u32 pid)
{
	struct imm_info_t *imm_info;

	down(&imm_sem);
	imm_debug("imm_register_user called for %s, vstart = %08X, pid = %d\n",
		current->comm, vstart, pid);
	if (!vstart || !pid ||
		((imm_info = imm_info_newentry(pid, current->comm, vstart))
		 == NULL)) {
		up(&imm_sem);
		return 0;
	}
	up(&imm_sem);
	return(pid);
}

#ifdef CONFIG_IMM_ALLOCATION_PRIORITY_SCHEME
int
imm_client_control(struct imm_client_info *info, u32 cmd)
{
	struct imm_info_t *head, *p;
	int i, res = 0;

	down(&imm_sem);
	head = &immid_list;
	p = head->next;
	switch(cmd) {
	case IMM_GET_NUM_CLIENTS:
		for (i = 0; p != head; i++, p = p->next);
		info->clientnum = i;
	case IMM_GET_INFO_BY_INDEX:
		for (i = 0; (p != head) && (i == info->clientnum);
			i++, p = p->next);
		if (p != head) {
			memcpy(info->name, p->name, IMM_CLIENT_NAME_SIZE);
			info->immid = p->immid;
			info->priority = p->priority;
		} else {
			res = -1;
		}
		break;
	case IMM_GET_INFO_BY_IMMID:
		for (i = 0; (p != head) && (p->immid != info->immid);
			i++, p = p->next);
		if (p != head) {
			memcpy(info->name, p->name, IMM_CLIENT_NAME_SIZE);
			info->clientnum = i;
			info->priority = p->priority;
		} else {
			res = -1;
		}
		break;
	case IMM_GET_INFO_BY_NAME:
		for (i = 0; (p != head) && (!NAME_MATCH(p->name, info->name));
			i++, p = p->next);
		if (p != head) {
			info->clientnum = i;
			info->immid = p->immid;
			info->priority = p->priority;
		} else {
			res = -1;
		}
		break;
	case IMM_SET_PRIORITY_BY_IMMID:
		if (VALID_PRIORITY(info->priority)) {
			/* loop through the list til you hit the end or
			 * the immid
			 */
			while ((p != head) && (p->immid != info->immid))
				p = p->next;
			if ((p != head) && (p->priority != info->priority))
				imm_priority_change(p, info->priority);
		} else {
			res = -1;
		}
		break;
	case IMM_SET_PRIORITY_BY_NAME:
		if (VALID_PRIORITY(info->priority) && VALID_NAME(info->name)) {
			/* loop through the list til you hit the end or
			 * the immid
			 */
			while ((p != head) &&
				(!NAME_MATCH(p->name, info->name)))
				p = p->next;
			if ((p != head) && (p->priority != info->priority))
				imm_priority_change(p, info->priority);
		} else {
			res = -1;
		}
		break;
	default:
		res = -1;
	}
	up(&imm_sem);
	return res;
}

#ifdef CONFIG_IMM_SIMPLE_PRIORITY_CONTROL
int
imm_add_priority_list_entry(char *name, s8 priority)
{
	struct imm_priority_t *head, *p, *new;
	struct imm_client_info clientinfo;

	if (!VALID_PRIORITY(priority))
		return -1;

	down(&imm_sem);
	head = &priority_list;
	p = head->next;

	/* loop through the list til you hit the end or the immid */
	while ((p != head) && (!NAME_MATCH(p->name, name)))
		p = p->next;
	if (p != head) {
		p->priority = priority;
		memcpy(clientinfo.name, p->name, IMM_CLIENT_NAME_SIZE);
	} else {
	        if ((new = kmalloc(sizeof(struct imm_priority_t), GFP_ATOMIC))
			       	== NULL) {
			up(&imm_sem);
	        	return -1;
		}
		new->priority = priority;
		memcpy(new->name, name, min(strlen(name)+1,
			(size_t)IMM_CLIENT_NAME_SIZE-1));
		new->name[IMM_CLIENT_NAME_SIZE-1] = '\0';

	        new->next = p;
	        new->prev = p->prev;
	        new->next->prev = new;
	        new->prev->next = new;

		memcpy(clientinfo.name, new->name, IMM_CLIENT_NAME_SIZE);
	}
	clientinfo.priority = priority;
	up(&imm_sem);
	imm_client_control(&clientinfo, IMM_SET_PRIORITY_BY_NAME);
	return 0;
}

int
imm_remove_priority_list_entry(char *name)
{
	struct imm_priority_t *head, *p;

	down(&imm_sem);
	head = &priority_list;
	p = head->next;

	while ((p != head) && (!NAME_MATCH(p->name, name))) p = p->next;
	if (p != head) {
		p->prev->next = p->next;
		p->next->prev = p->prev;
		kfree(p);
		up(&imm_sem);
		return 0;
	}
	up(&imm_sem);
	return -1;
}
#endif
#endif

/*************************************************************************
 *
 * Function: imm_error
 * Description: Return the last error for the given immid
 *
 * Return Value:
 * a const char pointer of the error string
 *
 *************************************************************************/

const char *
imm_error(u32 immid)
{
	struct imm_info_t *head, *p;

	down(&imm_sem);
	head = &immid_list;
	p = head->next;

	/* loop through the list til you hit the end or the immid */
	while ((p != head) && (p->immid != immid)) p = p->next;
	if (p->immid == immid) {
		if ((p->last_error >= 0)
			&& (p->last_error < IMM_ERROR_UNKNOWN)) {
			up(&imm_sem);
			return imm_errors[p->last_error];
		}
	}
	up(&imm_sem);
	return "The caller is not a client of IMM";
}

int
pprintf_helper(struct imm_proc_t *proc_head, char *page, char **start,
		int count, int *eof, int cmd, const char *fmt, ...)
{
	char buffer[1000];
	struct imm_proc_t *new;
	va_list args;
	int len, temp;

	switch(cmd) {
	case PCMD_CLEAN: /* init mode */
		for (new = proc_head->next; new != proc_head;
			new = proc_head->next) {
			new->prev->next = new->next;
			new->next->prev = new->prev;
			kfree(new);
		}
		break;
	case PCMD_STORE: /* input mode */
		va_start(args, fmt);
		len = vsprintf(buffer, fmt, args);
	        va_end(args);

		if ((new = kmalloc(sizeof(struct imm_proc_t), GFP_ATOMIC))
				!= NULL) {
			temp = strlen(buffer)+1;
			new->buffer = kmalloc(temp, GFP_ATOMIC);
			memcpy(new->buffer, buffer, temp);
			new->len = len;
			new->next = proc_head;
			new->prev = proc_head->prev;
			new->next->prev = new;
			new->prev->next = new;
		}
		break;
	case PCMD_EXPORT: /* output mode */
		len = 0;
		for (new = proc_head->next; new != proc_head;
			new = proc_head->next) {
			/* check if we're about to exceed count */
			if ((len + new->len) > count) {
				/* if we haven't reached count, write part
				 * of the current entry
				 */
				if (len < count) {
					/* copy out part of the entry */
					memcpy(page+len, new->buffer,
						count - len);
					/* remove the outputted data */
					memmove(new->buffer,
						new->buffer + (count - len),
						len + new->len - count + 1);
					new->len -= count - len;
				}
				/* we're not done yet, more printing is
				 * necessary
				 */
				*eof = 0;
				*start = page;
				return count;
			}

			memcpy(page+len, new->buffer, new->len);
			len += new->len;
			new->prev->next = new->next;
			new->next->prev = new->prev;
			kfree(new);
		}
		/* done printing */
		*eof = 1;
		*start = page;
		return len;
	}
	return 0;
}

#ifdef CONFIG_IMM_SIMPLE_PRIORITY_CONTROL
READ_PROC_PROTO(imm_priority_list_read_proc)
{
	READ_PROC_VARS
	struct imm_priority_t *p;
	int i;

        READ_PROC_INIT
	for (i = IMM_CLIENT_MAX_PRIORITY; i <= IMM_CLIENT_MIN_PRIORITY; i++) {
		for (p = priority_list.next; p != &priority_list; p = p->next){
			if (p->priority == i) {
				pprintf("Priority: %02d, Task Name: %s\n",
					p->priority, p->name);
			}
		}
	}
        READ_PROC_RETURN
}

READ_PROC_PROTO(imm_priority_control_read_proc) {
READ_PROC_VARS

	READ_PROC_INIT
	pprintf("Add or Remove IMM client priority list values.\n");
	pprintf("New tasks with names in this list will be automatically\n");
	pprintf("assigned the attached priority.\n");
	pprintf("The command \"add\" adds a new entry and \"del\" removes \
		one.\n");
	pprintf("example of add  %%> echo \"add taskname 4\" > \
		priority_control\n");
	pprintf("example of del  %%> echo \"del taskname\" > \
		priority_control\n");
	pprintf("%d is the highest priority, %d is the lowest\n",
		IMM_CLIENT_MAX_PRIORITY, IMM_CLIENT_MIN_PRIORITY);
	READ_PROC_RETURN
}

WRITE_PROC_PROTO(imm_priority_control_write_proc) {
char b[100], cmd[100], name[100];
int priority;

	copy_from_user(b, buffer, 100);
	sscanf(b, "%s %s %d", cmd, name, &priority);
	name[IMM_CLIENT_NAME_SIZE-1] = '\0';
	if (!strcmp(cmd, "add")) {
		if (VALID_PRIORITY(priority)) {
			if (imm_add_priority_list_entry(name, priority)) {
				printk(KERN_ERR "add %s %d failed\n", name,
					priority);
			}
		} else {
			printk(KERN_ERR "priority %d is out of range, must \
				be a value from %d (highest) to %d (lowest)\n",
				priority, IMM_CLIENT_MAX_PRIORITY,
				IMM_CLIENT_MIN_PRIORITY);
		}
	} else if (!strcmp(cmd, "del")) {
		if (imm_remove_priority_list_entry(name)) {
			printk(KERN_ERR "del %s failed\n", name);
		}
	} else {
		printk(KERN_ERR "Unknown command %s\n", cmd);
	}
	WRITE_PROC_RETURN
}
#endif

static int __init
imm_init (void)
{
#if defined(CONFIG_IMM_API_SRAM) || defined(CONFIG_IMM_API_CACHE)
int exit_cb_err = 1;
#endif
#ifdef CONFIG_IMM_API_CACHE
int unmap_cb_err = 1;
#endif
#ifdef CONFIG_IMM_SIMPLE_PRIORITY_CONTROL
struct proc_dir_entry *entry;
#endif

	kernel_immid = IMM_FIRST_KERNEL_IMMID;

#ifdef CONFIG_IMM_SIMPLE_PRIORITY_CONTROL
        priority_list.prev = &priority_list;
        priority_list.next = &priority_list;
#endif

#if defined(CONFIG_IMM_API_SRAM) || defined(CONFIG_IMM_API_CACHE)
	/* the exit callback is needed for SRAM and Cache garbage collection */
	exit_cb_err = profile_event_register(PROFILE_TASK_EXIT,
						&imm_task_exit_nb);
	if (exit_cb_err)
		printk(KERN_ERR "Process exit callback failed to register, \
				all APIs have been disabled\n");
#endif

#ifdef CONFIG_IMM_API_CACHE
	if (!exit_cb_err) {
		/* the unmap callback is needed for Cache garbage collection */
		unmap_cb_err = profile_event_register(PROFILE_DO_MUNMAP,
							&imm_do_munmap_nb);
		if (unmap_cb_err)
			printk(KERN_ERR "Process mem unmap callback failed \
					to register, the Cache lock API has \
					been disabled\n");
	}
#endif

	immid_list.immid = 0;
	immid_list.last_error = 0;
	immid_list.mm = NULL;
        immid_list.prev = &immid_list;
        immid_list.next = &immid_list;
#ifdef CONFIG_IMM_ALLOCATION_PRIORITY_SCHEME
	immid_list.priority = 0;
	immid_list.num_pages = 0;
#endif
#ifdef CONFIG_IMM_API_SRAM
	immid_list.malloc_list = NULL;
	immid_list.used_space = 0;
#endif

#ifdef CONFIG_IMM_API_CACHE
	immid_list.l2lock_list.addr = NULL;
	immid_list.l2lock_list.len = 0;
	immid_list.l2lock_list.prev = NULL;
	immid_list.l2lock_list.next = NULL;
#endif

	if ((proc_imm = proc_mkdir("imm", NULL)) == NULL) {
		printk(KERN_ERR "Failed to initialize the IMM proc \
				interface\n");
	} else {
#ifdef CONFIG_IMM_SIMPLE_PRIORITY_CONTROL
		REG_RW_PROC(entry, proc_imm, "priority_control",
				imm_priority_control_read_proc,
				imm_priority_control_write_proc)
		REG_READ_PROC(entry, proc_imm, "priority_list",
				imm_priority_list_read_proc)
#endif
	}

#ifdef CONFIG_IMM_API_SRAM
	/* if the exit cb is available */
	if (!exit_cb_err)
		imm_dau_sram_init(proc_imm);
#endif
#ifdef CONFIG_IMM_API_CACHE
	/* if the exit and unmap cbs are available */
	if (!exit_cb_err && !unmap_cb_err)
		imm_dau_cache_init(proc_imm);
#endif
#ifdef CONFIG_IMM_API_MEMATTR
	if (!exit_cb_err)
		imm_dau_memattr_init(proc_imm);
#endif
#ifdef CONFIG_IMM_POWER
	imm_autopower_init(proc_imm);
#endif
	pr_info("Intel(c) Memory Management is now Enabled\n");

	imm_initialized = 1;

	return 0;
}

__initcall(imm_init);
/* kernel hooks */

#if defined(CONFIG_IMM_API_SRAM) || defined(CONFIG_IMM_API_CACHE)
EXPORT_SYMBOL(imm_task_exit);
#endif

/* app/driver calls */
EXPORT_SYMBOL(imm_register_kernel);
EXPORT_SYMBOL(imm_register_user);
EXPORT_SYMBOL(imm_error);
EXPORT_SYMBOL(imm_get_physical);
EXPORT_SYMBOL(imm_get_virtual);
#ifdef CONFIG_IMM_ALLOCATION_PRIORITY_SCHEME
EXPORT_SYMBOL(imm_client_control);
#endif
