/*
 *  linux/arch/arm/mach-pxa/imm/power.c
 *
 *  Intel Memory Management
 *
 *  Memory Power Management Module
 *
 *  Todd Brandt
 *  Copyright (c) 2004, Intel Corporation (todd.e.brandt@intel.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.

 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/mm.h>
#include <linux/pm.h>
#include <linux/vmalloc.h>
#include <linux/linkage.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/pgalloc.h>
#include <asm/io.h>
#include <asm/bitops.h>
#include <asm/string.h>
#include <asm/arch/timex.h>
#include "sram.h"

/*************************************************************************
 *
 *  Defines and Macros
 *
 *************************************************************************/

#define NUM_DDR_PAGES			(PHYS_DDR_SIZE >> PAGE_SHIFT)
#define IMM_DDR_VSTART			(__phys_to_virt(PHYS_DDR_START))
#define IS_KERNEL_MEM(x) 		((x) <= (unsigned long)virt_to_phys(&_end))
#define IMM_MAX_PASR_DIVISION		(PHYS_PASR_LEVEL+2)
#define NEW_PASR_DIVISION(a, d)		((a <= (PHYS_DDR_START+(PHYS_DDR_SIZE/pasr_divisions[d])))?d:(d-1))

/*************************************************************************
 *
 *  External globals
 *
 *************************************************************************/

extern void imm_test_standby(void);

/* needed to determine the end of const kernel memory */
extern unsigned long _end;
extern unsigned long imm_pasr_stack_end;
extern unsigned long imm_pasr_stack_begin;

#ifdef CONFIG_IMM_POWER_TEST
/* cycles holders for testing the time for PASR */
extern unsigned int imm_pasr_cycles[6];
#endif

/*************************************************************************
 *
 *  Static globals
 *
 *************************************************************************/

/* used for power testing */
static int imm_pasr_disable = 0;

/* the set of pasr divisions available */
static const int pasr_divisions[] = {1, 2, 4, 8, 16};

/* the actual address boundaries for each pasr division */
static const int pasr_boundaries[] = {NUM_DDR_PAGES, NUM_DDR_PAGES>>1, NUM_DDR_PAGES>>2, NUM_DDR_PAGES>>3, NUM_DDR_PAGES>>4};

/* Extended Mode Register Setting values for enabling PASR */
static const unsigned int emrs_values[5] = {
#ifdef CONFIG_PXA3xx
        0x60010000, 0x40010001, 0x40010002, 0x40010005, 0x40010006
#else
        0x4000C000, 0x4000C001, 0x4000C002, 0x4000C005, 0x4000C006
#endif
};

/* A bitmap representing which pages must be moved to avoid PASR loss */
static unsigned long compressed_pagemap[NUM_DDR_PAGES >> 5];

/* The first page within the non-refreshed portion of memory */
static unsigned long pasr_boundary = NUM_DDR_PAGES;

/* The EMRS command to be sent to enable PASR */
static unsigned long emrs_enable = 0;

/* The EMRS command to be sent to disable PASR */
static unsigned long emrs_disable = 0;

/* The page table used just prior to PASR compression */
static unsigned long last_pgd = 0;

/* A switch used to determine whether or not SRAM powerdown should be conducted */
int imm_srampower_disable = 0;

/*************************************************************************
 *
 *  Public globals
 *
 *************************************************************************/

/* A page table whose physical memory is guaranteed to be safe from PASR loss */
unsigned long imm_pasr_safe_pgd = 0;

/* A switch used to determine whether or not PASR should be conducted */
int imm_pasr_disable_once = 0;

/*----------------------------------------------------------------------
 *
 *  High Level Call Tree (for power.c and pasr-stack.S):
 *
 *  [STANDBY - Called from mhn_cpu_standby (in mhn_suspend.S)]
 *	imm_power_down
 *		pasr_calculate
 *		sram_powersave_enable
 *		pasr_memory_compress
 *	imm_power_up
 *		pasr_memory_decompress
 *		sram_powersave_disable
 *
 *  [SLEEP - Called from mhn_cpu_sleep (in mhn_suspend.S)]
 *	imm_power_down
 *		pasr_calculate
 *		sram_powersave_enable
 *		pasr_memory_compress
 *	imm_power_up
 *		pasr_memory_decompress
 *		sram_powersave_disable
 *
 *  [Proc Filesystem]
 *	/proc/imm/ddr_compression (read-only)
 *		imm_ddr_compression_read_proc
 *	/proc/imm/ddr_powersave (read-only)
 *		imm_ddr_powersave_read_proc
 *	/proc/imm/zone_map (read-only)
 *		imm_zone_map_read_proc
 *	/proc/imm/pasr_test (read/write)
 *		imm_pasr_test_read_proc
 *		imm_pasr_test_write_proc
 *	/proc/imm/power_test (read/write)
 *		imm_power_test_read_proc
 *		imm_power_test_write_proc
 *	/proc/imm/ddr_usagemap (read-only)
 *		imm_ddr_usagemap_read_proc
 *
 *  [Module Initialization - from Policy Manager]
 *	imm_autopower_init
 *
 *--------------------------------------------------------------------*/

/*************************************************************************
 *
 * Function: sram_powersave_enable
 * Description: Determines current SRAM usage map and shuts down any
 * unused banks
 *
 *************************************************************************/

#ifdef CONFIG_IMM_API_SRAM
void
sram_powersave_enable(void)
{
int i, j, used;
#ifdef CONFIG_PXA3xx
u32 ad1r_val, ad2r_val, ad3r_val;
#else
u32 pstr_val, pslr_val, psnr_val;
#endif
	if(imm_srampower_disable) return;

#ifdef CONFIG_PXA3xx
	/* D1 SRAM control */
	ad1r_val = AD1R;
	/* D2 SRAM control */
	ad2r_val = AD2R;
	/* D3 SRAM control */
	ad3r_val = AD3R;
#else
	/* Standby SRAM control */
	pstr_val = PSTR;
	/* Sleep SRAM control */
	pslr_val = PSLR;
	/* Sense SRAM control */
	psnr_val = PSNR;
#endif
	for(i = IMM_SRAM_FIRST_BANK; i < PHYS_SRAM_BANK_COUNT; i++) {
		used = 0;
		for(j = SRAM_ARRAY_STARTPAGE(i);
			(j < SRAM_ARRAY_STARTPAGE(i+1))&&!used;
			j++)
		{
			if(PAGE_IS_MAPPED(page_list[j].flag))
				used = 1;
		}
#ifdef CONFIG_PXA3xx
		if(used) {
			ad1r_val |= (0x1 << i);
			ad2r_val |= (0x1 << i);
			ad3r_val |= (0x1 << i);
                } else { /* disable this array */
			ad1r_val &= ~(0x1 << i);
			ad2r_val &= ~(0x1 << i);
			ad3r_val &= ~(0x1 << i);
		}
#else
		if(used) {
			pstr_val |= (0x100 << i);
			pslr_val |= (0x100 << i);
			psnr_val |= (0x100 << i);
                } else { /* disable this array */
			pstr_val &= ~(0x100 << i);
			pslr_val &= ~(0x100 << i);
			psnr_val &= ~(0x100 << i);
		}
#endif
	}

#ifdef CONFIG_PXA3xx
	AD1R = ad1r_val | 0x100;
	AD2R = ad2r_val | 0x100;
	AD3R = ad3r_val | 0x100;
#else
	PSTR = pstr_val;
	PSLR = pslr_val;
	PSNR = psnr_val;
#endif
}

void
sram_powersave_disable(void)
{
#ifdef CONFIG_PXA3xx
u32 ad1r_val, ad2r_val, ad3r_val;

	if(imm_srampower_disable) return;

	/* D1 SRAM control */
	ad1r_val = AD1R;
	/* D2 SRAM control */
	ad2r_val = AD2R;
	/* D3 SRAM control */
	ad3r_val = AD3R;

	ad1r_val |= 0xff;
	ad1r_val &= ~(0x100);
	ad2r_val |= 0xff;
	ad2r_val &= ~(0x100);
	ad3r_val |= 0xff;
	ad3r_val &= ~(0x100);

	AD1R = ad1r_val;
	AD2R = ad2r_val;
	AD3R = ad3r_val;
#else
u32 pstr_val, pslr_val, psnr_val;

	if(imm_srampower_disable) return;

	/* Standby SRAM control */
	pstr_val = PSTR;
	/* Sleep SRAM control */
	pslr_val = PSLR;
	/* Sense SRAM control */
	psnr_val = PSNR;

	pstr_val |= 0xf00;
	psnr_val |= 0xf00;
	pslr_val |= 0xf00;

	PSTR = pstr_val;
	PSLR = pslr_val;
	PSNR = psnr_val;
#endif
}
#endif

void
pasr_memory_compress(void)
{
	int i = 0, j = pasr_boundary;
	unsigned long *src, *dst;

	if(imm_pasr_disable_once) return;

#ifdef CONFIG_PXA3xx
        MDMRS = emrs_enable;
#else
        MDMRSLP = emrs_enable;
#endif

	/* retrieve the current page table base pointer */
	asm( "mrc p15, 0, %0, c2, c0, 0" : "=r" (last_pgd) );

	/* change the page table to be the init table (in kernel mem) */
	cpu_do_switch_mm(imm_pasr_safe_pgd, NULL);

	/* now carry out the compression, i is the live side iterator, j is dead side */
	while(1)
	{
		/* find a destination page on the live side */
		while((i < pasr_boundary)&&!(__test_bit(i, compressed_pagemap))) i++;

		/* find a source page on the dead side */
		while((j < NUM_DDR_PAGES)&&!(__test_bit(j, compressed_pagemap))) j++;

		/* neither i or j should ever reach their limits separately */
		if(((i < pasr_boundary)&&(j >= NUM_DDR_PAGES))||
			((i >= pasr_boundary)&&(j < NUM_DDR_PAGES)))
		{
			BUG();
		}

		if((i >= pasr_boundary)&&(j >= NUM_DDR_PAGES)) break;

		/* now determine the src and dst page addresses for the copy */
		src = (unsigned long *)(IMM_DDR_VSTART + (j++ << PAGE_SHIFT));
		dst = (unsigned long *)(IMM_DDR_VSTART + (i++ << PAGE_SHIFT));

		copy_page(dst, src);
	}
}

void
pasr_memory_decompress(void)
{
	int i = 0, j = pasr_boundary;
	unsigned long *src, *dst;

	if(imm_pasr_disable_once) return;
#ifdef CONFIG_PXA3xx
	MDMRS = emrs_disable;
#else
	MDMRSLP = emrs_disable;
#endif

	/* now carry out the decompression, i is the live side iterator, j is dead side */
	while(1)
	{
		/* find a source page on the live side */
		while((i < pasr_boundary)&&!(__test_bit(i, compressed_pagemap))) i++;

		/* find a destination page on the dead side */
		while((j < NUM_DDR_PAGES)&&!(__test_bit(j, compressed_pagemap))) j++;

		/* neither i or j should ever reach their limits separately */
		if(((i < pasr_boundary)&&(j >= NUM_DDR_PAGES))||
			((i >= pasr_boundary)&&(j < NUM_DDR_PAGES)))
		{
			BUG();
		}

		if((i >= pasr_boundary)&&(j >= NUM_DDR_PAGES)) break;

		/* now determine the src and dst page addresses for the copy */
		src = (unsigned long *)(IMM_DDR_VSTART + (i++ << PAGE_SHIFT));
		dst = (unsigned long *)(IMM_DDR_VSTART + (j++ << PAGE_SHIFT));

		copy_page(dst, src);
	}

	/* change the page table to be the init table (in kernel mem) */
	cpu_do_switch_mm(last_pgd, NULL);
}

/*************************************************************************
 *
 * Function: pasr_calculate
 * Description: Determines current DDR usage map, then sets up all the
 * globals used by pxa_cpu_suspend, pasr_memory_compress, and
 * pasr_memory_decompress
 *
 *************************************************************************/

void
pasr_calculate(void)
{
	unsigned long flags, address, paddress;
	int i, j;
	struct page *p;
	struct list_head *curr;
	struct zone *zone;
	unsigned long pagenum = 0;

	imm_pasr_disable_once = imm_pasr_disable;

	if(imm_pasr_disable_once) return;

	/* by default PASR is unused */
	pasr_boundary = NUM_DDR_PAGES;
	emrs_enable = emrs_values[0];
	emrs_disable = emrs_values[0];

	/* start by switching all pages to used */
	for(i = 0; i < (NUM_DDR_PAGES >> 5); i++) {
		compressed_pagemap[i] = 0xFFFFFFFF;
	}

	/* set any pages that linux isn't using at all to free */
	for(i = (num_physpages >> 5); i < (NUM_DDR_PAGES >> 5); i++) {
		compressed_pagemap[i] = 0;
	}

	/* add unused pages to the count of free pages */
	if(NUM_DDR_PAGES > num_physpages) {
		pagenum = NUM_DDR_PAGES - num_physpages;
	}

	/* now let's switch off all the pages we know to be free */
	for_each_zone(zone) {
		/* if no pages are present in this zone, skip it */
		if (!zone->present_pages) continue;

		/* we must use the spinlock to safely access the zone data */
		spin_lock_irqsave(&zone->lock, flags);
		for(i = MAX_ORDER-1; i >= 0; i--) {
			list_for_each(curr, &zone->free_area[i].free_list) {
				p = list_entry(curr, struct page, lru);
				address = (unsigned long)page_address(p);
				if(!IS_KERNEL_MEM(address)) {
					pagenum += (1 << i);
					for(paddress = address;
						paddress < (address + (1 << (i+PAGE_SHIFT)));
						paddress += PAGE_SIZE) {
						__clear_bit((paddress-IMM_DDR_VSTART)>>PAGE_SHIFT, compressed_pagemap);
					}
				}
			}
		}
		spin_unlock_irqrestore(&zone->lock, flags);
	}

	/* what fraction of memory is used */
	pagenum = NUM_DDR_PAGES - pagenum;

	/* determine the most effective PASR level */
	for(i = IMM_MAX_PASR_DIVISION-1; (i >= 0)&&(pagenum > pasr_boundaries[i]); i--);

	/* if half or more of DDR is full, PASR is useless... abort */
	if(i == 0) {
		imm_pasr_disable_once = 1;
		return;
	}

	/* pasr boundary is the upper limit of where pages can stay active in standby/sleep */
	pasr_boundary = pasr_boundaries[i];
	emrs_enable = emrs_values[i];

	/* The physical DDR memory map is now broken into two regions during */
	/* standby/sleep according to the PASR boundary. Memory before */
	/* the boundary is retained (live side), memory after is lost (dead side) */
	/*  ----------------------------------------------------------  */
	/* |..live side...|PASR boundary|...........dead side.........| */
	/*  ----------------------------------------------------------  */
	/* We have to flag every page on the dead side to be moved to */
	/* free spots on the live side so no info gets lost. Pages already */
	/* on the live side are not flagged because nothing must be done */
	/* to them when we come back from standby/sleep */

	/* NOTE - the code that carries out the actual movement MUST exist */
	/* exclusively on the live side, and use only static memory */

	for(i = 0, j = pasr_boundary; j < NUM_DDR_PAGES; j++)
	{
		/* we found a page on the soon-to-be dead side */
		if(__test_bit(j, compressed_pagemap))
		{
			/* scan for the first empty page on the live side */
			while((i < pasr_boundary)&&(__test_bit(i, compressed_pagemap)))
			{
				/* zero all the used page flags since we won't need to move them */
				__clear_bit(i++, compressed_pagemap);
			}
			/* this case denotes a calculation error and should never occur */
			if(i >= pasr_boundary)
			{
				printk(KERN_ERR "ERROR: compressed bitmap calculation error\n");
				pasr_boundary = NUM_DDR_PAGES;
				return;
			}
			/* we set this bit to denote where j will be moved */
			__set_bit(i++, compressed_pagemap);
		}
	}
	/* now clean up the rest of the live side */
	while(i < pasr_boundary)
	{
		/* zero all the used pages since we won't need to move them */
		__clear_bit(i++, compressed_pagemap);
	}
}

/*************************************************************************
 *
 * Everything below this point is in service of the proc interface
 *
 *************************************************************************/

#define MGLINES 18
const char *mem_graphic[MGLINES] = {
        "  ________________________________________________________________ ",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        " |UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU|",
        "  ---------------------------------------------------------------- "
};
#define DDR_MEMGRAPHIC_LINE(a)        (((((a)-PHYS_DDR_START)<<4)/PHYS_DDR_SIZE)+1)
#define DDR_MEMGRAPHIC_CHAR(a)        ((((((a)-PHYS_DDR_START)%(PHYS_DDR_SIZE>>4))<<6)/(PHYS_DDR_SIZE>>4))+2)
#define DDR_MEMGRAPHIC_SETFREE(a)     chip_graphic[DDR_MEMGRAPHIC_LINE(a)][DDR_MEMGRAPHIC_CHAR(a)] = '.'

READ_PROC_PROTO(imm_ddr_usagemap_read_proc) {
READ_PROC_VARS
unsigned long flags, address, vaddress;
int i;
struct page *p;
struct list_head *curr;
struct zone *zone;
unsigned long paddress, total_free=0;
char *chip_graphic[MGLINES];

	READ_PROC_INIT
	/* create copies of the memory graphic to be displayed */
	for(i = 0; i < MGLINES; i++) {
		chip_graphic[i] = kmalloc(strlen(mem_graphic[i])+1, GFP_KERNEL);
		strcpy(chip_graphic[i], mem_graphic[i]);
	}

	/* let's loop through each zone and check which pages are used */
	for_each_zone(zone) {
		/* if no pages are present in this zone, skip it */
		if (!zone->present_pages) continue;

		/* we must use the spinlock to safely access the zone data */
		spin_lock_irqsave(&zone->lock, flags);

		for(i = MAX_ORDER-1; i >= 0; i--) {
			list_for_each(curr, &zone->free_area[i].free_list) {
				p = list_entry(curr, struct page, lru);
				vaddress = (unsigned long)page_address(p);
				address = __virt_to_phys(vaddress);

				if(!IS_KERNEL_MEM(vaddress)) {
					for(paddress = address;
						paddress < (address + (1 << (i+PAGE_SHIFT)));
						paddress += PAGE_SIZE)
					{
						DDR_MEMGRAPHIC_SETFREE(paddress);
						total_free++;
					}
				}
			}
		}
		spin_unlock_irqrestore(&zone->lock, flags);
	}

	/* set any pages that linux isn't using at all to free */
	for(i = num_physpages; i < NUM_DDR_PAGES; i++) {
		DDR_MEMGRAPHIC_SETFREE((i*PAGE_SIZE)+PHYS_OFFSET);
		total_free++;
	}

	/* at this point we now have the PASR divisor settings */
        pprintf("\nDDR: Start = %08X, Size = %dM (%dMx16)\n",
                 (unsigned int)PHYS_DDR_START, PHYS_DDR_SIZE/0x100000,
                 PHYS_DDR_SIZE/0x1000000);
        pprintf("DDR: Has %d free pages (%dK)\n", (unsigned int)total_free,
                 (unsigned int)(total_free << 2));
        for(i = 0; i < MGLINES; i++) {
		pprintf("%s\n", chip_graphic[i]);
        }
	pprintf("\n");
	for(i = 0; i < MGLINES; i++) {
		kfree(chip_graphic[i]);
	}
	READ_PROC_RETURN
}

#ifdef CONFIG_IMM_POWER_TEST
READ_PROC_PROTO(imm_power_test_read_proc) {
READ_PROC_VARS

	READ_PROC_INIT
        pprintf("Write a number to standby/sleep the system and test memory power:\n\n");
        pprintf("1: STANDBY with no IMM\n");
        pprintf("2: SLEEP   with no IMM\n");
        pprintf("3: STANDBY with SRAM Powersave\n");
        pprintf("4: SLEEP   with SRAM Powersave\n");
        pprintf("5: STANDBY with DRAM PASR\n");
        pprintf("6: SLEEP   with DRAM PASR\n");
        pprintf("7: STANDBY with full IMM\n");
        pprintf("8: SLEEP   with full IMM\n");
	READ_PROC_RETURN
}

WRITE_PROC_PROTO(imm_power_test_write_proc) {
WRITE_PROC_VAR_NUM(5)
char *state_name[6] = {"", "STANDBY", "", "SLEEP", "PM_SUSPEND_DISK", "PM_SUSPEND_MAX"};
int state = 0, num;

	imm_pasr_disable = 0;
	imm_srampower_disable = 0;
	WRITE_PROC_GET_NUM(num, 5)
	switch(num) {
	case 1: /* standby - no IMM */
		state = PM_SUSPEND_STANDBY;
		imm_pasr_disable = 1;
		imm_srampower_disable = 1;
		printk(KERN_INFO "POWER TEST: Standby with IMM Disabled\n");
		break;
	case 2: /* sleep - no IMM */
		state = PM_SUSPEND_MEM;
		imm_pasr_disable = 1;
		imm_srampower_disable = 1;
		printk(KERN_INFO "POWER TEST: Sleep with IMM Disabled\n");
		break;
	case 3: /* standby - SPD on */
		state = PM_SUSPEND_STANDBY;
		imm_pasr_disable = 1;
		printk(KERN_INFO "POWER TEST: Standby with SRAM Array Powerdown\n");
		break;
	case 4: /* sleep - SPD on */
		state = PM_SUSPEND_MEM;
		imm_pasr_disable = 1;
		printk(KERN_INFO "POWER TEST: Sleep with SRAM Array Powerdown\n");
		break;
	case 5: /* standby - PASR on */
		state = PM_SUSPEND_STANDBY;
		imm_srampower_disable = 1;
		printk(KERN_INFO "POWER TEST: Standby with PASR\n");
		break;
	case 6: /* sleep - PASR on */
		state = PM_SUSPEND_MEM;
		imm_srampower_disable = 1;
		printk(KERN_INFO "POWER TEST: Sleep with PASR\n");
		break;
	case 7: /* standby - full IMM */
		state = PM_SUSPEND_STANDBY;
		printk(KERN_INFO "POWER TEST: Standby with FULL IMM\n");
		break;
	case 8: /* sleep - full IMM */
		state = PM_SUSPEND_MEM;
		printk(KERN_INFO "POWER TEST: Sleep with FULL IMM\n");
		break;
	default:
		printk(KERN_INFO "Error, invalid power_test number (valid values are 1, 2, 3, or 4)\n");
	}
	if(state) {
		printk(KERN_INFO "ENTERING %s STATE\n", state_name[state]);
		imm_pasr_cycles[0] = OSCR;
		pm_suspend(state);
		imm_pasr_cycles[3] = OSCR;
		printk(KERN_INFO "RETURNED FROM %s STATE\n", state_name[state]);
		printk(KERN_INFO "Entry Time = %d ms (%d OSCR ticks)\n",
			(imm_pasr_cycles[1] - imm_pasr_cycles[0])/(CLOCK_TICK_RATE/1000),
			imm_pasr_cycles[1] - imm_pasr_cycles[0]);
		printk(KERN_INFO " Exit Time = %d ms (%d OSCR ticks)\n",
			(imm_pasr_cycles[3] - imm_pasr_cycles[2])/(CLOCK_TICK_RATE/1000),
			imm_pasr_cycles[3] - imm_pasr_cycles[2]);
	}
	imm_pasr_disable = 0;
	imm_srampower_disable = 0;
	WRITE_PROC_RETURN
}
#endif

READ_PROC_PROTO(imm_zone_map_read_proc) {
READ_PROC_VARS
int i;
struct page *p;
unsigned long flags, vaddress, total=0;
struct list_head *curr;
struct zone *zone;

	READ_PROC_INIT
	/* let's loop through each zone and check which pages are used */
	for_each_zone(zone) {
		/* if no pages are present in this zone, skip it */
		if (!zone->present_pages) continue;

		pprintf("Zone = %s\n", zone->name);

		/* we must use the spinlock to safely access the zone data */
		spin_lock_irqsave(&zone->lock, flags);

		for(i = MAX_ORDER-1; i >= 0; i--) {
			list_for_each(curr, &zone->free_area[i].free_list) {
				p = list_entry(curr, struct page, lru);
				vaddress = (unsigned long)page_address(p);
				if(IS_KERNEL_MEM(vaddress)) { pprintf("[KERNEL] "); }
				pprintf("ORDER %02d: Free %4dK block at %08X\n",
					i, (1 << (i+2)), (unsigned int)__virt_to_phys(vaddress));
				total += (1 << i);
			}
		}
		spin_unlock_irqrestore(&zone->lock, flags);
	}
	READ_PROC_RETURN
}

READ_PROC_PROTO(imm_ddr_powersave_read_proc) {
READ_PROC_VARS
int pasr, i;
struct page *p;
unsigned long frac, flags, freemem = 0;
struct list_head *curr;
struct zone *zone;

	READ_PROC_INIT
	/* add unused pages to the count of free pages */
	if(NUM_DDR_PAGES > num_physpages) {
		freemem = NUM_DDR_PAGES - num_physpages;
	}

	/* let's loop through each zone and check which pages are used */
	for_each_zone(zone) {
		/* if no pages are present in this zone, skip it */
		if (!zone->present_pages) continue;

		/* we must use the spinlock to safely access the zone data */
		spin_lock_irqsave(&zone->lock, flags);

		for(i = MAX_ORDER-1; i >= 0; i--) {
			list_for_each(curr, &zone->free_area[i].free_list) {
				p = list_entry(curr, struct page, lru);
				if(!IS_KERNEL_MEM((unsigned long)page_address(p)))
					freemem += (1 << i);
			}
		}
		spin_unlock_irqrestore(&zone->lock, flags);
	}

	if(freemem >= NUM_DDR_PAGES) {
		pprintf("Error: Free pages outnumber total pages\n");
		READ_PROC_RETURN
	}

	frac = NUM_DDR_PAGES / (NUM_DDR_PAGES - freemem);
	for(pasr = 0; pasr < IMM_MAX_PASR_DIVISION; pasr++) {
		if((frac >= pasr_divisions[pasr]) && (frac <= pasr_divisions[pasr+1])) {
			if(frac == pasr_divisions[pasr+1])
				pasr += 1;
			break;
		}
	}

	pprintf("Total Memory = %dK, Used Memory = %dK, Free Memory = %dK, PASR = 1/%d\n",
		PHYS_DDR_SIZE/1024, (unsigned int)(NUM_DDR_PAGES-freemem)<<2, (unsigned int)freemem<<2, pasr_divisions[pasr]);
	READ_PROC_RETURN
}

READ_PROC_PROTO(imm_ddr_compression_read_proc) {
READ_PROC_VARS
int i, j, cnt = 0;
unsigned long src, dst;

	READ_PROC_INIT
	pasr_calculate();
	if(pasr_boundary == NUM_DDR_PAGES) {
		pprintf("Memory Usage is greater than half of DDR, PASR is useless\n");
		READ_PROC_RETURN
	}
	for(i = 0, j = pasr_boundary; (i < pasr_boundary)&&(j < NUM_DDR_PAGES);)
	{
		while((i < pasr_boundary)&&!(__test_bit(i, compressed_pagemap))) i++;
		while((j < NUM_DDR_PAGES)&&!(__test_bit(j, compressed_pagemap))) j++;
		if((i < pasr_boundary)&&(j >= NUM_DDR_PAGES))
		{
			pprintf("ERROR: There are more destination than source page flags\n");
			READ_PROC_RETURN
		}
		if((i >= pasr_boundary)&&(j < NUM_DDR_PAGES))
		{
			pprintf("ERROR: There are more source than destination page flags\n");
			READ_PROC_RETURN
		}
		src = PHYS_DDR_START + (j++ << PAGE_SHIFT);
		dst = PHYS_DDR_START + (i++ << PAGE_SHIFT);
		cnt++;

		pprintf("Page move from %08X to %08X\n", (unsigned int)src, (unsigned int)dst);
	}

	pprintf("PASR Can be set to 1/%d of memory if:\n",
		(unsigned int)(NUM_DDR_PAGES / pasr_boundary));
	pprintf("A total of %d pages (%dK) are moved from upper memory\n",
		cnt, cnt<<2);
	READ_PROC_RETURN
}

void
imm_autopower_init (struct proc_dir_entry *imm_dir)
{
struct proc_dir_entry *entry;

	imm_pasr_safe_pgd = __pa(init_mm.pgd);

	if(imm_dir) {
#ifdef CONFIG_IMM_POWER_TEST
		REG_RW_PROC(entry, imm_dir, "power_test", imm_power_test_read_proc, imm_power_test_write_proc)
#endif
		REG_READ_PROC(entry, imm_dir, "zone_map", imm_zone_map_read_proc)
		REG_READ_PROC(entry, imm_dir, "ddr_powersave", imm_ddr_powersave_read_proc)
		REG_READ_PROC(entry, imm_dir, "ddr_usagemap", imm_ddr_usagemap_read_proc)
		REG_READ_PROC(entry, imm_dir, "ddr_compression", imm_ddr_compression_read_proc)
	}

	printk(KERN_INFO "Intel(c) Memory Management - Memory Power Savings\n");
}
