/*
 *  linux/arch/arm/mach-pxa/imm/sram.h
 *
 *  Intel Memory Management
 *
 *  SRAM API Defines/Globals/Functions
 *
 *  Todd Brandt
 *  Copyright (c) 2004, Intel Corporation (todd.e.brandt@intel.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

#ifndef _ARCH_PXA_SRAM_H
#define _ARCH_PXA_SRAM_H

#include "imm.h"

/* indexes for page_list */
#define SRAMSTART	(0)
#define SRAMEND		(IMM_SRAM_PAGES)
#define L2CACHESTART	(SRAMEND)
#define L2CACHEEND	(SRAMEND + IMM_L2CACHE_PAGES)
#define DRAMSTART	(L2CACHEEND)
#define DRAMEND		(IMM_MEMORY_PAGES)

#define MEMSTARTP	(&page_list[0])
#define MEMENDP		(&page_list[IMM_MEMORY_PAGES])
#define SRAMSTARTP	(&page_list[SRAMSTART])
#define SRAMENDP	(&page_list[SRAMEND])
#define L2CACHESTARTP	(&page_list[L2CACHESTART])
#define L2CACHEENDP	(&page_list[L2CACHEEND])
#define DRAMSTARTP	(&page_list[DRAMSTART])
#define DRAMENDP	(&page_list[DRAMEND])

#if (IMM_SRAM_PAGES > 0)
#define IMM_SRAM_TYPE		(1)
#else
#define IMM_SRAM_TYPE		(0)
#endif
#if (IMM_L2CACHE_PAGES > 0)
#define IMM_L2CACHE_TYPE	(1)
#else
#define IMM_L2CACHE_TYPE	(0)
#endif
#if (IMM_DRAM_PAGES > 0)
#define IMM_DRAM_TYPE		(1)
#else
#define IMM_DRAM_TYPE		(0)
#endif
#define NUM_MEMORY_TYPES	(IMM_SRAM_TYPE + IMM_L2CACHE_TYPE + IMM_DRAM_TYPE)

extern struct imm_page_t *mstart[NUM_MEMORY_TYPES];
extern struct imm_page_t *mend[NUM_MEMORY_TYPES];
extern const u32 mflag[NUM_MEMORY_TYPES];

#define MEMORY_PAGE_VADDRESS(i)	(((i) < IMM_SRAM_PAGES)?(sram_vmem + (PAGE_SIZE * (i))):(dram_vmem + (PAGE_SIZE * ((i)-IMM_SRAM_PAGES))))
#define MEMORY_PAGE_PADDRESS(i)	(((i) < IMM_SRAM_PAGES)?(sram_pmem + (PAGE_SIZE * (i))):(dram_pmem + (PAGE_SIZE * ((i)-IMM_SRAM_PAGES))))
#define SRAM_PHYS_ADDRESS(i)	(IMM_SRAM_START + ((i)*PAGE_SIZE))

/* imm internal flags */
#define IMM_INFO_USED           0x10
#define IMM_INFO_RESERVED   	0x20
#define IMM_INFO_REPRIORITIZE	0x40
#define IMM_INFO_DMA		IMM_MALLOC_HARDWARE
#define IMM_INFO_SRAM		IMM_MALLOC_SRAM
#define IMM_INFO_DRAM		IMM_MALLOC_DRAM
#define IMM_INFO_L2CACHE	IMM_MALLOC_L2CACHE

/* defines and macros */
#define PAGE_IS_MAPPED(x)	((x) & IMM_INFO_USED)
#define PAGE_IS_DMA(x)		((x) & IMM_INFO_DMA)
#if (IMM_SRAM_PAGES > 0)
#define PAGE_IS_SRAM(x)		((x) & IMM_INFO_SRAM)
#else
#define PAGE_IS_SRAM(x)		(0)
#endif
#if (IMM_DRAM_PAGES > 0)
#define PAGE_IS_DRAM(x)		((x) & IMM_INFO_DRAM)
#else
#define PAGE_IS_DRAM(x)		(0)
#endif
#if (IMM_L2CACHE_PAGES > 0)
#define PAGE_IS_L2CACHE(x)	((x) & IMM_INFO_L2CACHE)
#else
#define PAGE_IS_L2CACHE(x)	(0)
#endif
#define PAGE_IS_RESERVED(x)	((x) & IMM_INFO_RESERVED)
#define PAGE_IS_REPRIORITIZE(x)	((x) & IMM_INFO_REPRIORITIZE)
#define PAGE_IS_TYPE(x)		(PAGE_IS_L2CACHE(x)?"L2CACHE":(PAGE_IS_SRAM(x)?"SRAM":"DRAM"))
#define PAGE_IS_TYPEC(x)	(PAGE_IS_L2CACHE(x)?'L':(PAGE_IS_SRAM(x)?'S':'D'))

#define SRAM_ARRAY_STARTPAGE(x) (((x)-IMM_SRAM_FIRST_BANK)*(IMM_SRAM_PAGES/IMM_SRAM_BANK_COUNT))

/* entries for the global physical memory list */
struct imm_page_t {
	u8 flag;		/* flags */
	u32 immid;		/* each page is associated with a process id */
	u32 virt_addr;		/* virtual address for this page (into the immid's space) */
	u32 phys_addr;		/* fixed physical address associated with this page */
#ifdef CONFIG_IMM_ALLOCATION_PRIORITY_SCHEME
	s8 priority;		/* client priority */
#endif
#ifdef CONFIG_IMM_HETEROGENEOUS_MEMORY
	u32 kern_addr;		/* fixed virtual address used by IMM to access this page */
#endif
#ifdef CONFIG_IMM_DEBUG
	s16 index;
#endif
};

/* global data */
extern u32 num_free_sram_pages;
extern u32 num_free_dram_pages;
extern u32 cache_space_left;

#ifdef CONFIG_IMM_HETEROGENEOUS_MEMORY
#define MEMSPACELEFT (num_free_sram_pages + num_free_dram_pages + (cache_space_left/PAGE_SIZE))
#else
#define MEMSPACELEFT (num_free_sram_pages)
#endif

#ifdef CONFIG_IMM_ALLOCATION_PRIORITY_SCHEME
extern int imm_alloc_pages_hotswap(struct imm_page_t **, struct imm_info_t *, s16);
extern int imm_alloc_pages_hotswap_dma(struct imm_page_t **, struct imm_info_t *, s16);
extern int imm_free_pages_hotswap(struct imm_page_t **, s16, s16, struct imm_info_t *);

#define PRIORITY_INDEX(x)		(IMM_CLIENT_MIN_PRIORITY - (x))
#define PRIORITY_FROM_INDEX(x)		(IMM_CLIENT_MIN_PRIORITY - (x))
#define MAX_PRIORITY_INDEX		(IMM_CLIENT_MIN_PRIORITY - IMM_CLIENT_MAX_PRIORITY)

extern u32 num_l2cache_pages[IMM_CLIENT_NUM_PRIORITIES];
extern u32 num_sram_pages[IMM_CLIENT_NUM_PRIORITIES];
extern u32 num_dram_pages[IMM_CLIENT_NUM_PRIORITIES];

#define INCREMENT_PAGECOUNT(f, x)	if(PAGE_IS_L2CACHE(f)) { num_l2cache_pages[PRIORITY_INDEX(x)]++; } \
else if(PAGE_IS_SRAM(f)) { num_sram_pages[PRIORITY_INDEX(x)]++; } \
else { num_dram_pages[PRIORITY_INDEX(x)]++; }
#define DECREMENT_PAGECOUNT(f, x)	if(PAGE_IS_L2CACHE(f)) { num_l2cache_pages[PRIORITY_INDEX(x)]--; } \
else if(PAGE_IS_SRAM(f)) { num_sram_pages[PRIORITY_INDEX(x)]--; } \
else { num_dram_pages[PRIORITY_INDEX(x)]--; }
#endif

#ifdef CONFIG_IMM_DEBUG
extern void print_page_map(int, struct imm_page_t **, s16);
extern void set_plist(int, int, int);
#endif

/* physical memory page list */
extern struct imm_page_t page_list[IMM_MEMORY_PAGES];
extern struct imm_info_t immid_list;

static inline void
page_map_reverse(struct imm_page_t **page_map, int n)
{
int i, j;
struct imm_page_t *p;

	for(i = 0, j = n - 1; i < j; i++, j--)
	{
		p = page_map[i];
		page_map[i] = page_map[j];
		page_map[j] = p;
	}
}
#endif
