#include <linux/module.h>
#include <linux/err.h>
#include <linux/i2c.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/platform_device.h>
#include <asm/arch/pxa3xx_pmic.h>
#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <linux/delay.h>
#include <asm/arch/max8660.h>
#include <asm/arch/pxa3xx_gpio.h>
#include <asm/arch/mfp.h>

#define MAX8660_ADDRESS 0x34

struct max8660 {
	struct mutex lock;
	struct i2c_client *client;
	u8 shadow_regs[MAX8660_N_REGS];
};

static struct max8660 *the_max8660;

static const u8 defined_regs[] = {
	0x10, 0x12, 
	0x20, 0x23, 0x24, 0x29, 0x2A, 
	0x32, 0x33, 0x39, 
	0x80,
};

static int init_succ = 0;

void connect_si2c(void)
{
	pxa3xx_gpio_set_level(MFP_I2C_CTRL, GPIO_LEVEL_HIGH);
}

void disconnect_si2c(void)
{
	pxa3xx_gpio_set_level(MFP_I2C_CTRL, GPIO_LEVEL_LOW);
}

static inline int is_reg_valid(u32 reg)
{
	int i;
	for (i = 0; i < ARRAY_SIZE(defined_regs); i++)
		if (reg == defined_regs[i])
			return 1;

	return 0;
}

static int max8660_read(u8 reg, u8 *pval)
{
	if(!is_reg_valid(reg) || (the_max8660 == NULL) || !init_succ)
	{
		return -EINVAL;
	}
	
	*pval = the_max8660->shadow_regs[reg];
	
	return 0;
}

static int max8660_write(u8 reg, u8 mask, u8 val)
{
	int ret = 0;
	u8 reg_val;
		
	if (!is_reg_valid(reg) || (the_max8660 == NULL) || !init_succ)
	{
		return -EINVAL;
	}
	
	reg_val = (the_max8660->shadow_regs[reg] & mask) | val;
	
	connect_si2c();
	ret = i2c_smbus_write_byte_data(the_max8660->client, reg, reg_val);
	disconnect_si2c();
	if (ret == 0)
	{
		the_max8660->shadow_regs[reg] = reg_val;
	}
	
	return ret;
}

int pxa3xx_pmic_read(u8 reg, u8 *pval)
{
        return max8660_read(reg, pval);
}
EXPORT_SYMBOL(pxa3xx_pmic_read);

int pxa3xx_pmic_write(u8 reg, u8 val)
{
        return max8660_write(reg, 0x00, val);
}
EXPORT_SYMBOL(pxa3xx_pmic_write);

static int enable_max8660_voltage(int cmd, int enable)
{
	int ret = 0;

	if(!init_succ)
	{
		return -1;
	}
	
	switch(cmd)
	{
		case VCC_CIF:
		{			
			if(!enable)
			{
				ret = max8660_write(MAX8660_OVER2, ~0x02, 0x00);				
			}
			else
			{
				ret = max8660_write(MAX8660_OVER2, 0xFF, 0x02);
			}
			break;
		}
		case VCC_V7_RESERVED:
		{			
			if(!enable)
			{
				ret = max8660_write(MAX8660_OVER2, ~0x04, 0x00);
			}
			else
			{
				ret = max8660_write(MAX8660_OVER2, 0xFF, 0x04);
			}
			break;
		}
		default:
		{
			printk(KERN_INFO "max8660: error command\n");
			return -EINVAL;
		}
	}

	return ret;
}

static int is_max8660_voltage_enabled(int cmd, int *enabled)
{
	int ret = 0;
	u8 val = 0;
	
	*enabled = 0;

	if(!init_succ)
	{
		return -1;
	}
	
	switch(cmd)
	{
		case VCC_CIF:
		{
			ret = max8660_read(MAX8660_OVER2, &val);
			if(ret)
			{
				break;
			}
			if(val & 0x02)
			{
				*enabled = 1;
			}
			else
			{
				*enabled = 0;
			}
			break;
		}
		case VCC_V7_RESERVED:
		{
			ret = max8660_read(MAX8660_OVER2, &val);
			if(ret)
			{
				break;
			}
			if(val & 0x04)
			{
				*enabled = 1;
			}
			else
			{
				*enabled = 0;
			}
			break;
		}
		default:
		{
			printk(KERN_INFO "max8660: error command\n");
			return -EINVAL;
		}
	}

	return ret;
}

static int get_max8660_voltage(int cmd, int *pmv)
{
	u8 selector = 0;
	
	*pmv = 0;

	if(!init_succ)
	{
		return -1;
	}
	
	switch (cmd) 
	{
		case VCC_CORE: //V3
		{
#ifdef	CONFIG_DVFM_STDI2C
			selector = the_max8660->shadow_regs[MAX8660_ADTV2];
			*pmv = MAX8660_DCDC_MIN_UV + selector * MAX8660_DCDC_STEP;
#endif
			break;
		}
		case VCC_SRAM: //V4
		{
#ifdef	CONFIG_DVFM_STDI2C
			selector = the_max8660->shadow_regs[MAX8660_SDTV2];
			*pmv = MAX8660_DCDC_MIN_UV + selector * MAX8660_DCDC_STEP;
#endif
			break;
		}
		case VCC_MVT: //V5
		case VCC_PLL:
		case VCC_BG:
		case VCC_OSC13M:
		{
			selector = the_max8660->shadow_regs[MAX8660_MDTV2];
			*pmv = MAX8660_LDO5_MIN_UV + selector * MAX8660_LDO5_STEP;
			break;
		}
		case VCC_CIF: //V6
		{
			u8 val = 0;
			max8660_read(MAX8660_OVER2, &val);
			if(val & 0x02)
			{
				selector = the_max8660->shadow_regs[MAX8660_L12VCR] & 0xf;
				*pmv = MAX8660_LDO67_MIN_UV + selector * MAX8660_LDO67_STEP;
			}
			
			break;
		}
		case VCC_V7_RESERVED: //V7
		{
			u8 val = 0;
			max8660_read(MAX8660_OVER2, &val);
			if(val & 0x04)
			{
				selector = (the_max8660->shadow_regs[MAX8660_L12VCR] >> 4) & 0xf;
				*pmv = MAX8660_LDO67_MIN_UV + selector * MAX8660_LDO67_STEP;
			}
			break;
		}
		case VCC_3V_APPS:
		case VCC_USB:
		case VCC_TSI:
		case VCC_LCD:
		case VCC_CAMERA_IO:
		case VCC_MEM:
		case VCC_SDIO:
		case VCC_CAMERA_ANA:
		case VCC_CARD1:
		case VCC_CARD2:
		{
			break;
		}
		default:
		{
			printk(KERN_INFO "max8660: error command\n");
			return -EINVAL;
		}
	}

	return 0;
}

int set_max8660_voltage(int cmd, int mv)
{
	int ret = 0;
	u8 selector = 0;
	
	if(!init_succ)
	{
		return -1;
	}
	                                                               
	switch (cmd) 
	{
		case VCC_CORE: //V3
		{
#ifdef	CONFIG_DVFM_STDI2C
			if (mv < MAX8660_DCDC_MIN_UV || mv > MAX8660_DCDC_MAX_UV)
				return -EINVAL;

			selector = (mv - (MAX8660_DCDC_MIN_UV - MAX8660_DCDC_STEP + 1)) / MAX8660_DCDC_STEP;
			ret = max8660_write(MAX8660_ADTV2, 0x00, selector);
			if (ret == 0)
			{
				ret = max8660_write(MAX8660_VCC1, 0xff, 0x03);
			}
#endif

			break;
		}
		case VCC_SRAM: //V4
		{
#ifdef	CONFIG_DVFM_STDI2C
			if (mv < MAX8660_DCDC_MIN_UV || mv > MAX8660_DCDC_MAX_UV)
				return -EINVAL;

			selector = (mv - (MAX8660_DCDC_MIN_UV - MAX8660_DCDC_STEP + 1)) / MAX8660_DCDC_STEP;
			ret = max8660_write(MAX8660_SDTV2, 0x00, selector);
			if (ret == 0)
			{
				ret = max8660_write(MAX8660_VCC1, 0xff, 0x30);
			}
#endif

			break;
		}
		case VCC_MVT: //V5
		case VCC_PLL:
		case VCC_BG:
		case VCC_OSC13M:
		{		
			if (mv < MAX8660_LDO5_MIN_UV || mv > MAX8660_LDO5_MAX_UV)
				return -EINVAL;

			selector = (mv - (MAX8660_LDO5_MIN_UV - MAX8660_LDO5_STEP + 1)) / MAX8660_LDO5_STEP;
			ret = max8660_write(MAX8660_MDTV2, 0, selector);
			if (ret == 0)
			{
				ret = max8660_write(MAX8660_VCC1, 0xff, 0xc0);
			}

			break;
		}
		case VCC_CIF: //V6
		{
			if (mv < MAX8660_LDO67_MIN_UV || mv > MAX8660_LDO67_MAX_UV)
				return -EINVAL;

			selector = (mv - (MAX8660_LDO67_MIN_UV - MAX8660_LDO67_STEP + 1)) / MAX8660_LDO67_STEP;
			ret = max8660_write(MAX8660_L12VCR, 0xf0, selector);

			break;
		}
		case VCC_V7_RESERVED: //V7
		{
			if (mv < MAX8660_LDO67_MIN_UV || mv > MAX8660_LDO67_MAX_UV)
				return -EINVAL;

			selector = (mv - (MAX8660_LDO67_MIN_UV - MAX8660_LDO67_STEP + 1)) / MAX8660_LDO67_STEP;
			ret = max8660_write(MAX8660_L12VCR, 0x0f, selector << 4);
			
			break;
		}
		case VCC_3V_APPS:
		case VCC_USB:
		case VCC_TSI:
		case VCC_LCD:
		case VCC_CAMERA_IO:
		case VCC_MEM:
		case VCC_SDIO:
		case VCC_CAMERA_ANA:
		case VCC_CARD1:
		case VCC_CARD2:
		{
			break;
		}
		default:
		{
			printk(KERN_INFO "max8660: error command\n");
			return -EINVAL;
		}
	}

	return ret;
}

static int is_max8660_vbus_assert(void)
{
	return 0;
}

static int is_max8660_avbusvld(void)
{
	return 0;
}

static int is_max8660_asessvld(void)
{
	return 0;
}

static int is_max8660_bsessvld(void)
{
	return 0;
}

static int is_max8660_srp_ready(void)
{
	return 0;
}

static int max8660_set_pump(int enable)
{
	return 0;
}

static int max8660_set_vbus_supply(int enable, int srp)
{
	return 0;
}

static int max8660_set_usbotg_a_mask(void)
{
	return 0;
}

static int max8660_set_usbotg_b_mask(void)
{
	return 0;
}

static struct pmic_ops max8660_pmic_ops = {
        .get_voltage            = get_max8660_voltage,
        .set_voltage            = set_max8660_voltage,
        .enable_voltage		   = enable_max8660_voltage,
        .is_voltage_enabled	   = is_max8660_voltage_enabled,

        .is_vbus_assert         = is_max8660_vbus_assert,
        .is_avbusvld            = is_max8660_avbusvld,
        .is_asessvld            = is_max8660_asessvld,
        .is_bsessvld            = is_max8660_bsessvld,
        .is_srp_ready           = is_max8660_srp_ready,

        .set_pump               = max8660_set_pump,
        .set_vbus_supply        = max8660_set_vbus_supply,
        .set_usbotg_a_mask      = max8660_set_usbotg_a_mask,
        .set_usbotg_b_mask      = max8660_set_usbotg_b_mask,
};

static int max8660_initchip(void)
{
	int ret = 0;
	init_succ = 1;
	
	if(the_max8660 == NULL)
	{
		init_succ = 0;
		return -EINVAL;
	}
	
	memset(&the_max8660->shadow_regs, 0, sizeof(u8) * MAX8660_N_REGS);

#ifdef	CONFIG_DVFM_STDI2C
	ret = max8660_write(MAX8660_OVER1, 0x00, 0x00);
	if(ret)
	{
		init_succ++;
	}
#endif
	ret = max8660_write(MAX8660_OVER2, 0x00, 0x00);
	if(ret)
	{
		init_succ++;
	}
#ifdef	CONFIG_DVFM_STDI2C
	ret = max8660_write(MAX8660_ADTV1, 0x00, 0x1B);
	if(ret)
	{
		init_succ++;
	}
	ret = max8660_write(MAX8660_ADTV2, 0x00, 0x1A);
	if(ret)
	{
		init_succ++;
	}
	ret = max8660_write(MAX8660_SDTV1, 0x00, 0x1B);
	if(ret)
	{
		init_succ++;
	}
	ret = max8660_write(MAX8660_SDTV2, 0x00, 0x1B);
	if(ret)
	{
		init_succ++;
	}
#endif
	ret = max8660_write(MAX8660_MDTV1, 0x00, 0x04);
	if(ret)
	{
		init_succ++;
	}
	ret = max8660_write(MAX8660_MDTV2, 0x00, 0x04);
	if(ret)
	{
		init_succ++;
	}
#ifdef	CONFIG_DVFM_STDI2C
	ret = max8660_write(MAX8660_VCC1,  0x00, 0xFF);
	if(ret)
	{
		init_succ++;
	}
#else
	ret = max8660_write(MAX8660_VCC1, 0xff, 0xC0);
	if(ret)
	{
		init_succ++;
	}
#endif
	ret = max8660_write(MAX8660_L12VCR,0x00, 0xFA);
	if(ret)
	{
		init_succ++;
	}
	ret = max8660_write(MAX8660_FPWM,  0x00, 0x00);
	if(ret)
	{
		init_succ++;
	}

	if(init_succ > 1)
	{
		init_succ = 0;
	}
	else
	{
		init_succ = 1;
	}
	
	return init_succ - 1;
}

#ifdef CONFIG_PROC_FS
#define MAX8660_PROC_FILE "driver/max8660"
static struct proc_dir_entry *max8660_proc_file;
 
static int max8660_seq_show(struct seq_file *s, void *p)
{
	u8 val = 0;

	pxa3xx_pmic_read(MAX8660_OVER1, &val);
	seq_printf(s, "MAX8660_OVER1: 0x%02x\n", val);

	pxa3xx_pmic_read(MAX8660_OVER2, &val);
	seq_printf(s, "MAX8660_OVER2: 0x%02x\n", val);

	pxa3xx_pmic_read(MAX8660_ADTV1, &val);
	seq_printf(s, "MAX8660_ADTV1: 0x%02x\n", val);

	pxa3xx_pmic_read(MAX8660_ADTV2, &val);
	seq_printf(s, "MAX8660_ADTV2: 0x%02x\n", val);

	pxa3xx_pmic_read(MAX8660_SDTV1, &val);
	seq_printf(s, "MAX8660_SDTV1: 0x%02x\n",val);

	pxa3xx_pmic_read(MAX8660_SDTV2, &val);
	seq_printf(s, "MAX8660_SDTV2: 0x%02x\n",val);

	pxa3xx_pmic_read(MAX8660_MDTV1, &val);
	seq_printf(s, "MAX8660_MDTV1: 0x%02x\n",val);

	pxa3xx_pmic_read(MAX8660_MDTV2, &val);
	seq_printf(s, "MAX8660_MDTV2: 0x%02x\n",val);

	pxa3xx_pmic_read(MAX8660_L12VCR, &val);
	seq_printf(s, "MAX8660_L12VCR:0x%02x\n",val);
	
	pxa3xx_pmic_read(MAX8660_FPWM, &val);
	seq_printf(s, "MAX8660_FPWM:  0x%02x\n",val);

        return 0;
}

static int max8660_seq_open(struct inode *inode, struct file *file)
{
        return single_open(file, &max8660_seq_show, NULL);
}

static struct file_operations max8660_seq_fops = {
        .owner          = THIS_MODULE,
        .open           = max8660_seq_open,
        .read           = seq_read,
        .llseek         = seq_lseek,
        .release        = single_release,
};

static int create_max8660_proc_file(void)
{
        max8660_proc_file = create_proc_entry(MAX8660_PROC_FILE, 0644, NULL);
        if(!max8660_proc_file)
        {
                printk(KERN_INFO "Create proc file for max8660 failed\n");
                return -ENOMEM;
        }
        max8660_proc_file->proc_fops = &max8660_seq_fops;
        
        return 0;
}

static void remove_max8660_proc_file(void)
{
        remove_proc_entry(MAX8660_PROC_FILE, &proc_root);
}
#endif

/******************************************************************************
 *                                                                            *
 *                              max8660 Driver                                *
 *                                                                            *
 ******************************************************************************/

static int max8660_probe(struct platform_device *_dev)
{
	int ret = 0;

	ret = max8660_initchip();
	if(ret != 0)
	{
		printk(KERN_INFO "Initialize max8660 failed\n");
		return -1;
	}
         
        pmic_set_ops(&max8660_pmic_ops);
         
#ifdef CONFIG_PROC_FS
	ret = create_max8660_proc_file();
	if(ret)
	{
		printk(KERN_INFO "%s: Failed to create max8660 proc file.\n", __func__);
		return ret;
	}
#endif
	
	return 0;
}

static int max8660_remove(struct platform_device *_dev)
{	
#ifdef  CONFIG_PROC_FS
         remove_max8660_proc_file();
#endif
         pmic_set_ops(NULL);
         return 0;
}

#ifdef CONFIG_PM
static int max8660_suspend(struct platform_device *pdev, pm_message_t state)
{	
	printk(KERN_INFO "max8660_suspend\n");

#ifdef	CONFIG_DVFM_STDI2C	
	max8660_write(MAX8660_OVER1, 0xFF, 0x04);
#endif

        return 0;
}

static int max8660_resume(struct platform_device *pdev)
{	
	printk(KERN_INFO "max8660_resume\n");

#ifdef	CONFIG_DVFM_STDI2C
	max8660_write(MAX8660_OVER1, ~0x04, 0x00);
#endif

        return 0;
}
#else
#define max8660_suspend           NULL
#define max8660_resume            NULL
#endif

static struct platform_driver max8660_driver = {
         .driver = {
                 .name   = "pxa3xx_pmic",
         },
         .probe          = max8660_probe,
         .remove         = max8660_remove,
#ifdef CONFIG_PM
         .suspend        = max8660_suspend,
         .resume         = max8660_resume,
#endif
};

/******************************************************************************
 *                                                                            *
 *                              max8660 I2C Client Driver                     *
 *                                                                            *
 ******************************************************************************/

static int i2c_max8660_attach_adapter(struct i2c_adapter *adapter);
static int i2c_max8660_detach_client(struct i2c_client *client);

struct i2c_driver i2c_max8660_driver  =
{
        .driver = {
                .name   = "max8660 i2c client driver",
        },
        .id             = 0,
        .attach_adapter = &i2c_max8660_attach_adapter,
        .detach_client  = &i2c_max8660_detach_client,
};

static unsigned short normal_i2c[] = { MAX8660_ADDRESS, I2C_CLIENT_END };
I2C_CLIENT_INSMOD_1(max8660);

static int i2c_max8660_detect_client(struct i2c_adapter *adapter, int address, int kind)
{
	int ret = 0;
	struct i2c_client *new_client;
	struct max8660 *max8660;
	
	if(the_max8660)
	{
		printk(KERN_INFO "max8660: only one max8660 for now\n");
		return -ENODEV;
	}
        
	connect_si2c();
	if(!i2c_check_functionality(adapter,I2C_FUNC_SMBUS_WRITE_BYTE_DATA))
	{
		printk(KERN_INFO "max8660: write byte op is not permited.\n");
		disconnect_si2c();
		goto ERROR0;
	}
	disconnect_si2c();
	
	new_client = kzalloc(sizeof(struct i2c_client), GFP_KERNEL );
	if (!new_client)
	{
		printk(KERN_INFO "max8660: insufficient memory\n");
		ret = -ENOMEM;
		goto ERROR0;
	}
	new_client->addr = address;
	new_client->adapter = adapter;
	new_client->driver = &i2c_max8660_driver;
	new_client->flags = 0;

	max8660 = kzalloc(sizeof(struct max8660), GFP_KERNEL);
	if(max8660 == NULL)
	{
		printk(KERN_INFO "max8660: insufficient memory\n");
		ret = -ENOMEM;
                goto ERROR1;
	}
	strcpy(new_client->name, "max8660");
	max8660->client = new_client;
	the_max8660 = max8660;
	
	if((ret = i2c_attach_client(the_max8660->client)))
		goto ERROR2;
        
	i2c_set_clientdata(the_max8660->client, the_max8660);
	
	printk(KERN_INFO "i2c: max8660 detected\n");

	return 0;

ERROR2:
	kfree(the_max8660);
	the_max8660 = NULL;
ERROR1:
        kfree(new_client);
ERROR0:
        return ret;
}

static int i2c_max8660_attach_adapter(struct i2c_adapter *adap)
{	
	return i2c_probe(adap,&addr_data, &i2c_max8660_detect_client);
}

static int i2c_max8660_detach_client(struct i2c_client *client)
{
	int ret;
	struct max8660 *max8660;
	
	max8660 = i2c_get_clientdata(client);
	kfree(max8660);
	i2c_set_clientdata(client, NULL);
	the_max8660 = NULL;

	if ((ret = i2c_detach_client(client)))
	{
		printk(KERN_WARNING "max8660: client deregistration failed, client not detached.\n");
		return ret;
	}
	kfree(client);
	return 0;
}

static int __init max8660_init(void)
{
	int ret;

        if((ret = i2c_add_driver(&i2c_max8660_driver)))
        {
                printk(KERN_WARNING "max8660: driver registration failed, module not inserted.\n");
                return ret;
        }
	
        ret = platform_driver_register(&max8660_driver);
        return ret;
}

static void __exit max8660_exit(void)
{
        platform_driver_unregister(&max8660_driver);

        if(i2c_del_driver(&i2c_max8660_driver))
        {
                printk(KERN_WARNING "max8660: driver deregistration failed, module not removed.\n");
        }
}

module_init(max8660_init);
module_exit(max8660_exit);

MODULE_DESCRIPTION("MAX8660 Driver");
MODULE_LICENSE("GPL");
