/*
 * Wacom serial digitizer driver
 *
 * Copyright (c) 2010  breeze_yin<breeze_yin@asus.com>
 *
 * Based on MicroTouch driver (drivers/input/touchscreen/mtouch.c)
 * Copyright (c) 2010 Breeze Yin
 */

/*
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation.
 */

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/serio.h>
#include <linux/init.h>
#include <linux/socket.h>
#include <linux/skbuff.h>
#include <linux/netlink.h>
#include <net/sock.h>
#include <linux/wait.h>
#include <linux/freezer.h>
#include <linux/kthread.h>
#include <asm/atomic.h>
#include <linux/cdev.h>
#include "../../usb/gadget/wacom_tablet.h"

#ifdef CONFIG_IPM
#include <asm/arch/ipmc.h>
#endif

//#define DEBUG

#ifdef DEBUG
#define DBG(fmt, args...) \
	printk(fmt, ## args)
#define DBGMSG	\
	printk("FILE: %s, FUN: %s, LINE: %d\n", __FILE__, __FUNCTION__, __LINE__);
#else
#define DBG(fmt, args...) \
	do { } while (0)
#define DBGMSG  \
	do { } while (0);
#endif /* DEBUG */

#define DRIVER_DESC	"Wacom serial digitizer driver"

MODULE_AUTHOR("Breeze Yin <breeze_yin@asus.com>");
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

static const char shortname [] = "wac_tch";
static dev_t wacom_touch_devno;
static struct class *wacom_touch_class;

/*
 * Definitions & global arrays.
 */
#define PRODUCT_ID	0x6583

#define WAC_MAX_LENGTH 	11 //or 9 ?
#define WAC_LENGTH	9

#define WAC_MIN_XC 	0
#define WAC_MAX_XC 	0x4060
#define WAC_MIN_YC 	0
#define WAC_MAX_YC 	0x307A
#define WAC_MIN_PRESS 	0
#define WAC_MAX_PRESS 	0xFF

/* device IDs */
#define STYLUS_DEVICE_ID	0x02
#define TOUCH_DEVICE_ID		0x03
#define CURSOR_DEVICE_ID	0x06
#define ERASER_DEVICE_ID	0x0A
#define PAD_DEVICE_ID		0x0F

#define WACOM_TOUCH_NOT_ERROR	0x08
#define WACOM_TOUCH_SELECTED	0x10
#define WACOM_TOUCH_PAPER_EMPTY	0x20

int count = 0;
static int bTap = 0;

/*
 * Per-touchscreen data.
 */

struct wac {
	spinlock_t		lock;		/* lock this structure */
	struct input_dev *dev;
	struct serio *serio;
	struct wacom_tablet_dev *gadget_dev; 
	int idx;
	unsigned char data[WAC_MAX_LENGTH];
	unsigned char report[WAC_MAX_LENGTH];
	char phys[32];
	int  tool[1];
	int  id[1];
	bool bconnected;
	struct task_struct *capsensor_task;
	wait_queue_head_t	wait;
	bool  bCapStatus; /*1--> open , 0-->close*/
	struct cdev		wacom_touch_cdev;
	struct device	*pdev;
	long   wacom_touch_cdev_open;
	long   wacom_touch_status;
};

static struct wac g_wac_dev;


/*-------------------------------------------------------------------------*/

static int
wacom_touch_open(struct inode *inode, struct file *fd)
{
	struct wac				*dev;
	unsigned long			flags;
	int						ret = -EBUSY;

	lock_kernel();
	
	DBGMSG
	
	dev = container_of(inode->i_cdev, struct wac, wacom_touch_cdev);

	spin_lock_irqsave(&dev->lock, flags);

	if (!dev->wacom_touch_cdev_open) {
		dev->wacom_touch_cdev_open = 1;
		fd->private_data = dev;
		ret = 0;
		/* Change the wacom touch status to show that it's on-line. */
		dev->wacom_touch_status |= WACOM_TOUCH_SELECTED;
	}

	spin_unlock_irqrestore(&dev->lock, flags);

	DBG("wacom_touch_open returned %x\n", ret);
	
	unlock_kernel();
	
	return ret;
}

static int
wacom_touch_close(struct inode *inode, struct file *fd)
{
	struct wac	*dev = fd->private_data;
	unsigned long		flags;
	
	DBGMSG
	
	spin_lock_irqsave(&dev->lock, flags);
	
	dev->wacom_touch_cdev_open = 0;
	fd->private_data = NULL;
	/* Change wacom touch status to show that the wacom touch is off-line. */
	dev->wacom_touch_status &= ~WACOM_TOUCH_SELECTED;
	
	spin_unlock_irqrestore(&dev->lock, flags);
  	
	return 0;
}

static int
wacom_touch_write(struct file *fd, const char __user *buf, size_t len, loff_t *ptr)
{
	int status = 0;
	struct wac	*dev = fd->private_data;
	unsigned long		flags;
	unsigned char 		k_buf[2];
	
	DBGMSG
	
	if(!dev->bconnected)
	{
		printk("The wacom touch is not connected\n");
		return -EINVAL;
	}
	
	if(!dev->wacom_touch_cdev_open || !(dev->wacom_touch_status & WACOM_TOUCH_SELECTED))
	{
		printk("The device is closed\n");
		return -EINVAL;
	}
	
	if(len > 2)
		len = 2;
	
	spin_lock_irqsave(&dev->lock, flags);
	
	if(copy_from_user(k_buf, buf, len))
	{
		printk("copy_from_user failed.\n");
		spin_unlock_irqrestore(&dev->lock, flags);
		return -EAGAIN;
	}
	
	k_buf[1] = 0;
	
	serio_pause_rx(dev->serio);
	
	status = dev->serio->write(dev->serio, k_buf[0]);
	
	serio_continue_rx(dev->serio);
	
	spin_unlock_irqrestore(&dev->lock, flags);
	
	return status;
}

static struct file_operations wacom_touch_ops = {
	.owner =	THIS_MODULE,
	.open =		wacom_touch_open,
	.write =	wacom_touch_write,
	.release =	wacom_touch_close
};

/*-------------------------------------------------------------------------*/


static int wac_parsedata(struct wac *wac, const unsigned char* puchData, unsigned int uLength)
{
	int x=0, y=0, proxi=0, press=0, button=0, stylus=0, stylus2=0;
	int flags;
		
	struct input_dev *dev = wac->dev ;

	/* Tablet PC Supports: 256 pressure, eraser, 1/2 side-switch */

	if (uLength != 9) 
		return 1; 

	proxi = puchData[0] & 0x20 ? 1 : 0; //pen is proximity
	if (likely(proxi))
	{
		//eraser or pen ?
		if(!wac->id[0])
		{
			wac->tool[0] = (puchData[0] & 0x04) ? BTN_TOOL_RUBBER : BTN_TOOL_PEN;
			wac->id[0] = (wac->tool[0] == BTN_TOOL_RUBBER) ? ERASER_DEVICE_ID : STYLUS_DEVICE_ID ;
		}

		button = (puchData[0] & 0x01) ? 1 : 0;

		/* pen has side-switch(es), eraser has none */
		if(wac->tool[0] == BTN_TOOL_PEN)
		{
			stylus  = (puchData[0] & 0x02) ? 1 : 0;
			//stylus2 = (puchData[0] & 0x04) ? 1 : 0;
		}

		press = (((puchData[6] & 0x01) << 7) | (puchData[5] & 0x7F));

		x = (((int)puchData[6] & 0x60) >> 5) |
			((int)puchData[2] << 2) |
			((int)puchData[1] << 9);

		y = (((int)puchData[6] & 0x18) >> 3) |
			((int)puchData[4] << 2) |
			((int)puchData[3] << 9);
	}
	else
	{
		wac->id[0] = 0;
	}
	
	if(wac->gadget_dev)
	{
		//default right-handler mode
		int xtmp, ytmp;
		xtmp = WAC_MAX_XC - x;
		ytmp = WAC_MAX_YC - y;
		wac->report[1] =  puchData[0] &  0xff;
		wac->report[2] =  xtmp & 0xff ;
		wac->report[3] =  (xtmp >> 8) & 0xff ;
		wac->report[4] =  ytmp & 0xff;
		wac->report[5] =  (ytmp >> 8) & 0xff;
		wac->report[6] =  press & 0xff;
		wac->report[7] =  (press >> 8) & 0xff;
		wac->gadget_dev->start_tx(wac->gadget_dev, wac->report);
	}
	else
	{	
		//int touch = (button ? 1 : (press ? 1 : (stylus ? 1 : 0)));
		
		input_report_key(dev, BTN_TOUCH, button);   //left button
		input_report_key(dev, BTN_STYLUS, stylus);  //right buton
		//input_report_key(dev, BTN_STYLUS2, stylus2);//middle button

		input_report_abs(dev, ABS_X, x);
		input_report_abs(dev, ABS_Y, y);
		input_report_abs(dev, ABS_PRESSURE, press);

		input_report_key(dev, wac->tool[0], proxi);
		input_report_abs(dev, ABS_MISC, wac->id[0]);

		input_sync(dev);
		
		//if((touch && !bTap) || (!touch && bTap))
		//{
		//	printk("wakeup wait\n");	
		//	bTap = !bTap;
		//	wake_up_interruptible(&wac->wait);
		//}
		
	}

	return  !proxi;
}

extern int pm_updatetimer(int);
extern int bAllow_input;
static irqreturn_t wac_interrupt(struct serio *serio,
		unsigned char data, unsigned int flags)
{
	if(bAllow_input)
	{
		bAllow_input = 0;
		pm_updatetimer(IPM_COUNT_LIMIT_TIMER);
		/* Need to post event to policy maker.  */
		ipm_event_notify(IPM_EVENT_UI, IPM_EVENT_DEVICE_OUTD0CS, NULL, 0);//eric_zhu,0524
	}
	
	struct wac *wac = serio_get_drvdata(serio);
	//struct input_dev *dev = wac->dev;

	wac->data[wac->idx] = data;
	
	if(likely((wac->data[0] & 0xC0) == 0x80))
	{
		if(++wac->idx == WAC_LENGTH)
		{
			wac_parsedata(wac, wac->data, WAC_LENGTH);
			wac->idx = 0;
			
			if(((count++)%8) == 0){
				wake_up_interruptible(&wac->wait);
				count = 0 ;
			}
		}
	}

	return IRQ_HANDLED;
}

/*
static int wac_thread(void *dev)
{	
	struct wac *wac = (struct wac *)dev;
	
	printk("wac_thread: enter\n");
	while(!kthread_should_stop()) {
		
		printk("wac_thread: wait\n");
		wait_event_interruptible(wac->wait, !wac->bconnected);
		printk("wac_thread: wakeup\n");
		if(!bTap)
		{
			if(wac->bCapStatus == 1)
			{
				if(get_senflag() == 0)
				{
					//enable capsensor
					printk("wac_thread: enable capsensor\n");
					set_Sensorboard(1);
				}
				
				wac->bCapStatus = 0;
			}
		}
		else
		{
			if(get_senflag() == 1)
			{
				wac->bCapStatus = 1;
				//disable capsensor
				printk("wac_thread: disable capsensor\n");
				set_Sensorboard(0);
			}
		}
		printk("wac_thread: freeze?\n");
		try_to_freeze();
	} 

	printk(KERN_DEBUG "wacom digitizer: kWacThread exiting\n");
	return 0;
}
*/

static int wac_thread(void *dev)
{	
	struct wac *wac = (struct wac *)dev;
	
	while(!kthread_should_stop()) {
		
		long timeout = interruptible_sleep_on_timeout(&wac->wait, HZ/2);
		//printk("wac_thread: wakeup\n");
		if(timeout == 0)
		{
			if(wac->bCapStatus == 1)
			{
				//enable capsensor
				if(get_senflag() == 0)
				{
					//printk("wac_thread: enable capsensor\n");
					wac->bCapStatus = 0;
					set_Sensorboard(1);
				}
			}
		}
		else
		{
			if(get_senflag() == 1)
			{
				wac->bCapStatus = 1;
				//printk("wac_thread: disable capsensor\n");
				//disable capsensor
				set_Sensorboard(0);
			}
		}
		
		try_to_freeze();
	} 

	printk(KERN_DEBUG "wacom digitizer: kWacThread exiting\n");
	return 0;
}


static int wac_set_gadget_dev(struct wacom_tablet_dev* wacom_tablet)
{
	int flags;
	
	DBGMSG
	
	if(wacom_tablet != NULL)
	{
		if(g_wac_dev.bconnected == 1)
		{
			serio_pause_rx(g_wac_dev.serio);
			g_wac_dev.gadget_dev = wacom_tablet ;
			serio_continue_rx(g_wac_dev.serio);
		}
		else
		{
			printk("wacom digitizer: touchwacom isn't connected\n");
			return -EAGAIN;
		}
	}
	else
	{
		serio_pause_rx(g_wac_dev.serio);
		g_wac_dev.gadget_dev = wacom_tablet ;
		serio_continue_rx(g_wac_dev.serio);
	}
	
	return 0;
}

EXPORT_SYMBOL(wac_set_gadget_dev);

/*
 * wac_disconnect() is the opposite of wac_connect()
 */

static void wac_disconnect(struct serio *serio)
{
	struct wac *wac = serio_get_drvdata(serio);
	
	DBGMSG
	wac->bconnected = 0;
	
	/* Remove sysfs files */
	device_destroy(wacom_touch_class, wacom_touch_devno);

	/* Remove Character Device */
	cdev_del(&wac->wacom_touch_cdev);
	
	kthread_stop(wac->capsensor_task);
	
	input_get_device(wac->dev);
	input_unregister_device(wac->dev);
	serio_close(serio);
	serio_set_drvdata(serio, NULL);
	input_put_device(wac->dev);
	//kfree(wac);
	wac->dev = NULL;
	wac->gadget_dev = NULL;
}

/*
 * wac_connect() is the routine that is called when someone adds a
 * new serio device that supports the Touchwacom protocol and registers it as
 * an input device.
 */

static int wac_connect(struct serio *serio, struct serio_driver *drv)
{
	struct wac *wac;
	struct input_dev *input_dev;
	int err;
	
	DBGMSG
	//wac = kzalloc(sizeof(struct wac), GFP_KERNEL);
	wac = &g_wac_dev;
	wac->gadget_dev = NULL;
	
	DBG("init spin lock\n");
	spin_lock_init(&wac->lock);
	
	input_dev = input_allocate_device();
	if (!wac || !input_dev) {
		err = -ENOMEM;
		goto fail1;
	}

	wac->serio = serio;
	wac->dev = input_dev;
	snprintf(wac->phys, sizeof(wac->phys), "%s/input0", serio->phys);

	input_dev->private = wac;
	input_dev->name = "Wacom serial digitizer";
	input_dev->phys = wac->phys;
	input_dev->id.bustype = BUS_RS232;
	input_dev->id.vendor = SERIO_WACOM;
	input_dev->id.product = PRODUCT_ID;
	input_dev->id.version = 0x0100;
	input_dev->evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
	__set_bit(BTN_TOUCH, input_dev->keybit);
	__set_bit(BTN_STYLUS, input_dev->keybit);
	//__set_bit(BTN_STYLUS2, input_dev->keybit);
	__set_bit(BTN_TOOL_PEN, input_dev->keybit);
	__set_bit(BTN_TOOL_RUBBER, input_dev->keybit);

	input_set_abs_params(input_dev, ABS_X, WAC_MIN_XC, WAC_MAX_XC, 0, 0);
	input_set_abs_params(input_dev, ABS_Y, WAC_MIN_YC, WAC_MAX_YC, 0, 0);
	input_set_abs_params(input_dev, ABS_PRESSURE, WAC_MIN_PRESS, WAC_MAX_PRESS, 0, 0);

	__set_bit(ABS_MISC, input_dev->absbit);

	serio_set_drvdata(serio, wac);

	err = serio_open(serio, drv);
	if (err)
		goto fail2;

	err = input_register_device(wac->dev);
	if (err)
		goto fail3;
	
	init_waitqueue_head(&wac->wait);
	
	/*
	 * Register a character device as an interface to a user mode
	 * program that handles the wacom_touch specific functionality.
	 */
	cdev_init(&wac->wacom_touch_cdev, &wacom_touch_ops);
	
	wac->wacom_touch_cdev.owner = THIS_MODULE;
	
	err = cdev_add(&wac->wacom_touch_cdev, wacom_touch_devno, 1);
	if (err) {
		printk("Failed to add char device\n");
		/* Remove sysfs files */
		device_destroy(wacom_touch_class, wacom_touch_devno);
		/* Remove Character Device */
		cdev_del(&wac->wacom_touch_cdev);
		goto fail3;
	}
	
	/* Setup the sysfs files for the wacom_touch device. */
	wac->pdev = device_create(wacom_touch_class, NULL, wacom_touch_devno, shortname);  
	if (IS_ERR(wac->pdev)) {
		err = PTR_ERR(wac->pdev);
		printk("Failed to create wacom touch device\n");
		err = -ENOMEM;
		goto fail3;
	}
	
	printk("create char dev %s success\n", shortname);
	
	wac->wacom_touch_cdev_open = 0;
	wac->wacom_touch_status = WACOM_TOUCH_NOT_ERROR;
	wac->bconnected = 1;
	
	//get capsensor status
	wac->bCapStatus = 0;//get_senflag();
	
	/*init process thread*/
	wac->capsensor_task = kthread_run(wac_thread, (void*)wac, "kWacThread");
	if (IS_ERR(wac->capsensor_task)) {
		err = PTR_ERR(wac->capsensor_task);
		printk(KERN_ERR "wacom digitizer: Failed to start wacom thread, error: %d\n", err);
		goto fail3;
	}
	
	return 0;

 fail3:	serio_close(serio);
 fail2:	serio_set_drvdata(serio, NULL);
 fail1:	input_free_device(input_dev);
	//kfree(wac);
	wac->dev = NULL;
	return err;
}


static int  wac_reconnect(struct serio *serio)
{
	DBGMSG
	return 0;
}

/*
 * The serio driver structure.
 */

static struct serio_device_id wac_serio_ids[] = {
	{
		.type	= SERIO_RS232,
		.proto	= SERIO_WACOM,
		.id	= SERIO_ANY,
		.extra	= SERIO_ANY,
	},
	{ 0 }
};

MODULE_DEVICE_TABLE(serio, wac_serio_ids);

static struct serio_driver wac_drv = {
	.driver		= {
		.name	= "touchwacom", //"touchwacom",
		//.owner = THIS_MODULE,
	},
	.description	= DRIVER_DESC,
	.id_table	= wac_serio_ids,
	.interrupt	= wac_interrupt,
	.connect	= wac_connect,
	.disconnect	= wac_disconnect,
	.reconnect  = wac_reconnect,
};



/*
 * The functions for inserting/removing us as a module.
 */

static int __init wac_init(void)
{
	int status;
	
	DBGMSG

	status = alloc_chrdev_region(&wacom_touch_devno, 0, 1,
			shortname);
	if (status) {
		DBG("alloc_chrdev_region %d\n", status);
		//class_destroy(wacom_touch_class);
		return status;
	}
	
	wacom_touch_class = class_create(THIS_MODULE, shortname);
	if (IS_ERR(wacom_touch_class)) {
		status = PTR_ERR(wacom_touch_class);
		DBG("unable to create wacom touch class %d\n", status);
		unregister_chrdev_region(wacom_touch_devno, 1);
		return status;
	}
	
	//printk("the wacom touch devno-->%d\n", (u32)wacom_touch_devno);
		
	status = serio_register_driver(&wac_drv);
	if (status) {
		unregister_chrdev_region(wacom_touch_devno, 1);
		class_destroy(wacom_touch_class);
		DBG("serio_register_driver %x\n", status);
	}


	return status;
}

static void __exit wac_exit(void)
{
	DBGMSG
	
	unregister_chrdev_region(wacom_touch_devno, 1);
	class_destroy(wacom_touch_class);
	
	serio_unregister_driver(&wac_drv);
}

module_init(wac_init);
module_exit(wac_exit);
