/** @file if_sdio.c
 *  @brief This file contains SDIO IF (interface) module
 *  related functions.
 * 
 * (c) Copyright  2003-2006, Marvell International Ltd. 
 *
 * This software file (the "File") is distributed by Marvell International 
 * Ltd. under the terms of the GNU General Public License Version 2, June 1991 
 * (the "License").  You may use, redistribute and/or modify this File in 
 * accordance with the terms and conditions of the License, a copy of which 
 * is available along with the File in the gpl.txt file or by writing to 
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 
 * 02111-1307 or on the worldwide web at http://www.gnu.org/licenses/gpl.txt.
 *
 * THE FILE IS DISTRIBUTED AS-IS, WITHOUT WARRANTY OF ANY KIND, AND THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE 
 * ARE EXPRESSLY DISCLAIMED.  The License provides additional details about 
 * this warranty disclaimer.
 *
 */
/****************************************************
Change log:
	10/14/05: add Doxygen format comments 
****************************************************/

#include	"if_sdio.h"

#include	"sd8385.h"

#ifdef HELPER_IMAGE
#include	"helper.h"
#endif


#define SD_BLOCK_SIZE		32

#define SD_BLOCK_SIZE_512	512
#define SD_BLOCK_SIZE_320	320
#define SDIO_HEADER_LEN		4

#ifdef FW_WAKEUP_METHOD
#define GPIO_PORT_NUM		
#define GPIO_PORT_INIT()	
#define GPIO_PORT_TO_HIGH()	
#define GPIO_PORT_TO_LOW()	
#endif /* FW_WAKEUP_METHOD */

/********************************************************
		Local Variables
********************************************************/

static wlan_private *pwlanpriv;
static wlan_private *(* wlan_add_callback)(void *dev_id);
static int (* wlan_remove_callback)(void *dev_id);

/********************************************************
		Global Variables
********************************************************/

#ifdef FW_WAKEUP_TIME
extern u32 wt_pwrup_sending,wt_pwrup_sent,wt_int;
#endif

static int sdio_rw(struct mss_card *card, u8 func, u32 reg, u32 blksz, u32 nob, int opcode, u32 data, int rw)
{
	struct mss_request mreq;
	struct mss_rw_arg marg;
	struct mss_rw_result mres;
	struct scatterlist sg;
	int ret;

	memset(&mreq, 0x0, sizeof(mreq));	
	memset(&marg, 0x0, sizeof(marg));	
	memset(&mres, 0x0, sizeof(mres));	

	mreq.arg = &marg;
	mreq.result = &mres;
	mreq.card = card;

	mreq.action = rw ? MSS_WRITE_MEM: MSS_READ_MEM;
	marg.nob = nob;
	marg.func = func;
	marg.opcode = opcode;
	marg.block_len = blksz;
	marg.block = reg;
	if (rw)
		marg.val = data;

	/* Read/Write one register */
	if (blksz == 0 && nob == 1) {
		marg.sg = NULL;
		marg.sg_len = 0;
	}
	else {
		sg.page = virt_to_page((u8 *)data);
		sg.offset = offset_in_page((u8 *)data);
		sg.length = blksz * nob;
		marg.sg = &sg;
		marg.sg_len = 1;
	}

	ret = mss_send_request(&mreq);
	if (ret)
		return ret;
	if (blksz == 0 && nob == 1 && !rw) 
		*(u8 *)data = (u8)(mres.bytes_xfered);

	return 0;
}

static inline int sdio_read_ioreg(struct mss_card *card, u8 func, u32 reg, u8 *data)
{
	return sdio_rw(card, func, reg, 0, 1, 0, (u32)data, 0);
}

static inline int sdio_write_ioreg(struct mss_card *card, u8 func, u32 reg, u8 data)
{
	return sdio_rw(card, func, reg, 0, 1, 0, data, 1);
}

static inline int sdio_read_iomem(struct mss_card *card, u8 func, u32 reg, int opcode, u32 nob, u32 blksz, u8 *buf)
{
	return sdio_rw(card, func, reg, blksz, nob, opcode, (u32)buf, 0);
}

static inline int sdio_write_iomem(struct mss_card *card, u8 func, u32 reg, int opcode, u32 nob, u32 blksz, u8 *buf)
{
	PRINTM(INFO, "iomem: func:%d, reg:%d, opcode:%d, nob:%d, blksz:%d\n", func, reg, opcode, nob, blksz);
	return sdio_rw(card, func, reg, blksz, nob, opcode, (u32)buf, 1);
}		
/********************************************************
		Local Functions
********************************************************/

/** 
 *  @brief This function reads scratch registers
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @param dat	   A pointer to keep returned data
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
static int mv_sdio_read_scratch(wlan_private *priv, u16 *dat)
{
	int	ret = WLAN_STATUS_SUCCESS;
	u8	scr0;
	u8	scr1;
	u8	scr2;
	
	ret = sdio_read_ioreg(priv->wlan_dev.card, FN0, SCRATCH_0_REG, &scr0);

	if (ret < 0)
		return WLAN_STATUS_FAILURE;

	PRINTM(INFO, "SCRATCH_0_REG = %x\n",scr0);
	
	ret = sdio_read_ioreg(priv->wlan_dev.card, FN0, SCRATCH_1_REG, &scr1);

	if (ret < 0)
		return WLAN_STATUS_FAILURE;

	PRINTM(INFO, "SCRATCH_1_REG = %x\n",scr1);

	/* the SCRATCH_REG @ 0x8fc tells the cause for the Mac Event */
	
	ret = sdio_read_ioreg(priv->wlan_dev.card, FN0, 0x80fc, &scr2);
	
	if (ret < 0) { 
		PRINTM(WARN, "Unable to read \n");
		return WLAN_STATUS_FAILURE;
	}

	priv->adapter->EventCause = scr2;
	*dat = (((u16) scr1) << 8) | scr0;
	
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief This function polls the card status register.
 *  
 *  @param priv    	A pointer to wlan_private structure
 *  @param bits    	the bit mask
 *  @return 	   	WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
static int mv_sdio_poll_card_status(wlan_private *priv, u8 bits)
{
	int	tries;
	int	rval;
	u8	cs;

	for (tries = 0; tries < MAX_POLL_TRIES; tries++) {
		rval = sdio_read_ioreg(priv->wlan_dev.card, FN1,
					CARD_STATUS_REG, &cs);
		PRINTM(INFO, "read result:%d, cs:%d, bits:%d\n", rval, cs, bits);
		if (rval == 0 && (cs & bits) == bits) {
			return WLAN_STATUS_SUCCESS;
		}

		udelay(100);  
	}

	PRINTM(WARN, "mv_sdio_poll_card_status: FAILED!\n");
	return WLAN_STATUS_FAILURE;
}

/** 
 *  @brief This function programs the firmware image.
 *  
 *  @param priv    	A pointer to wlan_private structure
 *  @param firmware 	A pointer to the buffer of firmware image
 *  @param firmwarelen 	the length of firmware image
 *  @return 	   	WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
static int sbi_prog_firmware_image(wlan_private *priv,
				const u8 *firmware, int firmwarelen)
{
	int	ret = WLAN_STATUS_SUCCESS;
	u16	firmwarestat;
	u64	buf[(FIRMWARE_TRANSFER_NBLOCK * SD_BLOCK_SIZE / 8) + 1];
	u8	*fwbuf;
	int	fwblknow;
	u32	tx_len;
#ifdef FW_DOWNLOAD_SPEED
	u32	tv1,tv2;
#endif

	ENTER();

	fwbuf = (u8 *)buf;
	PRINTM(INFO, "buf addr :0x%x\n", fwbuf);
	/* Check if the firmware is already downloaded */
	if ((ret = mv_sdio_read_scratch(priv, &firmwarestat)) < 0) {
		PRINTM(INFO, "read scratch returned <0\n");
		goto done;
	}

	if (firmwarestat == FIRMWARE_READY) {
		PRINTM(INFO, "Firmware already downloaded!\n");
		ret = WLAN_STATUS_SUCCESS;
		goto done;
  	}

#ifdef HELPER_IMAGE
	PRINTM(INFO, "Downloading helper image (%d bytes), block size %d bytes "
		,firmwarelen, SD_BLOCK_SIZE);
#else
	PRINTM(INFO, "Downloading firmware image (%d bytes), block size %d bytes "
		,firmwarelen, SD_BLOCK_SIZE);
#endif

#ifdef FW_DOWNLOAD_SPEED
	tv1 = get_utimeofday();
#endif
	/* Perform firmware data transfer */
	tx_len = (FIRMWARE_TRANSFER_NBLOCK * SD_BLOCK_SIZE) - 4;
	for (fwblknow = 0; fwblknow < firmwarelen; fwblknow += tx_len) {

		/* The host polls for the DN_LD_CARD_RDY and CARD_IO_READY bits */
		ret = mv_sdio_poll_card_status(priv, CARD_IO_READY|DN_LD_CARD_RDY);
		if (ret < 0) {
			PRINTM(FATAL, "Firmware download died @ %d\n", fwblknow);
			goto done;
		}

		/* Set blocksize to transfer - checking for last block */
		if (firmwarelen - fwblknow < tx_len)
			tx_len = firmwarelen - fwblknow;

		fwbuf[0] = ((tx_len & 0x000000ff) >>  0);  /* Little-endian */
		fwbuf[1] = ((tx_len & 0x0000ff00) >>  8);
		fwbuf[2] = ((tx_len & 0x00ff0000) >> 16);
		fwbuf[3] = ((tx_len & 0xff000000) >> 24);

		/* Copy payload to buffer */
		memcpy(&fwbuf[4], &firmware[fwblknow], tx_len);

		PRINTM(INFO, ".");

		/* Send data */
		ret = sdio_write_iomem(priv->wlan_dev.card, FN1,
					priv->wlan_dev.ioport,  
					FIXED_ADDRESS, FIRMWARE_TRANSFER_NBLOCK,
					SD_BLOCK_SIZE, fwbuf);

		if (ret < 0) {
			PRINTM(FATAL, "IO error: transferring block @ %d\n", fwblknow);
			goto done;
		}
  	}

	PRINTM(INFO, "\ndone (%d/%d bytes)\n", fwblknow, firmwarelen);

#ifdef FW_DOWNLOAD_SPEED
	tv2 = get_utimeofday();
	PRINTM(INFO, "helper: %ld.%03ld.%03ld ",tv1 / 1000000, (tv1 % 1000000) / 1000,
								tv1 % 1000);
	PRINTM(INFO, " -> %ld.%03ld.%03ld ", tv2 / 1000000, (tv2 % 1000000)/1000,
								tv2 % 1000);
	tv2 -= tv1;
	PRINTM(INFO, " == %ld.%03ld.%03ld\n", tv2 / 1000000, (tv2 % 1000000) / 1000,
								tv2 % 1000);
#endif
  
	/* Write last EOF data */
	PRINTM(INFO, "Transferring EOF block\n");
	memset(fwbuf, 0x0, sizeof(fwbuf));
	ret = sdio_write_iomem((struct mss_card *) priv->wlan_dev.card, FN1, 
				priv->wlan_dev.ioport,  
				FIXED_ADDRESS, 1, SD_BLOCK_SIZE, fwbuf);

	if (ret < 0) {
		PRINTM(FATAL, "IO error in writing EOF firmware block\n");
		goto done;
	}

	ret = WLAN_STATUS_SUCCESS;

done:
	return ret;
}

#ifdef HELPER_IMAGE
/** 
 *  @brief This function downloads firmware image to the card.
 *  
 *  @param priv    	A pointer to wlan_private structure
 *  @param firmware	A pointer to firmware image buffer
 *  @param firmwarelen	the length of firmware image
 *  @return 	   	WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
static int sbi_download_wlan_fw_image(wlan_private *priv,
			const u8 *firmware, int firmwarelen)
{
	u8	base0;
	u8	base1;
	int	ret = WLAN_STATUS_SUCCESS;
	int	offset;
	int	fwblknow = 0;
	u64	buf[FIRMWARE_TRANSFER_NBLOCK * 32];
	u8	*fwbuf;
	int	timeout = 5000;
	u16	len;
	int	txlen = 0;
	int	tx_len = 0;
#ifdef FW_DOWNLOAD_SPEED
	u32	tv1,tv2;
#endif
	
	ENTER();

	fwbuf = (u8 *)buf;
	PRINTM(INFO, "wbuf size is :0x%x\n", fwbuf);
	PRINTM(INFO, "WLAN_FW: Downloading firmware of size %d bytes\n", firmwarelen);

#ifdef FW_DOWNLOAD_SPEED
	tv1 = get_utimeofday();
#endif

	/* Wait initially for the first non-zero value */
	do {
		if((ret = sdio_read_ioreg(priv->wlan_dev.card, FN1, 
					HOST_F1_RD_BASE_0, &base0))) {
			PRINTM(WARN, "Dev BASE0 register read failed:"
					" base0=0x%04X(%d)\n", base0, base0);
			ret = WLAN_STATUS_FAILURE;
			goto done;
		}
		if((ret = sdio_read_ioreg(priv->wlan_dev.card, FN1, 
					HOST_F1_RD_BASE_1, &base1))) {
			PRINTM(WARN, "Dev BASE1 register read failed:"
					" base1=0x%04X(%d)\n", base1, base1);
			ret = WLAN_STATUS_FAILURE;
			goto done;
		}
		len = (((u16) base1) << 8) | base0;
		mdelay(1);
	} while (!len && --timeout);
		
	if (!timeout) {
		PRINTM(MSG, "Timeout waiting for initial non-zero length\n");
		ret = WLAN_STATUS_FAILURE;
		goto done;
	}
	
	/* The host polls for the DN_LD_CARD_RDY and CARD_IO_READY bits */
	ret = mv_sdio_poll_card_status(priv,CARD_IO_READY|DN_LD_CARD_RDY);
	if (ret < 0) {
		PRINTM(FATAL, "Firmware download died @ the end\n");
		ret = WLAN_STATUS_FAILURE;
		goto done;
	}
	
	PRINTM(INFO, "WLAN_FW: Len got from firmware = 0x%04X(%d)\n", 
								len, len);
	len &= ~B_BIT_0;

	/* Perform firmware data transfer */
	for (offset = 0; offset < firmwarelen; offset += txlen) {
		txlen = len;
		
		/* Set blocksize to transfer - checking for last block */
		if (firmwarelen - offset < txlen) {
			txlen = firmwarelen - offset;
		}
		PRINTM(INFO, "WLAN_FW: offset=%d, txlen = 0x%04X(%d)\n", 
							offset,txlen,txlen);
	
 	    	tx_len = (FIRMWARE_TRANSFER_NBLOCK * SD_BLOCK_SIZE * 8);

    		for (fwblknow = 0; fwblknow < txlen; fwblknow += tx_len) {
			/* The host polls for the DN_LD_CARD_RDY and CARD_IO_READY bits */
			ret = mv_sdio_poll_card_status(priv, CARD_IO_READY|DN_LD_CARD_RDY);
			if (ret < 0) {
				PRINTM(FATAL, "Firmware download died @ %d\n", fwblknow);
				goto done;
			}
			if (txlen - fwblknow < tx_len)
				tx_len = txlen - fwblknow;
			PRINTM(INFO, "WLAN_FW:     fwblknow=%d, tx_len = 0x%04X(%d)\n", 
							offset,tx_len,tx_len);
			/* Copy payload to buffer */
			memcpy(fwbuf, &firmware[offset+fwblknow], tx_len);

			/* Send data */
			ret = sdio_write_iomem(priv->wlan_dev.card, FN1,
				priv->wlan_dev.ioport, FIXED_ADDRESS,
					FIRMWARE_TRANSFER_NBLOCK*8, SD_BLOCK_SIZE, fwbuf);

			if (ret < 0) {
				PRINTM(FATAL, "IO error:transferring @ %d\n", offset+fwblknow);
				goto done;
			}
	    	}
		
		do {
			udelay(10);
			if((ret = sdio_read_ioreg(priv->wlan_dev.card, FN1, 
					HOST_F1_CARD_RDY, &base0)) < 0) {
				PRINTM(WARN, "Dev CARD_RDY register read failed:"
					" base0=0x%04X(%d)\n",base0,base0);
				ret = WLAN_STATUS_FAILURE;
				goto done;
			}
			PRINTM(INFO, "offset=0x%08X len=0x%04X: FN1,"
					"HOST_F1_CARD_RDY: 0x%04X\n",
					offset, txlen, base0);
		} while(!(base0 & 0x08) || !(base0 & 0x01));
		
		if((ret = sdio_read_ioreg(priv->wlan_dev.card, FN1, 
					HOST_F1_RD_BASE_0, &base0)) < 0) {
			PRINTM(WARN, "Dev BASE0 register read failed:"
					" base0=0x%04X(%d)\n",base0,base0);
			ret = WLAN_STATUS_FAILURE;
			goto done;
		}
		if((ret = sdio_read_ioreg(priv->wlan_dev.card, FN1, 
					HOST_F1_RD_BASE_1, &base1)) < 0) {
			PRINTM(WARN, "Dev BASE1 register read failed:"
					" base1=0x%04X(%d)\n",base1,base1);
			ret = WLAN_STATUS_FAILURE;
			goto done;
		}
		len = (((u16) base1) << 8) | base0;

		if (!len) {
			PRINTM(INFO, "WLAN Firmware Download Over\n");
			break;
		}
		
		if (len & B_BIT_0) {
			PRINTM(FATAL, "CRC32 Error indicated by the helper:"
					" len=0x%04X(%d)\n",len,len);
			len &= ~B_BIT_0;
			/* Setting this to 0 to resend from same offset */
			txlen = 0;
		} else {
			PRINTM(INFO, "%d,%d bytes block of firmware downloaded\n",
							offset, txlen);
		}
  	}

	PRINTM(INFO, "Firmware Image of Size %d bytes downloaded\n", firmwarelen);
  
	ret = WLAN_STATUS_SUCCESS;
done:
#ifdef FW_DOWNLOAD_SPEED
	tv2 = get_utimeofday();
	PRINTM(INFO, "firmware: %ld.%03ld.%03ld ", tv1 / 1000000, 
						(tv1 % 1000000) / 1000,
						tv1 % 1000);
	PRINTM(INFO, " -> %ld.%03ld.%03ld ", tv2 / 1000000, 
						(tv2 % 1000000) / 1000,
						tv2%1000);
	tv2 -= tv1;
	PRINTM(INFO, " == %ld.%03ld.%03ld\n", tv2 / 1000000,
						(tv2 % 1000000) / 1000,
						tv2 % 1000);
#endif
	LEAVE();
	return ret;
}
#endif

/** 
 *  @brief This function reads data from the card.
 *  
 *  @param priv    	A pointer to wlan_private structure
 *  @param type	   	A pointer to keep type as data or command
 *  @param nb		A pointer to keep the data/cmd length retured in buffer
 *  @param payload 	A pointer to the data/cmd buffer
 *  @param nb	   	the length of data/cmd buffer
 *  @param npayload	the length of data/cmd buffer
 *  @return 	   	WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
static int mv_sdio_card_to_host(wlan_private *priv,
			 u32 *type, int *nb, u8 *payload, int npayload)
{
	int ret = WLAN_STATUS_SUCCESS;
	u16 buf_len = 0;
	int buf_block_len;
	int blksz;
	//struct mss_card * card = (struct mss_card *) priv->wlan_dev.card;
	
	if (!payload) {
		PRINTM(WARN, "payload NULL pointer received!\n");
		return WLAN_STATUS_FAILURE;
	}

	/* Read the length of data to be transferred */
	ret = mv_sdio_read_scratch(priv, &buf_len);
	if (ret < 0) {
		PRINTM(WARN, "Failed to read the scratch reg\n");
		return WLAN_STATUS_FAILURE;
	}

	PRINTM(INFO, "Receiving %d bytes from card at scratch reg value\n", buf_len);

	if (((buf_len - SDIO_HEADER_LEN) <= 0) || (buf_len > (MRVDRV_ETH_RX_PACKET_BUFFER_SIZE + SDIO_HEADER_LEN))) {
		PRINTM(WARN, "Invalid packet size from firmware, size = %d\n", 
								buf_len);
		return WLAN_STATUS_FAILURE;
	}
  
	/* Allocate buffer */
	if(1) {//card->block_size_512) {
		blksz = SD_BLOCK_SIZE_320;
		buf_block_len = (buf_len + blksz - 1) / blksz;
	}
	else {
		blksz = SD_BLOCK_SIZE;
		buf_block_len = (buf_len + blksz - 1) / blksz;
	}

	/* The host polls for the CARD_IO_READY bit */
	ret = mv_sdio_poll_card_status(priv, CARD_IO_READY);
	if (ret < 0) {
		PRINTM(WARN, "Poll failed in card_to_host : %d\n", ret);
		return WLAN_STATUS_FAILURE;
	}

 	ret = sdio_read_iomem(priv->wlan_dev.card, FN1, priv->wlan_dev.ioport,
				FIXED_ADDRESS, buf_block_len,
				blksz, payload);

	if (ret < 0) {
		PRINTM(WARN, "sdio_read_iomem failed - mv_sdio_card_to_host\n") ;
		return WLAN_STATUS_FAILURE;
	}
	*nb = buf_len;

	if (*nb <= 0) {
		PRINTM(WARN, "Null packet recieved \n");
		return WLAN_STATUS_FAILURE;
	}

	*type  = (payload[2] | (payload[3] << 8));
	
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief This function enables the host interrupts mask
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @param mask	   the interrupt mask
 *  @return 	   WLAN_STATUS_SUCCESS
 */
static int enable_host_int_mask (wlan_private *priv , u8 mask)
{
	int	ret = WLAN_STATUS_SUCCESS;
	struct mss_card *card = (struct mss_card *) priv->wlan_dev.card;

	/* Simply write the mask to the register */
	ret = sdio_write_ioreg(card, FN1, HOST_INT_MASK_REG, mask);

	if(ret < 0) {
		PRINTM(WARN, "ret = %d\n", ret);
		ret = WLAN_STATUS_FAILURE;
	}

	priv->adapter->HisRegCpy = 1;

	return ret;
}

/**  @brief This function disables the host interrupts mask.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @param mask	   the interrupt mask
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
static int disable_host_int_mask(wlan_private *priv,u8 mask)
{
	int	ret = WLAN_STATUS_SUCCESS;
	u8	host_int_mask;

	/* Read back the host_int_mask register */
	ret = sdio_read_ioreg(priv->wlan_dev.card, FN1, HOST_INT_MASK_REG, 
							&host_int_mask);
	if (ret < 0) {
		ret = WLAN_STATUS_FAILURE;
		goto done;
	}

	/* Update with the mask and write back to the register */
	host_int_mask &= ~mask;
	ret = sdio_write_ioreg(priv->wlan_dev.card, FN1, HOST_INT_MASK_REG, 
							host_int_mask);
	if(ret < 0) {
		PRINTM(WARN, "Unable to diable the host interrupt!\n");
		ret = WLAN_STATUS_FAILURE;
		goto done;
	}

done:
	return ret;
}
/********************************************************
		Global Functions
********************************************************/
/** 
 *  @brief This function handles the interrupt.
 *  
 *  @param irq 	   The irq of device.
 *  @param dev_id  A pointer to net_device structure
 *  @param fp	   A pointer to pt_regs structure
 *  @return 	   n/a
 */
void sbi_interrupt(struct mss_card *card)
{
	wlan_private   		*priv;

	priv = dev_get_drvdata(&card->dev);

	PRINTM(INFO, "sbi_interrupt called IntCounter= %d\n",
					priv->adapter->IntCounter);

#ifdef FW_WAKEUP_TIME
	if((wt_pwrup_sending!=0L) && (wt_int==0L))
		wt_int = get_utimeofday();
#endif

	wlan_interrupt(priv);
}

/** 
 *  @brief This function probes the card.
 *  
 *  @param card_p  A pointer to the card
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
static int sbi_probe_card(struct device *dev)
{
	int		ret = WLAN_STATUS_SUCCESS;
	struct mss_card *card;
	struct sdio_card *sdio_card;
	u8		bic;
	u8		chiprev;

	card = container_of(dev, struct mss_card, dev);
	if (card->card_type != MSS_SDIO_CARD)
		return 0;
	sdio_card = card->prot_card;
	ENTER();  

	if (!card) {
		ret = WLAN_STATUS_FAILURE;
	       	goto done;
	}

	/* Check for MANFID */
	ret = sdio_card->ccis.manufacturer;

#define MARVELL_VENDOR_ID 0x02df	
	if (ret == MARVELL_VENDOR_ID) {
		printk("Marvell SDIO card detected!\n");
	} else {
		PRINTM(MSG, "Ignoring a non-Marvell SDIO card...\n");
		ret = WLAN_STATUS_FAILURE;
		goto done;
	}

	/* read Revision Register to get the hw revision number */
	if (sdio_read_ioreg(card, FN1, CARD_REVISION_REG, &chiprev) < 0) {
		PRINTM(FATAL, "cannot read CARD_REVISION_REG\n");
	} else {
		printk("revsion=0x%x\n",chiprev);
		switch(chiprev) {
		case 0x00:	/* SD8385-A0 */
			PRINTM(FATAL, "SD8385-A0 not supported\n");
			ret = WLAN_STATUS_FAILURE;
			goto done;
		case 0x10:	/* SD8385-B0 */
		case 0x12:	/* SD8385-B1 */
		case 0x13:	/* SD8385-B2 */
		default:
		/*	card->block_size_512 = TRUE;
			card->async_int_mode = TRUE;
		*/
			/* enable async interrupt mode */
			sdio_read_ioreg(card, FN0, BUS_IF_CTRL,	&bic);
			sdio_write_ioreg(card, FN0, BUS_IF_CTRL, bic | 0x20);
			break;
		}
	}

	if (!fmimage) {
		ret = WLAN_STATUS_FAILURE;
		goto done;
	}

	ret = WLAN_STATUS_SUCCESS;
done:
	LEAVE();
	if (!wlan_add_callback)
		return WLAN_STATUS_FAILURE;

	pwlanpriv = wlan_add_callback(card);
	if(!pwlanpriv)
		return WLAN_STATUS_FAILURE;

	pwlanpriv->adapter->chip_rev = chiprev;
	//dev_set_drvdata(dev, pwlanpriv);
	return ret;
}

static int sbi_remove_card(struct device *dev)
{
	struct mss_card *card;

	card = container_of(dev, struct mss_card, dev);
	if (!wlan_remove_callback)
		return WLAN_STATUS_FAILURE;
		
	pwlanpriv = NULL;
	mss_set_sdio_int(card->slot->host, 0);
	return wlan_remove_callback(card); 
}

static struct mss_driver wlan_driver = {
	.driver = {
		.name = "wlandriver",
		.probe = sbi_probe_card,
		.remove = sbi_remove_card,
	},
	.sdio_int_handler = sbi_interrupt,
};


/** 
 *  @brief This function registers the IF module in bus driver.
 *  
 *  @param add	   wlan driver's call back funtion for add card.
 *  @param remove  wlan driver's call back funtion for remove card.
 *  @param arg     not been used
 *  @return	   An int pointer that keeps returned value
 */
int *sbi_register(wlan_notifier_fn_add add, wlan_notifier_fn_remove remove, 
								void *arg)
{
	int *sdio_ret;

	wlan_add_callback = add;
	wlan_remove_callback = remove;

	register_mss_driver(&wlan_driver);
	return 0;
}

/** 
 *  @brief This function de-registers the IF module in bus driver.
 *  
 *  @return 	   n/a
 */
void sbi_unregister(void)
{
	unregister_mss_driver(&wlan_driver);
	wlan_add_callback = NULL;
	wlan_remove_callback = NULL;
}

/** 
 *  @brief This function reads the IO register.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @param func	   funcion number
 *  @param reg	   register to be read
 *  @param dat	   A pointer to variable that keeps returned value
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_read_ioreg(wlan_private *priv, u8 func, u32 reg, u8 *dat)
{
	return sdio_read_ioreg((struct mss_card *)priv->wlan_dev.card, func, reg, dat);
}

/** 
 *  @brief This function writes the IO register.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @param func	   funcion number
 *  @param reg	   register to be written
 *  @param dat	   the value to be written
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_write_ioreg(wlan_private *priv, u8 func, u32 reg, u8 dat)
{
	return sdio_write_ioreg((struct mss_card *)priv->wlan_dev.card, func, reg, dat);
}

/** 
 *  @brief This function checks the interrupt status and handle it accordingly.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @param ireg    A pointer to variable that keeps returned value
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_get_int_status(wlan_private *priv , u8 *ireg)
{
	int 		ret = WLAN_STATUS_SUCCESS;
	u8  		*cmdBuf;
	wlan_dev_t	*wlan_dev = &priv->wlan_dev;
	struct mss_card *card = (struct mss_card *) wlan_dev->card;
	struct sk_buff	*skb;
	
	if ((ret = sdio_read_ioreg( card, FN1, HOST_INTSTATUS_REG, ireg)) < 0) {
		PRINTM(WARN, "sdio_read_ioreg: reading interrupt status register"
								" failed\n");
		ret = WLAN_STATUS_FAILURE;
		goto done;
	}

	if(*ireg != 0) {
		/*
		 * DN_LD_HOST_INT_STATUS and/or UP_LD_HOST_INT_STATUS
		 * Clear the interrupt status register
		 */
		PRINTM(INFO, "ireg = 0x%x, ~ireg = 0x%x\n",*ireg,~(*ireg));
		if ((ret = sdio_write_ioreg(card, FN1, HOST_INTSTATUS_REG, 
			~(*ireg) & (DN_LD_HOST_INT_STATUS | UP_LD_HOST_INT_STATUS))) < 0) {
			PRINTM(WARN, "sdio_write_ioreg: clear interrupt status" 
							" register failed\n");
			ret = WLAN_STATUS_FAILURE;
			goto done;
		}
	}
	

	if (*ireg & DN_LD_HOST_INT_STATUS) {	/* tx_done INT */
		*ireg |= HIS_TxDnLdRdy;
		if (!priv->wlan_dev.dnld_sent) { /* tx_done already received */
			PRINTM(INFO, "warning: tx_done already received:"
					" dnld_sent=0x%x ireg=0x%x\n",
					priv->wlan_dev.dnld_sent,*ireg);
		} else {
			PRINTM(INFO, "tx_done received: dnld_sent=0x%x ireg=0x%x\n",
					priv->wlan_dev.dnld_sent, *ireg);
			priv->wlan_dev.dnld_sent = DNLD_RES_RECEIVED;
		}
	}

	if (
#ifdef THROUGHPUT_TEST
	(
#endif /* THROUGHPUT_TEST */
		*ireg & UP_LD_HOST_INT_STATUS
#ifdef THROUGHPUT_TEST
		||  !priv->adapter->NumTransfersRx) && 
					(priv->adapter->ThruputTest & 0x02)
#endif /* THROUGHPUT_TEST */
	) {
		/* 
		 * DMA read data is by block alignment,so we need alloc extra block
		 * to avoid wrong memory access.
		 */
		if(!(skb = dev_alloc_skb(MRVDRV_ETH_RX_PACKET_BUFFER_SIZE+SD_BLOCK_SIZE_512 + 8))) {
			PRINTM(WARN, "No free skb\n");
			priv->stats.rx_dropped++;
			return WLAN_STATUS_FAILURE;
		}

		if ((u32)skb->tail & 0x7) {
			skb_put(skb, (u32)skb->tail & 7);
			skb_pull(skb, (u32)skb->tail & 7);
		}
		/* 
		 * Transfer data from card
		 * skb->tail is passed as we are calling skb_put after we
		 * are reading the data
		 */
		if (mv_sdio_card_to_host(priv,&wlan_dev->upld_typ,
			(int *)&wlan_dev->upld_len, skb->tail,
					MRVDRV_ETH_RX_PACKET_BUFFER_SIZE+SD_BLOCK_SIZE_512) < 0) {
			u8		cr = 0;

			PRINTM(WARN, "Card to host failed: ireg=0x%x\n", *ireg);
			if (sdio_read_ioreg(wlan_dev->card, FN1,
						CONFIGURATION_REG,&cr) < 0)
				PRINTM(WARN, "sdio_read_ioreg failed\n");

			PRINTM(INFO, "Config Reg val = %d\n",cr);
			if (sdio_write_ioreg(wlan_dev->card, FN1,
					CONFIGURATION_REG, (cr | 0x04)) < 0)
				PRINTM(WARN, "write ioreg failed\n");

			PRINTM(INFO, "write success\n");
			if (sdio_read_ioreg(wlan_dev->card, FN1,
					CONFIGURATION_REG, &cr) < 0 )
				PRINTM(WARN, "sdio_read_ioreg failed\n");

			PRINTM(INFO, "Config reg val =%x\n",cr);
			ret = WLAN_STATUS_FAILURE;
			kfree_skb(skb);
			goto done;
		}
			
		if (wlan_dev->upld_typ == MVSD_DAT) {
			*ireg |= HIS_RxUpLdRdy;
			skb_put(skb, priv->wlan_dev.upld_len);
			skb_pull(skb, SDIO_HEADER_LEN);
			list_add_tail((struct list_head *) skb,
				(struct list_head *) &priv->adapter->RxSkbQ);
		} else if (wlan_dev->upld_typ == MVSD_CMD) {
			*ireg &= ~(HIS_RxUpLdRdy);
			*ireg |= HIS_CmdUpLdRdy;

			/* take care of CurCmd = NULL case by reading the 
			 * data to clear the interrupt */
			if(!priv->adapter->CurCmd) {
				cmdBuf = priv->wlan_dev.upld_buf;
				priv->adapter->HisRegCpy &= ~HIS_CmdUpLdRdy;
			} else {
#ifdef WLAN_LOOPBACK
				cmdBuf = priv->wlan_dev.upld_buf;
#else
				cmdBuf = priv->adapter->CurCmd->BufVirtualAddr;
#endif
			}
			
			priv->wlan_dev.upld_len -= SDIO_HEADER_LEN;
			memcpy(cmdBuf, skb->data + SDIO_HEADER_LEN,
				       		priv->wlan_dev.upld_len);
			kfree_skb(skb);
		}
		else if (wlan_dev->upld_typ == MVSD_EVENT) {
			*ireg |= HIS_CardEvent;
			kfree_skb(skb);
		}
		
		*ireg |= HIS_CmdDnLdRdy;
	}

	ret = WLAN_STATUS_SUCCESS;
done:

	sbi_reenable_host_interrupt(priv, 0);
	return ret;
}

/** 
 *  @brief This function is a dummy function.
 *  
 *  @return 	   WLAN_STATUS_SUCCESS
 */
int sbi_card_to_host(wlan_private *priv, u32 type,
	       	u32 *nb, u8 *payload, u16 npayload)
{
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief This function is a dummy function.
 *  
 *  @return 	   WLAN_STATUS_SUCCESS
 */
int sbi_read_event_cause(wlan_private *priv)
{
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief This function reenables the host interrupts.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @param bits    the bit mask
 *  @return 	   WLAN_STATUS_SUCCESS
 */
int sbi_reenable_host_interrupt(wlan_private *priv, u8 bits)
{
	mss_set_sdio_int(((struct mss_card *)((priv->wlan_dev).card))->slot->host, 1);
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief This function disables the host interrupts.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_disable_host_int(wlan_private *priv)
{
	return disable_host_int_mask(priv,HIM_DISABLE);
}

/** 
 *  @brief This function enables the host interrupts.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS
 */
int sbi_enable_host_int (wlan_private *priv)
{

	return enable_host_int_mask(priv,HIM_ENABLE); 
}

/** 
 *  @brief This function de-registers the device.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS
 */
int sbi_unregister_dev(wlan_private *priv)
{
	struct mss_card *card = priv->wlan_dev.card;
	
	ENTER();

	if (priv->wlan_dev.card != NULL) {
		/* Release the SDIO IRQ */
		//sdio_free_irq(priv->wlan_dev.card, priv->wlan_dev.netdev);
		PRINTM(WARN, "Making the sdio dev card as NULL\n");
	}
	dev_set_drvdata(&card->dev, NULL);
	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief This function registers the device.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_register_dev(wlan_private *priv)
{
	int	ret = WLAN_STATUS_SUCCESS;
	struct mss_card *card = priv->wlan_dev.card;
	u8	reg;
	//struct mss_card *card = (struct mss_card *) priv->wlan_dev.card;
	ENTER();  

	dev_set_drvdata(&card->dev, priv);
	/* Initialize the private structure */
	strncpy(priv->wlan_dev.name, "sdio0", sizeof(priv->wlan_dev.name));
	priv->wlan_dev.ioport = 0;
	priv->wlan_dev.upld_rcv = 0;
	priv->wlan_dev.upld_typ = 0;
	priv->wlan_dev.upld_len = 0;

	/* Read the IO port */
	ret = sdio_read_ioreg(priv->wlan_dev.card, FN1, IO_PORT_0_REG, &reg);
	if (ret < 0)
		goto failed;
	else
		priv->wlan_dev.ioport |= reg;

	ret = sdio_read_ioreg(priv->wlan_dev.card, FN1, IO_PORT_1_REG, &reg);
	if (ret < 0)
		goto failed;
	else
		priv->wlan_dev.ioport |= (reg << 8);

	ret = sdio_read_ioreg(priv->wlan_dev.card, FN1, IO_PORT_2_REG, &reg);
	if (ret < 0)
		goto failed;
	else
		priv->wlan_dev.ioport |= (reg << 16);

	PRINTM(INFO, "SDIO FUNC1 IO port: 0x%x\n", priv->wlan_dev.ioport);

	/* Disable host interrupt first. */
	if ((ret = disable_host_int_mask(priv, 0xff)) < 0) {
		PRINTM(WARN, "Warning: unable to disable host interrupt!\n");
	}
	return WLAN_STATUS_SUCCESS;

failed:
	priv->wlan_dev.card = NULL;

	return WLAN_STATUS_FAILURE;
}

/** 
 *  @brief This function sends data to the card.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @param type	   data or command
 *  @param payload A pointer to the data/cmd buffer
 *  @param nb	   the length of data/cmd
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_host_to_card(wlan_private *priv, u8 type, u8 *payload, u16 nb)
{
	int	ret = WLAN_STATUS_SUCCESS;
	int	buf_block_len;
	int	blksz;
	//struct mss_card *card = (struct mss_card *) priv->wlan_dev.card;

	ENTER();

	priv->adapter->HisRegCpy = 0;

	/* Allocate buffer and copy payload */
	if(1) {//card->block_size_512) {
		blksz = SD_BLOCK_SIZE_320;
		buf_block_len = (nb + SDIO_HEADER_LEN + blksz - 1) / blksz;
	}
	else {
		blksz = SD_BLOCK_SIZE;
		buf_block_len = (nb + SDIO_HEADER_LEN + blksz - 1) / blksz;
	}

	priv->adapter->TmpTxBuf[0] = (nb + SDIO_HEADER_LEN) & 0xff;
	priv->adapter->TmpTxBuf[1] = ((nb + SDIO_HEADER_LEN) >> 8) & 0xff;
	priv->adapter->TmpTxBuf[2] = (type == MVSD_DAT) ? 0 : 1;
	priv->adapter->TmpTxBuf[3] = 0x0;

	if (payload != NULL && nb > 0) {
		memcpy(&priv->adapter->TmpTxBuf[SDIO_HEADER_LEN], payload, nb);
	}
	else {
		PRINTM(WARN, "sbi_host_to_card(): Error: payload=%p, nb=%d\n",
								payload,nb);
	}
  
	/* The host polls for the CARD_IO_READY bit */
	ret = mv_sdio_poll_card_status(priv, CARD_IO_READY);
	if (ret < 0) {
		PRINTM(WARN, "Poll failed in host_to_card : %d\n", ret);
		return WLAN_STATUS_FAILURE;
	}

	/* Transfer data to card */
	ret = sdio_write_iomem(priv->wlan_dev.card, FN1, priv->wlan_dev.ioport,
			FIXED_ADDRESS, buf_block_len,
			blksz, (u8 *)priv->adapter->TmpTxBuf);
	if (ret < 0) {
		PRINTM(WARN, "sdio_write_iomem failed: ret=%d\n", ret);
		return WLAN_STATUS_FAILURE;
	}
	else {
		PRINTM(INFO, "sdio write -dnld val =>%d\n",ret);
	}

	if(type == MVSD_DAT)
		priv->wlan_dev.dnld_sent = DNLD_DATA_SENT;
	else
		priv->wlan_dev.dnld_sent = DNLD_CMD_SENT;
	
	LEAVE();
	return ret;
}

/** 
 *  @brief This function reads CIS informaion.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_get_cis_info(wlan_private *priv)
{
	struct mss_card *card = (struct mss_card *) priv->wlan_dev.card;
	wlan_adapter	*Adapter = priv->adapter;
	u8 		tupledata[255];
	char 		cisbuf[512];
	int 		ofs=0, i;
	u32 		ret = WLAN_STATUS_SUCCESS;

	ENTER();
	
	/* Read the Tuple Data */
  	for (i = 0; i < sizeof(tupledata); i++) {
		ret = sdio_read_ioreg(card, FN0, 
			((struct sdio_card *)(card->prot_card))->cccr.pccis + i,
							&tupledata[i]);
		if (ret < 0)
			return WLAN_STATUS_FAILURE;
	}

	memset(cisbuf, 0x0, sizeof(cisbuf));
    	memcpy(cisbuf+ofs, tupledata, sizeof(cisbuf));
	
	/* Copy the CIS Table to Adapter */
	memset(Adapter->CisInfoBuf, 0x0, sizeof(cisbuf));
	memcpy(Adapter->CisInfoBuf, &cisbuf, sizeof(cisbuf));
	Adapter->CisInfoLen = sizeof(cisbuf);

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


#ifdef HELPER_IMAGE
/** 
 *  @brief This function calls sbi_download_wlan_fw_image to download
 *  firmware image to the card.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_prog_firmware_w_helper(wlan_private *priv)
{
	return sbi_download_wlan_fw_image(priv,fmimage,sizeof(fmimage));
}
#endif

/** 
 *  @brief This function calls sbi_prog_firmware_image to program
 *  firmware image to the card.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_prog_firmware(wlan_private *priv)
{
	return sbi_prog_firmware_image(priv,fmimage,sizeof(fmimage));
}

#ifdef HELPER_IMAGE
/** 
 *  @brief This function programs helper image.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_prog_helper(wlan_private *priv)
{
	return sbi_prog_firmware_image(priv,helperimage,sizeof(helperimage));
}
#endif

/** 
 *  @brief This function checks if the firmware is ready to accept
 *  command or not.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_verify_fw_download(wlan_private *priv)
{
	int	ret = WLAN_STATUS_SUCCESS;
	u16	firmwarestat;
	int	tries;
	u8	rsr;

	/* Wait for firmware initialization event */
	for (tries = 0; tries < MAX_FIRMWARE_POLL_TRIES; tries++) {
		if ((ret = mv_sdio_read_scratch(priv, &firmwarestat)) < 0)
			continue;

		if (firmwarestat == FIRMWARE_READY) {
			ret = WLAN_STATUS_SUCCESS;
			PRINTM(INFO, "Firmware successfully downloaded\n");
			break;
		} else {
			mdelay(10);
			ret = WLAN_STATUS_FAILURE;
		}
	}

	if (ret < 0) {
		PRINTM(MSG, "Timeout waiting for firmware to become active\n");
		goto done;
	}

	ret = sdio_read_ioreg(priv->wlan_dev.card, FN1, HOST_INT_RSR_REG, &rsr);
	if (ret < 0) {
		PRINTM(WARN, "sdio_read_ioreg: reading INT RSR register failed\n");
		return WLAN_STATUS_FAILURE;
	} else
		PRINTM(INFO, "sdio_read_ioreg: RSR register 0x%x\n",rsr);
	
	ret = WLAN_STATUS_SUCCESS;
done:
	return ret;
}

#ifdef DEEP_SLEEP
/** 
 *  @brief This function makes firmware exiting from deep sleep.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_exit_deep_sleep(wlan_private *priv)
{
	int ret = WLAN_STATUS_SUCCESS;
	struct mss_card *card = (struct mss_card *)((priv->wlan_dev).card);

	PRINTM(INFO, "Trying to wakeup device... Conn=%d IntC=%d PS_Mode=%d PS_State=%d\n",
		priv->adapter->MediaConnectStatus, priv->adapter->IntCounter,
		priv->adapter->PSMode, priv->adapter->PSState);			
	
//	start_bus_clock(((mmc_card_t)((priv->wlan_dev).card))->ctrlr);
	mss_set_clock(card->slot->host, card->slot->host->f_max);

#ifdef FW_WAKEUP_METHOD
	if (priv->adapter->fwWakeupMethod == WAKEUP_FW_THRU_GPIO) {
		GPIO_PORT_TO_LOW();
	}
	else
#endif
	ret = sdio_write_ioreg(priv->wlan_dev.card, FN1, CONFIGURATION_REG, 
								HOST_POWER_UP);
#ifdef FW_WAKEUP_TIME
	wt_pwrup_sent = get_utimeofday();
#endif

	return ret;
}

/** 
 *  @brief This function resets the setting of deep sleep.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_reset_deepsleep_wakeup(wlan_private *priv)
{
	ENTER();

	int ret = WLAN_STATUS_SUCCESS;

#ifdef FW_WAKEUP_METHOD
	if (priv->adapter->fwWakeupMethod == WAKEUP_FW_THRU_GPIO) {
		GPIO_PORT_TO_HIGH();
	}
	else
#endif
	ret = sdio_write_ioreg(priv->wlan_dev.card, FN1, CONFIGURATION_REG, 0);

	LEAVE();

	return ret;
}
#endif  /* DEEP_SLEEP */

#ifdef ENABLE_PM
/** 
 *  @brief This function suspends the device.
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
 */
int sbi_suspend(wlan_private *priv)
{
	return sdio_suspend(priv->wlan_dev.card);
}

/** 
 *  @brief This function resumes the device from sleep
 *  
 *  @param priv    A pointer to wlan_private structure
 *  @return 	   WLAN_STATUS_SUCCESS or WLAN_STATUS_FAILURE
*/
int sbi_resume(wlan_private *priv)
{
	return sdio_resume(priv->wlan_dev.card);
}
#endif

MODULE_LICENSE("GPL");
